//                              -*- Mode: C++ -*-
//:ECDcalIO.hh
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 9/4/97
// Last Update ........
//
#ifndef ECDCALIO_HH_
#define ECDCALIO_HH_
#include "care/ECALtypes.hh"
typedef unsigned short unshort;

/**
class  ECDcalIO.

It provides function to operate on the stream of bytes produced by
the DAQ system. Actually it has the possibility to write a simulation
of that stream, so that it can be used in the MC simulations.

USAGE of the class:
  <PRE>
  ECDcalIO dcal;                  // define the handler
  dcal.start(event);              // let the object know where the event is

  unsigned short head1, head2;       // bank headers
  dcal.getBankHeader(head1, head2);  // read them (if you are interested in)

  // Now loop on FED blocks
  unshort fed_id = dcal.nextFED(); // read the Front End Detector IDentifier
  while( fed_id>0) {
      // read the FED header (if you are interested in)     
      // it returns the number of ADC counts stored
      int nch = dcal.getFedHeader(siz, sfed_id, idmk, bx);
 
      // Now loop on ADC counts
      for(int i=0; i<nch; i++){
        int ofs = 0;               // used to store the channel number
                                   // inside the read-out board [0-31]
	unsigned short adc = 0;    // The ADC count stored
        dcal.getNextShort(ofs, adc);     // read them!
	                                 // mind: for not-sparsified
					 // data ofs values are continuous
					 // from 0 to nch (=siz/2)
					 // for sparsified data ofs values
					 // are increasing but not continuous
	// process the channel (fed_id, ofs) with value adc
      }

      fed_id = dcal.nextFED(); // read the next FED data block
      
  }

  <PRE>
*/

class ECDcalIO {
public:
  // Constructors
  //--
  /// Default constructor
  ECDcalIO();
  //
  //--
  ///  Start operations
  void start(EcalEvent_t* event);
  //--
  ///  Start operations
  void start();

  /* ****** Input functions *********** */

  //--
  /// Get the whole bank headers (2 shorts)
  inline void getBankHeader(unshort & head1, unshort & head2) const; 

  //--
  /// Get the FED header
  /// Returns the number of adc counts stored
  inline unsigned int getFedHeader(unshort & siz, unshort & fed_id, 
		   unshort & idmk, unshort & bx) const; 

  //--
  /// Get the next available short
  inline void getNextShort(int & offset, unshort & adc);

  /* ****** Output functions *********** */
  //--
  /// Put the whole bank headers (2 shorts)
  inline void putBankHeader(unshort head1, unshort head2) const; 
  //--
  /// Get the FED header
  /// Returns the number of adc counts stored
  inline unsigned int putFedHeader(unshort siz, unshort fed_id, 
		   unshort idmk, unshort bx) const; 
  /// put sparsified data
  inline void putNextShort(int ofs, unshort channel);
  /// put not-sparsified data
  inline void putNextShort(unshort channel);

  /// 
  inline int FEDdone();
  /* ****** Input/Output functions *********** */

  /// Get the next FED ID 
  /// Returns the FED ID 
  inline unshort nextFED();

  //--
  /// Test if the current FED block is sparsified or not
  inline bool sparse() const;

private:


  inline void putShort(unsigned char * p, unshort data) const ;
  inline unshort getShort(unsigned char * p) const ;

  //--
  /// Pointer to dcal
  unsigned char * dcalbank;
  /// pointer to the current fed
  unsigned char * fedblock;
  /// pointer to the current data item
  unsigned char * data;
  /// number of chars read/written in the current data block
  unsigned int nchar;
  /// lenght of the current data stream
  unsigned int lenght;  

};

#include "care/ECDcalIO.icc"
#endif // ECDCALIO_HH_
