
// *****************************************************************************
//
// source:    faLine.C
//
// type:      source code
//
// created:    7 Jul 97
// updated:   26 May 99 Andreas Gellrich
//
// author:    Andreas Gellrich
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: Andreas.Gellrich@desy.de
//
// symbol:    faLine()
//
// input:     void
// output:    int  flag = 0: farm mode off (normal Arte)
//                      = 1: farm mode on  (farm I/O)
//
// called by: atinit_()
//
// purpose:   evaluate line command arguments and set globals accordingly
//
// hints:     FARM mode can be enabled or disabled by giving either:
//            DAQ file name
//            EMG partition name
//
// changes:   line command arguments to steer rpm (AG: 25-Nov-98)
//            line command argument for delayed reading (AG: 26-Nov-98)
//            multiple file input (AG: 27-Nov-98)
//            daq format logging (AG: 1-Dec-98)
//            number of emg segmens to 8 (AG: 12 Jan 99)
//            more polishing (AG: 15 Jan 99)
//            print version description in help (AG: 21 Jan 99)
//            polish rpm handling (AG: 2 Feb 99)
//            submission date of sources (AG: 3 Feb 99)
//            increased default buffer size (AG: 11 Feb 99)
//            introduce argument to cycle until stop (AG: 12 Feb 99)
//            use same buffer size as everywhere else (AG: 17 Feb 99)
//            polishing (AG: 19 Feb 99)
//            init grr (AG: 25 Feb 99)
//            add flag for online display (AG: 25 Mar 99)
//            add flag to suppress log files (AG: 26 Mar 99)            
//            add -length:<event_buffer_length> (AG: 14 May 99)
//            change STAGE_DISK -> STAGE_PATH (AG: 26 May 99)
//
// *****************************************************************************

// includes
#include "arte/ArteFlag.h"
#include "farm/fa.hh"

// globals
int  FARM      = 0;          // default FARM mode
int  FAON      = 0;          // default FARM online=off
int  RPMINIT   = 0;          // init rpm
int  NOSPAR    = 1;          // default sparsification
int  FADEB     = 0;          // default NO debug mode
int  DAQLOG    = 0;          // daq format logging
int  CYCLE     = 0;          // cycling
int  DISPLAY   = 0;          // display
int  FAGRR     = 0;          // grr logger
char *PARNAME  = 0;          // default emg partition name
int  SEG_NUM   = 4;          // default emg number of segments
int  SEG_LEN   = 150000;     // default event buffer size
int  WAIT      = 0;          // delay for reading event
char *DBNAME   = 0;          // default data base file name
char *IN_FILE[50];           // input files (NO defaults)
char *HBSTAGE  = STAGE_PATH; // stage disk
int  OLD       = 0;          // old file mode
int  nFiles    = 0;          // number of files

// externals
extern int  nArg;
extern char **vArg;

//
// program code
//
int
faLine()
{
  // inits
  char *files = 0;
  char *ptr   = 0;

  // evaluate line command arguments arguments
  for( int i=1; i<nArg; i++ ){
    
    // help
    if( nArg - i ){
      if( !strcmp( vArg[i], "-help" ) ||
	  !strcmp( vArg[i], "-h" )    ||
	  !strcmp( vArg[i], "-H" )       ){
#ifdef _OUT_
	cout << "**********************************" << endl;
	cout << "* compiled: " << __DATE__ << " " << __TIME__ << " *" << endl;
	cout << "**********************************" << endl;
	cout << "usage: " << vArg[0] << " <kumac> inter/batch"              
	     << endl;
	cout << "       [-help][-h][-H]"                                      
	     << endl;
	cout << "       [-debug                       (DEFAULT=" << FADEB      << ")]"
	     << endl;
	cout << "       [-noprint]                    (DEFAULT=" << arteflag_.noprint    << ")]"
	     << endl;
	cout << "       FARM offline options:"
	     << endl;
	cout << "       [-f <new_file1,new_file2,...>  (DEFAULT=none)]" 
	     << endl;
	cout << "       [-length:<event_buffer_length> (DEFAULT=" << SEG_LEN    << ")]" 
	     << endl;
	cout << "       EMG options (FARM online mode):"
	     << endl;
	cout << "       [-EMGpart:<partition>          (DEFAULT=" << PARNAME    << ")]" 
	     << endl; 
	cout << "       [-EMGnum:<segments>            (DEFAULT=" << SEG_NUM    << ")]" 
	     << endl;
	cout << "       [-EMGlen:<length>              (DEFAULT=" << SEG_LEN    << ")]" 
	     << endl;
	cout << "       RPM initialization:"
	     << endl;
	cout << "       [-RPMinit                      (DEFAULT=" << RPMINIT    << ")]"
	     << endl;
	cout << "       [-Process:<process>] name      (DEFAULT=" << "none"     << ")]"
	     << endl;
	cout << "       [-Domain:<domain>] domain      (DEFAULT=" << "/OTHERS/" << ")]"
	     << endl;
	cout << "       Further options:"
	     << endl;
	cout << "       [-L4Ostage:<stage_disk>]       (DEFAULT=" << HBSTAGE    << ")]"
	     << endl;
	cout << "       [-L4Ogrr] error logger         (DEFAULT=" << FAGRR      << ")]"
	     << endl;
	cout << "       [-L4Ocycle] do NOT block emg   (DEFAULT=" << CYCLE      << ")]"
	     << endl;
	cout << "       [-L4Odisplay] online display   (DEFAULT=" << DISPLAY    << ")]"
	     << endl;
	cout << "       [-L4Owait:<seconds> read delay (DEFAULT=" << WAIT       << ")]"
	     << endl;
	cout << "       [-L4Atest][-L4Areal] logging   (DEFAULT=" << "none"     << ")]"
	     << endl;
#endif
	// terminate
	exit( ARTE_GOOD );
      }
    }

    if( nArg - i ){ 
      
      // debugging mode
      if( !strcmp( vArg[i], "-debug" ) ){
	FADEB = 1;
      }

      // debugging mode
      if( !strcmp( vArg[i], "-noprint" ) ){
	arteflag_.noprint = 1;
      }
      
      // rpm init
      if( !strcmp( vArg[i], "-RPMinit" ) ){
	RPMINIT = 1;
      }
      
      // rpm register
      if( !strncmp( vArg[i], "-Process:", 9 ) || 
          !strncmp( vArg[i], "-Domain:", 8 )     ){
	RPMINIT = 1;
      }

      // error logging
      if( !strcmp( vArg[i], "-L4Ogrr" ) ){
	FAGRR = 1;
      }
      
      // daq format logging
      if( !strcmp( vArg[i], "-L4Atest" )           ||
          !strcmp( vArg[i], "-L4Areal" )           ||
          !strncmp( vArg[i], "-L4Adetector:", 13 )    ){
	RPMINIT = 1;
	DAQLOG  = 1;
      }
      
      // delayed reading 
      if( !strncmp( vArg[i], "-L4Owait:", 9 ) ){
	WAIT = atoi( vArg[i] + 9 );
      }
      
      // cycling until file cycle.stop appears
      if( !strcmp( vArg[i], "-L4Ocycle" ) ){
	CYCLE = 1;
      }

      // cycling until file cycle.stop appears
      if( !strcmp( vArg[i], "-L4Odisplay" ) ){
	DISPLAY = 1;
      }

      // bugger size
      if( !strncmp( vArg[i], "-length:", 8 ) ){
	SEG_LEN = atoi( vArg[i] + 8 );
      }
      
      // emg partition name
      if( !strncmp( vArg[i], "-EMGpart:", 9 ) ){
	FARM    = 1; // farm mode
	FAON    = 1; // farm online mode
	PARNAME = new char[20];
	memset( PARNAME, 0, 20 );
	strcpy( PARNAME, vArg[i] + 9 );
      }
      
      // emg partition number
      if( !strncmp( vArg[i], "-EMGnum:", 8 ) ){
	SEG_NUM = atoi( vArg[i] + 8 );
      }
      
      // emg partition size
      if( !strncmp( vArg[i], "-EMGlen:", 8 ) ){
	SEG_LEN = atoi( vArg[i] + 8 );
      }

      // stage disk
      if( !strncmp( vArg[i], "-L4Ostage:", 10 ) ){
	HBSTAGE = new char[20];
	memset( HBSTAGE,0, 20 );
	strcpy( HBSTAGE, vArg[i] + 10 );
      }
    }
    
    // input file
    if( nArg - i - 1 ){ 
      if( !strcmp( vArg[i], "-x" ) || 
	  !strcmp( vArg[i], "-f" )    ){
	FARM = 1; // farm mode
        files = new char[strlen( vArg[i+1] ) + 1];
        strcpy( files,vArg[i+1] );        
        ptr = strtok( files, "," ); // find comma
        while( ptr ){
          IN_FILE[nFiles] = new char[strlen( ptr ) + 1];
          strcpy( IN_FILE[nFiles], ptr );
	  //cout << "n="    << nFiles + 1 << " file=" << ptr        << endl;
          nFiles++;
          ptr = strtok( NULL, "," );
	}
        delete[] files;
      }

      // old format
      if( !strcmp( vArg[i], "-x" ) ){
	OLD = 1; // old format (no logger)
      }
    }
  }
    
  // farm mode?
  if( !FARM ) return FARM;
  
  // print globals
  cout << endl;
  cout << "SETTINGS: "
       << "debug="        << FADEB   << " "
       << "rpm_init="     << RPMINIT << " "
       << "logging="      << DAQLOG  << " "
       << "delay="        << WAIT    << " "
       << "cycle="        << CYCLE   << " "
       << "grr="          << FAGRR   << " ";
  
  // farm online
  if( FAON ){
    cout << "partition="    << PARNAME  << " "
	 << "number="       << SEG_NUM  << " "
	 << "length="       << SEG_LEN  << endl;
  }
  // farm offline
  else {
    cout << "old="   << OLD     << " ";
    cout << "stage=" << HBSTAGE << " ";
    cout << endl;
    for( int i=0; i<nFiles; i++ ){
      cout << "          file" << i + 1 << "=" << IN_FILE[i] << endl;;
    }
    cout << endl;
  }

  // return  
  return FARM;
}
