#include "gean/ptlset.hh"
#include "ptset/detector_menu.hh"
#include "arte/ArteMessage.hh"
#include <fstream.h>
//#include <cstdlib> // for getenv(..)

// The HERA-B First Level Trigger Simulation

// Author V.Balagura (balagura@mail.desy.de)   (27.10.98).

Detector*  HERA_B_detector = 0;
HPT_Setup* hpt_inner_setup = 0;
HPT_Setup* hpt_outer_setup = 0;

/* The programmer should control whether to keep the *HERA_B_detector
   structure after filling PTGE banks or delete it.
   
   If you use only ARTE and do not need C++ Detector structure
   you should delete HERA_B_detector;
   otherwise, e.g. when you use FLTSIM, you should keep it since FLTSIM
   programs determine horizontal pretrigger rows, possible coincidences
   between different layers, the corresponding ROI at TC02 etc. using 
   this structure. 
   In the latter case you may delete HERA_B_detector after its last use.
   Or this memory will be freed automatically after the program 
   terminates.

   By default this structure is deleted and HERA_B_detector pointer is 
   set to zero.
   If you'd like to change this behaviour:

   1) copy ptlset_() definition to your own file, 
   2) comment out 2 lines in the end of this function:
      delete HERA_B_detector;
      HERA_B_detector = 0;
   3) compile it and link with the standard libraries. In this case your
   version of ptlset_() will be taken, overwriting the standard version 
   in the library.
   */

extern "C" {
  void ptlset_() {
    create_hpt_setup();
    if ( global_error() ) {
      arteMsg  << "can not create high-pt setup";   arteMsg.error();
    }
    fill_ARTE_hpt_geometry_table();

    delete HERA_B_detector; // if you'll need the C++ Detector structure 
    HERA_B_detector = 0;    // *HERA_B_detector: comment out these 2 lines

    cout << " === C++ Setup structure is converted to PTGE bank and ";
    if ( HERA_B_detector == 0 ) cout << "DELETED\n";
    else                        cout << "RETAINED\n";
  }
}

#include "gean/ZGESL.h"                 // to define LGESL_PTC;
#include "gean/zgvers.hh"               // to find the (HPT) directory name
#include <stdio.h>                      // for struct FILE and fileno()
FILE* utfilp(const char*, const char*); // where the file with the setup
                                        // geometry description is placed
// utfilp is defined in gean/utfilp.C; there is no corresponding header

FILE* get_setup_geometry_description_FILE
(const string& file_name_without_dir) {
  zgvers ZGVERS;
  string dir_name(ZGVERS.Get_vers_datafile(LGESL_PTC));
  return utfilp(file_name_without_dir.c_str(),dir_name.c_str()); 
}
  
void create_hpt_setup() {
  //  cout << "Creating High-Pt setup ...\n";
  HERA_B_detector = new Detector("HERA-B");
  DetectorMenu menu(HERA_B_detector);
  
  // Get geometry file:

  // string geom_file_name = string(getenv("HB")) + '/'
  // + "/geo/PTL/hpt_setup_geometry";

  string geom_file_name = "hpt_setup.dat";
  ifstream is(fileno(get_setup_geometry_description_FILE
		     (geom_file_name)));
  // here fileno(FILE*) returns file descriptor,
  // ifstream(int file_descriptor) associates istream with it
  if ( !is ) { 
    global_error()("can not open file " + geom_file_name); 
    return;
  }
  //  cout << "--> Geometry file " << geom_file_name << " opened\n";
  
  menu(&is);
  if ( global_error() ) return;

  cout << " === High-Pt setup C++ structure created:\n"
    " === ";
  bool first_hpt_setup = true;
  for (Detector::iterator iter=HERA_B_detector->begin(); 
       iter!=HERA_B_detector->end(); ++iter) {
    __SetupBase* ps = (*iter).second();
    if ( ps->type() == HPT_Setup::setup_type() ) {
      // print comma ',' between setups:
      if ( first_hpt_setup == false) cout << ", ";
      else first_hpt_setup = false;

      HPT_Setup* hps = (HPT_Setup*)ps;
      cout << '\"' << (*iter).first << "\" (" << hps->size() 
	   << " chambers, ";
      unsigned int ncells = 0;
      for (HPT_Setup::IChamber ich=hps->begin(); ich!=hps->end(); ++ich) {
	HPT_Chamber& ch = (*ich).second;
	for (HPT_Setup::ISection isect=ch.begin(); isect!=ch.end(); 
	     ++isect) ncells += (*isect).size();
      }
      cout << ncells << " cells)";
    }
  }
  cout << '\n';  

  Detector::iterator iter;
  if ((iter=HERA_B_detector->find("hpt_inner")) != HERA_B_detector->end())
    hpt_inner_setup = (HPT_Setup*)(__SetupBase*)((*iter).second());
  /*
  else {
    global_error()("setup \"hpt_inner\" is not defined");
    return;
  }
  */
  if ((iter=HERA_B_detector->find("hpt_outer")) != HERA_B_detector->end())
    hpt_outer_setup = (HPT_Setup*)(__SetupBase*)((*iter).second());
  /*
  else {
    global_error()("setup \"hpt_outer\" is not defined");
    return;
  }
  */
}
    
#include <arte/Arte.hh>
#include <arte/Pointer.hh>
#include <auto/PTGE.hh>
#include "ptset/setup_algorithms.hh"
#include "ptset/setup_tools.hh"
#include <string.h>
#include <strings.h>

extern void fill_ARTE_hpt_geometry_table_for_one_setup
(const vector<vector<HPT_Setup::IChamber> >& chambers,
 double chamber_width,
 const string& setup_name, // for comment
 const string& section_name,
 const string& material_name);

void fill_ARTE_hpt_geometry_table() {
  const double chamber_width_inner = 3.5;
  const double chamber_width_outer = 1.;

  vector<vector<HPT_Setup::IChamber> > chambers;
  // contains chambers ordered by z
  /*  Inner part  */
  sort_setup_chambers_by_z(*hpt_inner_setup, &chambers, 1.e-6);
  fill_ARTE_hpt_geometry_table_for_one_setup(chambers, chamber_width_inner,
					     "inner", "PTC", "PT:InnGas+Fe");
  /*  Outer part  */
  chambers.erase(chambers.begin(), chambers.end());
  sort_setup_chambers_by_z(*hpt_outer_setup, &chambers, 1.e-6);
  fill_ARTE_hpt_geometry_table_for_one_setup(chambers, chamber_width_outer,
					     "outer", "PTO", "PT:OutGasCMW");
}

void fill_ARTE_hpt_geometry_table_for_one_setup
(const vector<vector<HPT_Setup::IChamber> >& chambers,
 double chamber_width,
 const string& setup_name, // for comment
 const string& section_name,
 const string& material_name) {
  //  cout << "Start ARTE PTGE bank filling ...\n";
  for (unsigned int ilayer=0; ilayer<chambers.size(); ++ilayer) {
    unsigned int i_rect = 1;
    for (unsigned int inside_layer=0; inside_layer<chambers[ilayer].size();
	 ++inside_layer) {
      HPT_Chamber& ch = (*chambers[ilayer][inside_layer]).second;
      vector<ChamberRectangularPartDescription> rectangles;
      make_chamber_from_rectangles(ch, &rectangles);

      string superlayer_name = "PTL";
      superlayer_name << ilayer+1;
      
      for (unsigned int i=0; i<rectangles.size(); ++i, ++i_rect) {
	ChamberRectangularPartDescription& rp = rectangles[i];
	Rectangle r(rp.ind1 * rp.grid, rp.ind2 * rp.grid);
	
	PTGE x;

	x.flag = 1;
      
	HPT_Setup::ISection isect = ch.isection(r.center());
	unsigned int section_number = isect - ch.begin();

	string name = section_name;
	if (section_number<10) name << section_number;
	else if (section_number-10<=int('Z')-int('A'))
	  name << char(int('A')+section_number-10);
	else name << '*';

	memcpy(x.name, name.c_str(), 4);
	memcpy(x.supl, superlayer_name.c_str(), 4);
	memcpy(x.moth, x.supl, 4);
	memcpy(x.shap, "BOX ", 4);
	memcpy(x.mate, material_name.c_str(), 12);
	memcpy(x.maon, "ONLY", 4);
	x.irot = 0;
	bzero(x.rotm, sizeof(float)*9);
	x.colo = section_number+2;
	x.fill = 1;
	x.seen = 1;
	Grid& grid = (*isect).grid;
	string comment = "HPT_";
	comment << setup_name << "_#" << i_rect << ".Cell_"
		<< grid.x << "x" << grid.y;
	memcpy(x.comm, comment.c_str(), comment.size());
	memset(x.comm+comment.size(), ' ', 80-comment.size()); // fill the rest
	// with periods
      
	x.zmin = ch.z - chamber_width / 2.;
	x.zmax = ch.z + chamber_width / 2.;
      
	x.dx = grid.x;
	x.dy = grid.y;
      
	x.xmin = r.a.x;
	x.xmax = r.b.x;
	x.ymin = r.a.y;
	x.ymax = r.b.y;

	Index ind = rp.ind2 - rp.ind1;
	x.nx = ind.ix;
	x.ny = ind.iy;
	/*
	cout << "Flag: " << x.flag << ", "
	     << "name: " << string(x.name,4) << ", "
	     << "superlayer: " << string(x.supl,4) << ",\n"
	     << "mother: " << string(x.moth,4) << ", "
	     << "shape: " << string(x.shap,4) << ", "
	     << "material: " << string(x.mate,12) << ",\n"
	     << "many/only: " << string(x.maon,4) << ", "
	     << "rotation: " << x.irot << ", "
	     << " rotation[5]: " << x.rotm[5] << ",\n"
	     << "x: " << x.xmin << "-" << x.xmax << "; "
	     << "y: " << x.ymin << "-" << x.ymax << "; "
	     << "z: " << x.zmin << "-" << x.zmax << ";\n"
	     << "dx/dy: " << x.dx << " / " << x.dy << ", "
	     << "color: " << x.colo << ", "
	     << "fill: " << x.fill << ", "
	     << "seen: " << x.seen << ",\n"
	     << "comment: " << x.comm << endl;
	     */
	ArtePointer<PTGE>::append(x);
      }
    }
  }
}
