//--------------------------------------------------------------------
// Created: 17.03.99 by   Olya Igonkina
/* MARPLE function for matching RICH segments and ECAL clusters
   works both with magnet ON and OFF

   FORTRAN_ROUTINE(matchrichecal)( float * threshold_cut, 
                                   float * minCherenkov_angle )
   if threshold_cut > 200 - it is Ktrig cut, otherwise it is pt_cut  

 */
//--------------------------------------------------------------------

#include "marple/marple.hh"
#include "marple/RcSegClusBuilder.hh"
#include "arte/RSEG_cmp.hh"

#include "marple/RcMcFunction.hh"

//definition of constants for matching ( see \Ref{RcMatchConst})
// flag, z_plane, z_min, z_max, use_pair_cut, n_pair_left, n_pair_right, 
//use_delta_cut, delta_x, delta_y, delta_tx, delta_ty, delta_pe, chi2

// match with Ecluster  {tx, ty}  magnet is ON 
RcMatchConst rcRichEcalConst = { 
  110,     1330.,   0.,  1500.,      1,          1,            2,
  1,     1.e+10  ,     1.e+10,   0.05,  0.015,   10000.,   150
};

// match with Ecluster  {tx, ty}  magnet is OFF 
RcMatchConst rcRichEcalNoMagnetConst = { 
  110,     1330.,   0.,  1500.,      1,          2,            3,
  1,     1.e+10  ,     1.e+10,   0.015,  0.016,   10000.,   140
};



extern "C"  
void FORTRAN_ROUTINE(matchrichecal)( float * threshold_cut,
				     float * minCherenkov_angle) 
{
  RcSegClusBuilder RichEcalMatch;  // builder for charge track search


  //--------------------------------------------------------------------
  // put constants to the MatchBuilder
  //--------------------------------------------------------------------
  if( !maMagnetON() )
    RichEcalMatch.setConst(     rcRichEcalNoMagnetConst );
  else
    RichEcalMatch.setConst(     rcRichEcalConst );

  RichEcalMatch.setBremConst( rcRichEcalConst );

  //--------------------------------------------------------------------
  // create vector of clusters from CARE
  //--------------------------------------------------------------------

  int iErr  = 0;
  for( ArteTable<RCCL>::iterator irccl = ArteTable<RCCL>::begin();
       irccl != ArteTable<RCCL>::end(); ++irccl )
    {
      irccl->flag = -1;
      int takeRCCL = 0;
      if( *threshold_cut > 200 )
	{
	  // Ktrig cut
	  HepDouble radius = sqrt( (*irccl)->x*(*irccl)->x + 
				   (*irccl)->y*(*irccl)->y);
	  HepDouble radMag = sqrt( (*irccl)->x*(*irccl)->x + 
				   fabs((*irccl)->y*(*irccl)->y*(*irccl)->y) );
	  if( radMag && radius && 
	      (*irccl)->esum > *threshold_cut * ( 1/radius+1/radMag ) )
	    takeRCCL = 1;
	}else{
	  // pt_cut
	  HepDouble radius = sqrt( (*irccl)->x*(*irccl)->x + 
				   (*irccl)->y*(*irccl)->y);
	  HepDouble Et = (*irccl)->esum * radius/
	    sqrt( radius*radius + (*irccl)->z*(*irccl)->z );
	  if( Et > *threshold_cut ) takeRCCL = 1;
	}
      if( takeRCCL)
	{
	  RcClusPoint * ptPoint = new RcClusPoint( (*irccl));
	  RichEcalMatch.addPointRight( *ptPoint);
	  ptPoint = 0;
	}
    }  

  //--------------------------------------------------------------------
  // create vector of RICH segments 
  //--------------------------------------------------------------------
 
  for( ArteTable<RSEG>::iterator irseg = ArteTable<RSEG>::begin();
       irseg != ArteTable<RSEG>::end(); ++irseg )
    {
      if(  (*irseg)->cmp & Rsegc::rich && 
	   (*irseg)->time > *minCherenkov_angle)
	{ 	  
	  RcSegPoint * ptToPoint = new RcSegPoint( irseg , 0); // firstPoint
	  RichEcalMatch.addPointLeft( *ptToPoint);
	  ptToPoint = 0;
	}
    }
  //--------------------------------------------------------------------
  // do Matching 
  //--------------------------------------------------------------------

  RichEcalMatch.matchRings();
  //--------------------------------------------------------------------
  // outprint
  //--------------------------------------------------------------------

  //RichEcalMatch.fillMatch(1000);
  //RichEcalMatch.fillDeltas(2000,3000);

  /* 
    static int firstSVDCARE = 1;
    if( firstSVDCARE )
     {
    cout << "================= match ============="<< endl; 
    RichEcalMatch.stat(1,1);
    cout << endl << endl << endl;
    firstSVDCARE = 0;
     }else{
    
    cout << "================= match ============="<< endl; 
    RichEcalMatch.stat(2,1);
    }
    */
  //--------------------------------------------------------------------
  // save found relation to the table
  //--------------------------------------------------------------------

  RichEcalMatch.saveRelation(iErr);
  if( iErr ){ cout << "MARPLE: some relations were not saved " << endl;}

  //--------------------------------------------------------------------
  // clean everything
  //--------------------------------------------------------------------
  RichEcalMatch.deleteAll();  
 
}
