/* Some classes for Setups with cells which can be fired/non fired.

   (Note: not all Setups fall into this category. E.g. calorimeter
   which has cells with analog output /ADC/ can not be considered 
   in this way. 
   In fact it can be considered, but only if one fixes the threshold
   which distinguishes fired/non fired amplitudes. This threshold is
   an external entity).

   Includes:
     Fired_Cell contains n_hits, occupancy, 
                         enum { Ok=1, BAD, NOISY, UNKNOWN } state;
     Section_With_Fired_Cells<CellType> contains 
               vector<__SetupSection<CellType>::ICell> fired_cells;
	       double average_occupancy;
   Vector fired_cells can be used for fast traversing over all fired
   cells in the event. It is worthwhile to use it if the occupancies
   are small. It should be reset in the beginning of each event.
   (See below).

   Usage: derive YourCellType from Fired_Cell,
                 YourSectionType -
		 from Section_With_Fired_Cells<YourCellType>
		 (or make it the same, if no fields should be added in 
		 addition)
   Then make Setup (adding extra YourCamberType, YourSetupType):
   Setup<YourCellType, YourSectionType, YourCamberType, YourSetupType>. 

   See high_pt.hh as an example.
   When your Setup is defined in this way you may use tools from
   fired_cell_tools.hh.
......................................................................
   Note also that to use write_data_to_fired_setup_cell(...)
   from fired_cell_tools.hh the programmer should define
   is_fired() method and "data" field of type CellType::DataType in 
   the his Cell class. See fired_cell_tools.hh for details.
......................................................................
   */
   
     
// Author V.Balagura (balagura@mail.desy.de)   (09.04.98).

#ifndef _fired_cells_hh
#define _fired_cells_hh 1

#include "ptset/setup.hh"

/* -------------------- Fired_Cell -------------------- */
class Fired_Cell {
public:
  int n_hits;
  double occupancy;
  enum State { Ok=1, BAD, NOISY, UNKNOWN };
  State state;

  Fired_Cell() : n_hits(0), occupancy(-1.), state(UNKNOWN) { }
};

/* -------------------- Section_With_Fired_Cells -------------------- */
template<class CellType>
class Section_With_Fired_Cells {
public:
public:
  vector<__SetupSection<CellType>::ICell> fired_cells;
  double average_occupancy;

  Section_With_Fired_Cells() : average_occupancy(-1.) { } 
};

#endif
