// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 
#include <iostream.h>
#include <math.h>
#include "arte/arte.hh"
#include "arte/ArteMessage.hh"
#include "rang/Node.hh"
#include "rang/HitWire.hh"
#include "rang/HitStrip.hh"
#include "rang/HitPad.hh"
#include "rang/HitUnsigned.hh"
#include "rang/ECClus.hh"
#include "rang/ThinPlane.hh"
#include "rang/ThinWall.hh"
#include "rang/ThickWall.hh"
#include "rang/State.hh"
#include "rang/StateP.hh"
#include "rang/FitMode.hh"

//-- Constructor
rangerNode::rangerNode() {
  z = 0;
  qPExpand = 0;
  state = rangerState::makeState( ModeP );
  predictedState = 0;
  transportMatrix = 0;
  cout << "Warning: rangerNode standard constructor done" << endl;
};

rangerNode::rangerNode(rangerFitMode fm) {
  z = 0;
  qPExpand = 0;
  state = rangerState::makeState( fm );
  predictedState = 0;
  transportMatrix = 0;
};

rangerNode::rangerNode(ArtePointer<RHIT> rhit,rangerFitMode fm) {
  qPExpand = rhit->p;
  state = rangerState::makeState(rhit,fm);
  predictedState = 0;
  transportMatrix = 0;
#ifdef FULL
  cout << "rangerNode rhit constructor done " << endl;
#endif
};

rangerNode::rangerNode(ArtePointer<HITB> hb,rangerFitMode fm) {
  qPExpand = 0.;
  state = rangerState::makeState(hb,fm);
  predictedState = 0;
  transportMatrix = 0;
#ifdef FULL
  cout << "rangerNode hitb constructor done " << endl;
#endif
};




rangerNode::rangerNode(ArtePointer<RCCL> rccl,int charge,rangerFitMode fm) {
  if (rccl->esum > 0) {
    qPExpand = charge /rccl->esum;
  }
  else {
    qPExpand = charge * 9999;
  }
  state = rangerState::makeState(rccl,charge,fm);
  predictedState = 0;
  transportMatrix = 0;
#ifdef FULL
  cout << "rangerNode rccl constructor done " << endl;
#endif
};


rangerNode::~rangerNode() {
//  cout << "rangerNode standard destructor deletes predictions" << endl;
  delete state;
  delete predictedState;
  delete transportMatrix;
};


ostream& operator<<(ostream &os,const rangerNode &n){
  os << " z=" << n.z << " qPE=" << n.qPExpand;
  if (n.transportMatrix) os << " [F]";
  else                   os << " [ ]";
  //os << (n.state)->vec();
  os << " dim=" << n.dim()
     << " x=" << (n.state)->x()
     << " y=" << (n.state)->y()
     << " tx=" << (n.state)->tx()
     << " ty=" << (n.state)->ty()
     << " qP=" << (n.state)->qP();
  return os;
}


rangerNode* rangerNode::makeNode( ArtePointer<RHIT>& rhit,
				  rangerFitMode fm) {
  ArtePointer<GEDE> pl = rhit->hitb->gede;
//    cout << rhit->hitb->gede->cmp << endl;
  rangerNode* w=0;
  switch (pl->cmp) {
  case 1:
    w = new rangerHitStrip( rhit, fm );
    break;
  case 2:
    w = new rangerHitWire( rhit, fm );
    //w = new rangerHitUnsigned( rhit, fm );
    // cout << "                      plane=" << w->plane();
    break;
  case 3:
    w = new rangerHitStrip( rhit, fm );
    break;
  case 4:
    w = new rangerHitPad( rhit, fm );
    break;
  default:
    cout << "Construct no hit" << endl;
    w = 0;
  }

  return w;
}

rangerNode* rangerNode::makeNode( ArtePointer<RCCL>& rccl,
				  int charge,rangerFitMode fm) {
  return (new rangerECClus( rccl, charge, fm, 20 ));
}



rangerNode* rangerNode::makeNode( const ArtePointer<GEDE>& ge,
				  rangerNode* n,const float path) {
  // here we should decide if thick or thin plane
  rangerNode* w = new rangerThinPlane(ge,n,path);
  return w;
}

rangerNode* rangerNode::makeNode( const ArtePointer<GWAL>& gw,
				  rangerNode* n,const float path) {
  rangerNode* w;
  if (fabs(gw->zmax - gw->zmin) < 20) {
    w = new rangerThinWall(gw,n,path);   // create a thin wall
  }
  else {
    w = new rangerThickWall(gw,n,path);  // create a thick wall
  }
  return w;
}



int rangerNode::filter() { 

#ifdef FULL
  cout << "rangerNode::filter(): this is the node at z=" << z << endl;
#endif
  if (! predictedState) {
    predictedState = rangerState::makeState( state->fitMode() );
    //predictedState = new rangerState();
  }
  else {
  }

  *(predictedState) = *state;     // copy initialization into prediction
  
  predictedState->covInit();      // initialize cov matrix
                                  // no transport matrix needed
  *state = *predictedState;       // just copy the prediction

  return 0;
}



int rangerNode::filter(rangerNode* n,rangerParticleType pType) {

// default filter in case of no measurement

#ifdef FULL
  cout << "rangerNode::filter(...): this is the node at z=" << z << endl;
  cout << "                         call prediction" << endl;
#endif

  int returnStatus = n->predict(z,predictedState,transportMatrix);

#ifdef FULL
  cout << "  filter(...)               predictedState->fitMode()=" 
       << predictedState->fitMode() << endl;
  cout << "rangerNode::filter(...): the predicted state is" 
       << predictedState->vec() << endl
       << predictedState->cov() << endl
       << "                         the transport matrix is "
       << *transportMatrix << endl;
#endif
  
  if (returnStatus >= 0) {
    if (plane() != n->plane())    predProcessNoise(pType);
    *state = *predictedState;  // just copy the prediction
#ifdef FULL
    cout << "rangerNode::filter(...): state->fitMode()=" 
       << state->fitMode() << endl;
    cout << "end default filtering " << endl;
#endif
    return 0;
  }
  else {
    return returnStatus;
  }
#ifdef FULL
   cout << "end default filtering " << endl << endl << endl << endl;
#endif
}
  




int rangerNode::predict(const double& zDestination,rangerState*& tState,
			HepMatrix*& FTmat) {
  
#ifdef FULL
  cout << "rangerNode::predict: from state->z=" << state->z()
       << " to zDestination=" << zDestination
       << " fitMode=" << state->fitMode() << endl;
#endif

//  return state.predict(zDestination,targetState,FTransport);
  int retflag = state->predict(zDestination,qPExpand,tState,FTmat);
  return retflag;
}
  



int rangerNode::smooth(rangerNode* n) {
  // see the formulae in Hera-B 95-239, p4
  // this node  corresponds to (k)
  // the node n corresponds to (k+1)
  int stat;
  HepMatrix A( state->cov() * n->getF()->T() 
		  * n->getPrediction()->cov().inverse(stat));
  if (stat < 0) {
    cout << "rangerNode::smooth(): inversion failed " 
	 << " invert returns stat=" << stat 
	 << " for plane " << plane()
	 << " at z=" << z << endl;
    return -1;
  }
  state->vec() += A * (n->state->vec() - n->getPrediction()->vec());
  state->cov() += (n->state->cov() - n->getPrediction()->cov()).similarity(A);

  if ( fabs(state->x()) < 700 &&   // check validity of rkmain input
       fabs(state->y()) < 700 &&
       fabs(state->tx()) < 10   &&
       fabs(state->ty()) < 10   &&
       fabs(state->qP()) < 100 ) ;
  else {
    cout << "Node::smooth: invalid state="
	 << state->vec() 
	 << " at z=" << state->z() << endl;
    if (predictedState) {
      cout << "The prediction was state=" << predictedState->vec()
	   << " cov=" << predictedState->cov() << endl;
    }
    if (transportMatrix) {
      cout << "The transport matrix is " << *transportMatrix << endl;
    }
    static int id;
    arteMsg << "Node::smooth: invalid state";
    arteMsg.allp(id,1000);
    //return -13;
  }



  residUpdate(n,A); // update residual and chiSq

  if (fabs(state->qP()) < 10);
  else {
    static int id2;
    arteMsg << "Node::smooth: Funny qP()=" << state->qP()
	    << " at z=" << state->z() << endl;
    arteMsg.allp(id2,1);
#ifdef FUNNYQP
    cout << "The state was " << state->vec()
	 << " with cov=" << state->cov() << endl;
    if (predictedState) {
      cout << "The prediction was state=" << predictedState->vec()
	   << " cov=" << predictedState->cov() << endl;
    }
#endif
    return -14;
  }

  qPExpand = state->qP(); // update expansion point
#ifdef QPEXPWARN
  static int id3;
#endif
  if (qPExpand > 2) {
#ifdef QPEXPWARN
    arteMsg << "Node::smooth(): qPExpand fixed from " << qPExpand
	    << " at z=" << state->z() << endl;
    arteMsg.allp(id3,1000);
#endif
    qPExpand = 2;
  }
  else {
    if (qPExpand < -2) {
#ifdef QPEXPWARN
      arteMsg << "Node::smooth(): qPExpand fixed from " << qPExpand
	    << " at z=" << state->z() << endl;
      arteMsg.allp(id3,1000);
#endif
      qPExpand = -2;
    }
  }


#ifdef FULL
  cout << "Node::smooth(): Expansion point updated to " << qPExpand << endl;
#endif
  return 0;
}

void rangerNode::list() {
  cout << " Node " << this << " plane=" << plane()
       << " z=" << z
       << " chiSq=" << getChiSq()
       << " hasHitb=" << hasHitb()
       << " ms=" << ms() << endl 
       << "                 " << " radlength=" 
       << radiationThickness()
       << " state dims " << getState().dim()
       << " x=" << getState().x() << " tx=" << getState().tx();
  if (getPrediction()) {
    cout << " " << getPrediction()->dim();
  }
  else {
    cout << " (no pred) ";
  } 
  cout << endl;
}

void rangerNode::dump(){
  cout << "----------------------------- Node z=" << z 
       << " qPExpand=" << qPExpand << " plane=" << plane() << endl;
  if (state) {
    cout << "state=" << endl;
    state->dump();
  }
  if (predictedState) {
    cout << "predictedState=" << endl;
    predictedState->dump();
  }
  if (transportMatrix) {
    cout << "transportMatrix=" << endl;
    cout << *transportMatrix << endl;
  }
}  


float rangerNode::radiationThickness() {
  float teff=msPlain()/(0.0136*0.0136);
  float t;
  if (teff>4e-12) {        // undo the logarithmic correction
    t = teff / ((1+0.038 * log(teff)) * (1+0.038 * log(teff))); // 1st It
    t = teff / ((1+0.038 * log(teff)) * (1+0.038 * log(t)));    // 2nd It
  }
  else {
    t = teff;
  }
  return t;
}


float rangerNode::radiationThickEff() {
  float teff=ms()/(0.0136*0.0136);
  float t;
  if (teff>4e-12) {        // undo the logarithmic correction
    t = teff / ((1+0.038 * log(teff)) * (1+0.038 * log(teff))); // 1st It
  }
  else {
    t = teff;
  }
  return t;
}




void rangerNode::covInit() { 
  state->covInit();
}
  

