// ----------------------------------------------------------------------
//
// Fills the GTAR bank from the DTAR bank.
//
// The DTAR bank must have to be filled with a TGTS_RAW structure 
// on the SLT/TLT farm.
//
//
// TGTS_RAW structure (-> include/tgts_raw.h in the ONLINE library):
// ------------------
//
//    #define TGTS_Nwires 8
//    struct TGTS_RAW {
//      int time; /* somehow mizzi is using int for timestamps, not long */
//      float ia;
//      float bg;
//      float pos[TGTS_Nwires];
//    };
//    typedef struct TGTS_RAW TGTS_RAW;
//
//
// DTAR format:   1.     row:   recid
// ------------   2./ 3. row:   SLB_SUBHDR
//                4.-14. row:   TGTS_RAW
//
//
// Its assumed that the order in the DTAR bank is the same as in the
// target database:
//
// 1999 data:  1. GTAR row  Ia    Al           z=-1.0 cm 
// ----------  2.      row  Ib    Ti           z=-1.0 cm 
//             3.      row  Ii    Al           z=-0.7 cm 
//             4.      row  Io    Al           z=-0.7 cm 
//             5.      row IIa    Fe  (45 deg) z=-5.0 cm 
//             6.      row IIb    Al  (45 deg) z=-5.0 cm 
//             7.      row IIi    C            z=-4.7 cm 
//             8.      row IIo    Ti           z=-4.7 cm 
//
// Dimensions for all targets:
//          Length   : 3.0   cm
//          Width    : 0.005 cm =  50 um
//          Thickness: 0.05  cm = 500 um
//
//
// Author: Rainer Wanke, 22-Feb-1999
//
// ----------------------------------------------------------------------

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>

// Arte banks,  C++ version
#include "arte/arte.hh"
//extern "C" void mmdrop_(const int &);

// target structure in DTAR bank
#include "tgts_raw.h"

// event structures
#include "event.h"

// byte swapping
#include "esp.h"

// return codes
#include "grr.h"


int rc99AfillGTARfromDTAR()
{
  // Do nothing for MC events
  if (ArteTable<MTRA>::begin() != ArteTable<MTRA>::end())
    return GRR_OK;
  
  // print out DTAR record id
  ArtePointer<DTAR> dtar(1);
  //  int* DTARrecid = (int*) dtar->bits;
  //  printf( "DTAR recid = 0x%x\n", *DTARrecid );

  // check correct size of DTAR bank
  int dtar_size = ArteTable<DTAR>::size() * sizeof(DTAR);
  if ( dtar_size != sizeof(int) + sizeof(SLB_SUBHDR) + sizeof(TGTS_RAW) ) {
    cout << "rc99AfillGTARfromDTAR: ERROR: DTAR size = " << dtar_size
	 << ", should be " << sizeof(int) + sizeof(TGTS_RAW) << endl;
    cout << "                  No GTAR update done !" << endl;
    return GRR_FISHY;
  }

  // set up target structure from DTAR bank
  dtar += sizeof(SLB_SUBHDR)/sizeof(int) + 1;
  TGTS_RAW *target = (TGTS_RAW*) dtar->bits;
   
  // print out target entries
  //  cout << "Target Structure: Time = " << target->time
  //       << ", IA rate = " << target->ia
  //       << ", Bkg rate = " << target->bg << endl;
  //  for (int i=0; i<TGTS_Nwires; i++)
  //    cout << "                  Target " << i
  //	 << ": " << target->pos[i] << endl;

  // Drop GTAR bank
  //mmdrop_( ArtePointer<GTAR>::getLink().getInt() );
  ArteTable<GTAR>::clear();
  // ----------------------
  // Now fill the GTAR bank
  // ----------------------
  float rote[9] = { 1.0, 0.0, 0.0,    // unit matrix
		    0.0, 1.0, 0.0,
		    0.0, 0.0, 1.0};
  float rot[9];       // rotation matrix for 45 degree wires 
  rot[0] =  sqrt(0.5);
  rot[1] =  sqrt(0.5);
  rot[2] =  0.0;
  rot[3] = -sqrt(0.5);
  rot[4] =  sqrt(0.5);
  rot[5] =  0.0;
  rot[6] =  0.0;
  rot[7] =  0.0;
  rot[8] =  1.0;

  GTAR gtar;
  gtar.cmp   = 0;
  gtar.x0bm  = 0.;
  gtar.y0bm  = 0.;
  gtar.sxbm  = 0.;
  gtar.sybm  = 0.;
  gtar.shap  = 0;
  gtar.ms    = 0.;

  // loop over wires
  for (int i=0; i<TGTS_Nwires; i++) {

    // convert wire position to cm and correct sign if necessary
    float WirePosition[TGTS_Nwires];
    ESP_get_int( (int*) target->pos, (int*) WirePosition, TGTS_Nwires );
    if ( i==1 || i==5 || i==3 || i==7 )
      WirePosition[i] =  0.1 * WirePosition[i];
    else
      WirePosition[i] = -0.1 * WirePosition[i];

    // wire number and thickness
    gtar.wnr = i+1;
    gtar.dz  = 0.05;

    // I above/below
    if ( i==0 || i== 1 ) {
      gtar.z   = -1.0;
      gtar.dx  = 3.0;
      gtar.dy  = 0.005;
      gtar.x   = 0.;
      gtar.y   = WirePosition[i];
      gtar.rot = rote;
    }

    // I inner/outer
    if ( i==2 || i== 3 ) {
      gtar.z   = -0.7;
      gtar.dx  = 0.005;
      gtar.dy  = 3.0;
      gtar.x   = WirePosition[i];
      gtar.y   = 0.;
      gtar.rot = rote;
    }

    // II above/below  (rotated by 45 degrees!)
    if ( i==4 || i== 5 ) {
      gtar.z   = -5.0;
      gtar.dx  = 3.0;
      gtar.dy  = 0.005;
      gtar.x   = 0.;
      gtar.y   = WirePosition[i];
      gtar.rot = rot;
    }

    // II inner/outer
    if ( i==6 || i== 7 ) {
      gtar.z   = -4.7;
      gtar.dx  = 0.005;
      gtar.dy  = 3.0;
      gtar.x   = WirePosition[i];
      gtar.y   = 0.;
      gtar.rot = rote;
    }

    // Store new GTAR table
    ArteTable<GTAR>::push_back( gtar );

  }

  return GRR_OK;
}




