#ifndef __PMTYPE_H__
#define __PMTYPE_H__

// ------------------------------------------------------------------------
/// PMT--type specific information
/**
  The class PMType is used for the description of the common 
  properties of all M16 respectivily M4 pmt's. The geometrical
  properties of the channel grid and the common wavelength
  dependency (assumption!) of all M16 or M4 pmt's can be found
  here. 
  The channel numbers are defined in the following way. If you are
  looking from the back of one M16-pmt:\\[4mm]
  \begin{tabular}{lll}
  lowerleft  corner:&index=0 &(row=0,column=0)\\
  lowerright corner:&index=3 &(row=0,column=3)\\
  upperleft  corner:&index=12&(row=3,colunm=0)\\
  upperleft  corner:&index=15&(row=3,colunm=3)
  \end{tabular}\\[4mm]
  Similar in the M4 case. 
 */
class PMType {
protected:
  char *ttype;
  ArtePointer<CPMT> a;
  int  numchan,nnx,nny;
  double xxmin,yymin,xxdiv,yydiv;
  Table eff;

  inline PMType() { ttype=strdup("undefined");};
  inline void name(char *n) {
    if (ttype) free(ttype);
    ttype=strdup(n);
    }
  
public:
/// return type-name 
  inline const char*  name() const { return ttype; }

/// returns the number of channels
  inline int    numchannel() const { return numchan; }

/// returns the number of columns (number of channels in x-direction) 
  inline int    numcol() const { return nnx; }

/// returns the number of rows    (number of channels in y-direction) 
  inline int    numrow() const { return nny; }

/// returns the pad size of a column (padsize in x-direction) 
  inline double    colsize() const { return xxdiv; }

/// returns the pad size of a row (padsize in y-direction) 
  inline double  rowsize() const { return yydiv; }


/// returns pmt-channelindex: mapping of row and column to linear index
  inline int index(int i,int j) const { 
    int k=i*nnx+j;
    return (k>=numchannel()) ? k=-1 : k;
    }
    
/// evaluate the row from the pmt-channelindex
  inline int row(int index) const { 
    return index/nnx;
    }

/// evaluate the column from the pmt-channelindex
  inline int col(int index) const { 
    return index%nnx;
    }
   
/// returns pmt-channelindex: depending on position where photon hits the pmt
  inline int index(double x,double y) const { 
    return index((int)((y-yymin)/yydiv),(int)((x-xxmin)/xxdiv));
    }
    
/// return x-position as function of the pmt-channelindex
  inline double x(int channel) const {
    return (channel%nnx+0.5)*xxdiv+xxmin;
    }

/// return y-position as function of the pmt-channelindex
  inline double y(int channel) const {
    return (channel/nnx+0.5)*yydiv+yymin;
    }

/// return quantum-efficiency as function of the wavelength
  inline float  efficiency(double l)  { return eff.f(l); }
  inline Table& Efficiency()          { return eff; }

///  print the contents of the class to an ASCII-File 
  void print(FILE *fp=stdout);

///  constructor reading information from an ASCII-File produced using 'print'
  PMType(ScriptFile& f);

/// constructor using a record in the CPMT-ARTE-Table
  PMType(ArtePointer<CPMT> a);
  
///  destructor
  ~PMType();
  };
   

#endif
