// -*-C++-*-
#ifndef __RICHDAQ_H__
#define __RICHDAQ_H__

extern "C" {
#include "event.h"
#include "etb.h"
#include "emg.h"
#include "component.h"
#include "grr.h"
#include "libr.h"
#include "log.h"
}

#include "rich/ricutil.h"
#include "rich/PhotonDetection.h"
#include "rich/PhiLambda.h"


//#define INDEXCHECK


// ---------------------------------------------------------------------------
/// wire map of the M16-base
class M16Base {
int pm[4*16] ,pmChannel[4*16];
int asdID[4*16],asdChannel[4*16];

int numchan;

inline int index(int asdid,int asdchannel) const {
  int ret=asdid*16+asdchannel;
#ifdef INDEXCHECK
  if (ret<0||ret>=numchan) {
    fprintf(stderr,"M16base::index out of range ! %d %d %d %d\n",
           asdid,asdchannel,ret,numchan);
    Exit(1);
    }
#endif
  return ret;
  }

public:
/// contructor reading the wiring from as ASCII-File
  M16Base(ScriptFile& f);
/// returns the pmt (0...3) as function of the asd-ID(0...3) and the ASD-Channel
  inline int  pmt(int asdid,int asdchannel) const {
    return pm[index(asdid,asdchannel)];
    }
/// returns the pmt-channel (0...15) as function of the asd-ID(0...3) and the ASD-Channel
  inline int  pmtChannel(int asdid,int asdchannel) const {
    return pmChannel[index(asdid,asdchannel)];
    }
/// returns the asdid (0...3) as function of the pm index(0..3) and the pm channel (0.15)
  inline int  asdid(int pmi,int pmc) const { return asdID[pmi*16+pmc]; }
/// returns the asd-channel (0...16) as function of the pm index(0..3) and the pm channel (0.15)
  inline int  asdchannel(int pmi,int pmc) const {
    return asdChannel[pmi*16+pmc];
    }
  };

// --------------------------------------------------------------------
/// wire map of the M4-base

class M4Base {
int pm[4*4],pmChannel[4*4],asdChannel[4*4];
int numchan;

inline int index(int asdchannel) const {
#ifdef INDEXCHECK
  if (asdchannel<0||asdchannel>=numchan) {
    fprintf(stderr,"M4base::index out of range ! %d %d\n",
           asdchannel,numchan);
    Exit(1);
    }
#endif
  return asdchannel;
  }

public:
  M4Base(ScriptFile& f);
/// return the pmt (0...3) as function of the ASD-Channel(0..15)
  inline int  pmt(int asdchannel) const {
    return pm[index(asdchannel)];
    }
/// returns the pmt-channel (0...3) as function of the ASD-Channel(0..15)
  inline int  pmtChannel(int asdchannel) const {
    return pmChannel[index(asdchannel)];
    }
/// returns the asdid
  inline int  asdid() const { return 0; }
/// returns the asd-channel (0...16) as function of the pm index(0..3) and the pm channel (0.15)
  inline int  asdchannel(int pmi,int pmc) const {
    return asdChannel[pmi*4+pmc];
    }
  };



// --------------------------------------------------------------------

class RICHFEDInfo {
public:
/// return Number of Daughters per FED-system
  inline int NumDaughterPerFED() const { return 4; }
/// return Number of Connectors per FED-system
  inline int NumConnectorPerFED()  const { return 64; }
/// return Number of Connectors per FED-system
  inline int NumChannelPerFED()  const { return 1024; }

/// return Number of Connectors per Daughter
  inline int NumConnectorPerDaughter()  const { return 16; }
/// return Number of Channels per Daughter
  inline int NumChannelPerDaughter()  const { return 16*16; }

/// return Number of Channels per Connector
  inline int NumChannelPerConnector()  const { return 16; }
  inline int NumChannelPerASD()  const { return 16; }
  };


// --------------------------------------------------------------------

class RICHChannel : public RICHFEDInfo {
short b;
char f,s,r,c,pcha,pm_ncha,stat;

public:
  inline char&  fedid()     { return f; }
  inline short& bit()       { return b; }
  inline char&  status()    { return stat; }
  inline char&  sm()        { return s; }
  inline char&  row()       { return r; }
  inline char&  col()       { return c; }
  inline char&  pm_channel()    { return pcha; }
  inline char&  pm_numchannel() { return pm_ncha; }

/// return fed-id 
  inline int fedid() const {  return f; }
/// return bit-number 
  inline short bit() const {  return b; }
  
/// return daughter-number 
  inline int daughter() const {  return b/NumChannelPerDaughter(); }
/// return connector number (0..15) 
  inline int connector() const {  
    return (b%NumChannelPerDaughter())/NumChannelPerConnector(); 
    }
/// return asd channel number (0..15) 
  inline int channel() const {  
    int i=b%NumChannelPerConnector();
    return (i&0x8)|(7-(i&0x7)); 
    }

/// return status
  inline int status() const { return stat; }
/// return status
  inline int bad()  const { return (stat!=1); }
/// return status
  inline int off()  const { return (stat==0); }
/// return status
  inline int ok()   const { return (stat==1); }
/// return status
  inline int hot()  const { return (stat==2); }
/// return status
  inline int cold() const { return (stat==3); }


/// return supermodule index
  inline int sm() const { return s;  }
/// return supermodule row (0..31)
  inline int row() const { return r; }
/// return supermodule column (0..11)
  inline int col() const { return c; }
    
/// return pmt number of channel (m16:0..15   m4:0..3)
  inline int pm_numchannel() const { return pm_ncha; }
/// return if pmt has 16 channel
  inline int m16() const { return (pm_ncha==16); };
/// return if pmt has 4 channel
  inline int m4()  const { return (pm_ncha==4); };
/// return pm channel index (m16: 0..15   m4: 0..3)
  inline int pm_channel() const { return pcha; }  
/// return  number of channel rows on pm (m16: 4,   m4:2)
  inline int pm_numrow() const { return (m16()) ? 4 : 2; }  
/// return  number of channel columns on pm (m16: 4,   m4:2)
  inline int pm_numcol() const { return (m16()) ? 4 : 2; }  
/// return row of channel on pm (m16: 0..3,   m4:0..1)
  inline int pm_row() const { return pcha/pm_numcol(); }  
/// return column of channel on pm (m16: 0..3,   m4:0..1)
  inline int pm_col() const { return pcha%pm_numcol(); }  

/// return baseboard row (0 .. max(row())/2-1)
  inline int bb_row() const { return r/2; }
/// return baseboard column  (0 .. max(col())/2-1)
  inline int bb_col() const { return c/2; }
  
/// return pmt row (0..1) on baseboard 
  inline int bb_pmrow() const { return r%2; }
/// return pmt column (0..1) on baseboard 
  inline int bb_pmcol() const { return c%2; }
/// return pmt index (0..3)
  inline int bb_pm() const { return bb_pmrow()*2+bb_pmcol(); }
/// print class
  inline void print(FILE* fp=stdout) {
    fprintf(fp,
      "f:%02d d:%02d con:%02d ch:%02d bit:%04d s:%02d r:%02d c:%02d pch:%02d "
      "st:%02d\n",
      fedid(),daughter(),connector(),channel(),bit(),
      sm(),row(),col(),pm_channel(),status());
    }
  };

// --------------------------------------------------------------------
///  
/// Look up table to translate bit info in FED-data to detector position 
///  

class  RICHDAQ  : public RICHFEDInfo {

protected:
M16Base* m16bas;
M4Base*  m4bas;

struct {
  unsigned char bit[8];
  } pattern[256];
  
unsigned short *fedbitindex;

int numfed;
int numchannel;

ETB_fed*   etb_fed;
EVENT_HDR* event;
int        numfedevt;



int  sel_channelindex;
RICHChannel  *allcha,*selcha;
unsigned long* numhit;
unsigned short *hitlist;
int  nhit;
unsigned short bx;
unsigned short fltevt;
int numevent;


public:
/// constructor
  RICHDAQ(char *configdir=0);
/// event output to a stream
  void printevent(FILE *fp=stdout);
/// return the total number of FED's of the RICH
  inline int NumFED() const { return numfed; }
/// return the number of channels in the RICH
  inline int NumChannel() const { return numchannel; }


/// get fed data from EVENT_HDR* and execute hit search. return 0 at success
  int NextEvent(EVENT_HDR* event);
/// read next event from file and call NextEvent(EVENT_HDR*). return 0 at success
  int NextEvent(FILE* fp);
/// MC simulation of RICH DAQ, read hits from DIGB. Returns 0 upon success.
  int NextEvent();
/// Reads next event from common run file and calls NextEvent(EVENT_HDR*). Returns 0 upon success.
  int NextEventL4(FILE *fp);

/// read data from DRIC table and fill HITB table.
  int fillHITBfromDRIC();


/// return the number of FED's of the RICH in the current event
  inline int NumFEDInEvent() const   { return numfedevt; }
/// return the  event number of the current event
  inline int EventNo()       const   { return fltevt; }
/// return the bx number of the current event
  inline int BX()            const   { return bx; }
/// return the number of channels with a hit in the current event
  inline int NumHitChannel() const { return nhit; }

/// set accumulated hit counters to zero
  inline void ZeroNumHit() {  numevent=0;memset(numhit,0,NumChannel()*sizeof(*numhit)); } 
/// return the accumulated number of hits of the selected channel
  inline int NumHit() const { return numhit[ChannelIndex()]; }
/// return the number of accumulated events
  inline int NumEvents()      const   { return numevent; }


/// return the position in the bit stream 
  inline int bit(int d,int con,int cha) const {
    return d*NumChannelPerDaughter()+con*NumChannelPerConnector()+cha;
    }
/// select the i'th channel for further use 
  inline int ChannelExist(int i) { 
    int ret=(i>=0&&i<NumChannel());
    if (ret) {
      selcha=&allcha[i];
      if (off()) ret=0;
      else sel_channelindex=fedbitindex[fedid()*NumChannelPerFED()+bit()];
      }
    return ret;
    }  
/// select the i'th channel with a hit for further use (i.e. for (i=0;daq.HitChannelExist(i);i++) ...)
  inline int HitChannelExist(int i) { 
    int ret=(i<NumHitChannel());
    if (ret) {
      selcha=&allcha[hitlist[i]];
      sel_channelindex=fedbitindex[fedid()*NumChannelPerFED()+bit()];
      }
    return ret;
    }
    
/// return select channel as function of FED-id (0..27) and bit number
  inline int ChannelExist(int fedid,int bit) { 
    int ret;
#ifdef INDEXCHECK
    if (fedid<0||fedid>=numfed||bit<0||bit>=NumChannelPerFED()) ret=0;
    else 
#endif
      {
      sel_channelindex=fedbitindex[fedid*NumChannelPerFED()+bit];
      if (sel_channelindex==(unsigned short)0xFFFF) ret=0;
      else {
        selcha=&allcha[sel_channelindex]; 
	ret=1;
	}
      }
    return ret;
    }
    
/// return the channel-index of the selected channel 
  inline int          ChannelIndex() const { return sel_channelindex; }
/// return the channel info of the selected channel 
  inline  RICHChannel& Channel() { return *selcha; }
/// return fed-id of the selected channel 
  inline int fedid() const {  return selcha->fedid(); }
/// return bit-number  of the selected channel
  inline short bit() const {  return selcha->bit(); }
  
/// return daughter-number  of the selected channel
  inline int daughter()  const {  return selcha->daughter(); }
/// return connector number (0..15)  of the selected channel
  inline int connector() const {  return selcha->connector(); }
/// return asd channel number (0..15) 
  inline int channel()   const {  return selcha->channel(); }

/// return status  of the selected channel
  inline int status() const { return selcha->status(); }
/// return status is bad (of the selected channel)
  inline int bad()  const { return selcha->bad(); }
/// return status is bad (of the selected channel)
  inline int off()  const { return selcha->off(); }
/// return status is ok (of the selected channel) 
  inline int ok()   const { return selcha->ok(); }
/// return status is hot (of the selected channel)
  inline int hot()  const { return selcha->hot(); }
/// return status is cold (of the selected channel )
  inline int cold() const { return selcha->cold(); }


/// return supermodule index (of the selected channel )
  inline int sm()  const { return selcha->sm();  }
/// return supermodule row (0..31) (of the selected channel )
  inline int row() const { return selcha->row(); }
/// return supermodule column (0..11) (of the selected channel )
  inline int col() const { return selcha->col(); }
    
/// return pmt number of channels (m16:16   m4:4)  (of the selected channel )
  inline int pm_numchannel() const { return selcha->pm_numchannel(); }
/// return pmt number of rows (m16:4   m4:2)  (of the selected channel )
  inline int pm_numrow() const { return selcha->pm_numrow(); }
/// return pmt number of columns (m16:4   m4:2)  (of the selected channel )
  inline int pm_numcol() const { return selcha->pm_numcol(); }
/// return pm channel index (m16: 0..15   m4: 0..3) (of the selected channel )
  inline int pm_channel() const { return selcha->pm_channel(); }  
/// return row of channel on pm (m16: 0..3,   m4:0..1)
  inline int pm_row() const { return selcha->pm_row(); }  
/// return column of channel on pm (m16: 0..3,   m4:0..1)
  inline int pm_col() const { return selcha->pm_col(); }  
/// return if pmt has 16 channel (of the selected channel )
  inline int m16() const { return selcha->m16(); };
/// return if pmt has 4 channel (of the selected channel )
  inline int m4()  const { return selcha->m4(); };

/// return baseboard row (0 .. max(r())/2-1) (of the selected channel )
  inline int bb_row() const { return selcha->bb_row(); }
/// return baseboard column  (0 .. max(c())/2-1) (of the selected channel )
  inline int bb_col() const { return selcha->bb_col(); }
  
/// return pmt row (0..1) on baseboard (of the selected channel )
  inline int bb_pmrow() const { return selcha->bb_pmrow(); }
/// return pmt column (0..1) on baseboard (of the selected channel )
  inline int bb_pmcol() const { return selcha->bb_pmcol(); }
/// return pmt index (0..3) (of the selected channel )
  inline int bb_pm() const { return selcha->bb_pm(); }

/// return asdid (of the selected channel )
  inline int asdid() const { 
    return (m16())
         ? m16bas->asdid(bb_pm(),pm_channel())
	 : m4bas->asdid();
    }


/// return the daq-event-structure of the current event
  inline const EVENT_HDR* daq_event() const { return event; }
/// return the daq-fed-structure of the current event
  inline const ETB_fed* daq_etb_fed() const { return etb_fed; }

/// return the wiremap of the m16-base
  inline const M16Base& m16base() const { return *m16bas; }
/// return the wiremap of the m4-base
  inline const M4Base&  m4base() const { return *m4bas; }

  static ArtePointer<GEDE> RichGede;

  };

extern RICHDAQ *RICH_DAQ;
 
#endif 
