#include <stdio.h>
#include <cmath>
#include <cstdio>
#include <vector>

#include <arte/arte.hh>
#include "silarte/silarte.hh"
#include "silarte/EVinfo.hh"

#include "grr.h"
#include "aframelib.h"
#include "sildat.h"
#include "holmes.h"

#include "Class_HOLMES.h"
#include "Class_GEDE.h"
#include "Struct_Track.h"
#include "Struct_Hitb.h"
#include "Struct_Gede.h"

//
// Fill Holmes' Tracks into RSEG table
//
RSEG getRSEG( Track_t *Track )
{
  RSEG r;
  r.zf  = r.ze  = Track->z;
  r.xf  = r.xe  = Track->x;
  r.yf  = r.ye  = Track->y;
  r.txf = r.txe = Track->tx;
  r.tyf = r.tye = Track->ty;
  r.pf  = r.pe  = .1;
  for ( unsigned int i= 0; i<10; ++i ) r.cvf[i] = r.cve[i] = Track->C[i];
  for ( unsigned int i=10; i<14; ++i ) r.cvf[i] = r.cve[i] = 0;
  r.cvf[14] = r.cve[14] = 1e4; 
  
  r.time = 0;
  r.dtim = 0;
  r.chi  = Track->chi2;
  r.fit  = 0;
  r.cmp  = 1;
  
  return r;
}


//
// Create RHITs for a segment
//
void RSEGtoRHIT( ArtePointer<RSEG> rseg )
{
  vector<ArtePointer<RHIT> > rhits;
  for( ArteRelationIterator<HITB> hitb = rseg->hitb.begin();
       hitb != rseg->hitb.end(); ++hitb ) {
    
    RHIT rh;
    rh.x  = rseg->xf+((*hitb)->z-rseg->zf)*rseg->txf;
    rh.y  = rseg->yf+((*hitb)->z-rseg->zf)*rseg->tyf;
    rh.tx = rseg->txf;
    rh.ty = rseg->tyf;
    rh.p  = 0;
    rh.q  = 0;
    ArtePointer<RHIT> rhit = ArteTable<RHIT>::push_back( rh );
    
    // createRelation( rseg,  rhit, &RSEG::rhit );
    rhits.push_back( rhit );
    createRelation( *hitb, rhit, &HITB::rhit );
  }
  for( vector<ArtePointer<RHIT> >::reverse_iterator rhit
	 = rhits.rbegin(); rhit!=rhits.rend(); ++rhit ) {
    createRelation( rseg, *rhit, &RSEG::rhit );
  }
}

//
// =========================================
// silicon standalone reconstruction program 
// =========================================
//
int HolmesEvent()
{
  static HOLMES_class *HOLMESp = 0; 

  char *routine = "HolmesEvent";

  // Fill GEDE_class in case of MC
  if ( !HOLMESp && CheckMC() && holmes_iol.use_holmes.value )
    sildat_notifyGede( HolmesFillGEDEfromARTE() );

  // initialize holmes
  HOLMESp = holmes_init();
  if ( !HOLMESp )
    return GRR_OK;

  // get target spot(s)
  float *xtarget, *ytarget, *ztarget;
  int   ntarget = HolmesGetTargets( &xtarget, &ytarget, &ztarget );

  // fill Holmes-internal hit list
  HolmesFillHITBfromARTE( HOLMESp );

  // Do track finding - loop over target spots and sector combinations
  HOLMESp->clearTracklist();
  for ( int i=0; i<ntarget; i++ )
    for ( int icomb=0; icomb<HOLMESp->nsector; icomb++ )
      HOLMESp->findTracks( HOLMESp->sector[icomb],
			   xtarget[i], ytarget[i], ztarget[i],
			   holmes_iol.target_window.value );

  // print-out if requested
  if ( holmes_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO, "Found %d hits and %d tracks.",
		HOLMESp->nhits(), HOLMESp->ntracks() );
  if ( holmes_iol.dump_hits.value )
    HOLMESp->dumpHitlist();
  if ( holmes_iol.dump_tracks.value )
    HOLMESp->dumpTracklist();

  // loop over tracks
  for ( int itrack=0; itrack<HOLMESp->ntracks(); itrack++ ) {

    // store track in RSEG table
    Track_t *Track = HOLMESp->trackP( itrack, -4.7 );

    ArteTable<RSEG>::push_back( getRSEG( Track ) );
    ArtePointer<RSEG> rseg( ArteTable<RSEG>::size() );

    // store corresponding hits in RHIT table
    for (int ihit=0; ihit<Track->nhit; ihit++) {

      int ihitb = HOLMESp->hitbP( Track->hit[ihit] )->hitb;

      ArtePointer<HITB> hitb( ihitb );
      createRelation( hitb, rseg, &RSEG::hitb );
    }

    // update RHIT table 
    RSEGtoRHIT( rseg );

  }

  return GRR_OK;
}


// --------------------------------
//
//      Some Utility Routines
//
// --------------------------------

GEDE_class* HolmesFillGEDEfromARTE()
{
  // Create empty GEDE class
  GEDE_class* GEDEp = new GEDE_class( (char*) 0 );

  // loop over GEDE planes
  for ( ArteTable<GEDE>::iterator igede = ArteTable<GEDE>::begin();
	igede != ArteTable<GEDE>::end(); igede++ ) {
    ArtePointer<GEDE> gede = *igede;
    if ( gede->cmp == 1 ) {
  
      // fill Gede_t structure with ARTE values
      Gede_t mygede;
      mygede.gede   = gede.index();
      mygede.side   = gede.index();
      char slname[5];
      strncpy( slname, gede->gesl->name, 4 ); 
      slname[4] = '\0';
      mygede.sl     = atoi( slname+2 );
      mygede.lay    = gede->lay;
      mygede.sec    = gede->sec;
      mygede.pitch  = gede->pit;
      mygede.alpha  = 180. / M_PI * atan2( gede->sin, gede->cos );
      if ( mygede.alpha < -10 )
	mygede.alpha += 360.;
      mygede.s      = gede->sin;
      mygede.c      = gede->cos;
      mygede.d0     = - gede->cos * gede->sxmn
                      + gede->sin * gede->symn
	              + gede->xfw[0];
      mygede.dmin   = gede->xfw[0];
      mygede.sxmn   = gede->sxmn;
      mygede.sxmx   = gede->sxmx;
      mygede.symn   = gede->symn;
      mygede.symx   = gede->symx;
      mygede.z      = ( gede->zmin + gede->zmax ) / 2.;      
      mygede.status = 1;            // all MC sides are perfect of course
      mygede.sign   = gede->xfw[2];
      *mygede.text  = '\0';
      
      // determine the type of view of the current plane
      if( -10 < mygede.alpha && mygede.alpha <  10 ||
	  170 < mygede.alpha && mygede.alpha < 190 ||
	  350 < mygede.alpha && mygede.alpha < 370   )
	mygede.view = 0;
      else
	mygede.view = 1;

      // Determine back side of the current detector
      //
      // !!!                                          !!!
      // !!!  here we use the assumption that layers  !!!
      // !!!  1 <-> 2 and 3 <-> 4 build pairs         !!!
      // !!!                                          !!!
      mygede.pair  = 0;
      for ( ArteTable<GEDE>::iterator igedeback = ArteTable<GEDE>::begin();
	    igedeback != ArteTable<GEDE>::end(); igedeback++ ) {
	ArtePointer<GEDE> gedeback = *igedeback;
	if (     gedeback->cmp == gede->cmp
	     && !strncmp( gedeback->gesl->name, gede->gesl->name, 4 )
	     &&  gedeback->sec == gede->sec
	     && (   gedeback->lay == 2 && gede->lay == 1
	         || gedeback->lay == 1 && gede->lay == 2
	         || gedeback->lay == 4 && gede->lay == 3
	         || gedeback->lay == 3 && gede->lay == 4 ) )
	  mygede.pair = gedeback.index();
      }

      // fill additional information of a gede struct
      mygede.dd    = 0;
      mygede.grad  = 0;   
      mygede.hesse = 0;   
      mygede.v[0]  =   mygede.c;
      mygede.v[1]  = - mygede.s;
      mygede.v[2]  =   mygede.c * mygede.z;
      mygede.v[3]  = - mygede.s * mygede.z;
      for (int i=0; i<4; i++)
	for (int j=0; j<4; j++)
	  mygede.vv[4*i+j] = mygede.v[i] * mygede.v[j];

      // now append to gede table
      GEDEp->gedeAppend( &mygede );
      
    }
  }

  return GEDEp;

}

//
//   Fill Holmes-internal hitlist from ARTE HITB table
//
void HolmesFillHITBfromARTE( HOLMES_class *HOLMESp )
{

  // clear the hit maps
  HOLMESp->clearHitlist();

  // refill the map from the HITB table
  for( ArteTable<HITB>::iterator iHITB = ArteTable<HITB>::begin();
       iHITB != ArteTable<HITB>::end(); iHITB++ ) {
    ArtePointer<HITB> ihit = *iHITB;

    if ( ihit->gede == 0 )              continue;

    Gede_t *GedeP = HOLMESp->gedeTableP( ihit->gede.index() );
    
    if ( !GedeP )                       continue;
    if ( ihit->q * GedeP->sign < 0 )    continue; 
    //    if ( fabs(ihit->q) > 100 )          continue;

    // fill next hitb into the map
    Hitb_t hitb;
    hitb.u      = ihit->x;
    hitb.cog    = -1;
    hitb.charge = ihit->q;
    hitb.sn     =  1;
    hitb.flag   =  1;
    hitb.gede   = GedeP;
    hitb.width  =  1;
    hitb.hitb   = ihit.index();

    HOLMESp->addtoHitlist(hitb);
  }

  return;
}
