#include "NormalDistribution.hh"
#include <math.h>

extern "C" float erf_(const float&) ;
extern "C" float erfc_(const float&) ;
extern "C" float prob_(const float&, const int&);
extern "C" float chisin_(const float&, const int&); //inverse of prob_
extern "C" float gausin_(const float&);
extern "C" float gamdis_(const float&, const float&);
extern "C" float gammf_(const float&) ;

double NormalDistribution::Prob(double chi2, int dof) {
  return prob_(float(chi2),dof);
}

double NormalDistribution::InvProb(double prob, int dof) {
  return chisin_(float(prob),dof);
}

double NormalDistribution::Erf(double noSigmas) {
  return double(erf_(float(noSigmas/sqrt(2.0))));
};

double NormalDistribution::InvErf(double fraction) {
  return double(gausin_((1+fraction)/2));
};

double NormalDistribution::InvErfAlt(double fraction) {
  if(fraction<=0) return 0;
  if(fraction>1) return 1000;
  
  double a=1.0;
  double res, dresda, da;
  int i=0;
  
  double dresda_alt;
  do {
    res = erf_(a) - fraction ;
    dresda = 2/sqrt(M_PI) * exp(-a*a) ;
    
    dresda_alt = (erf_(a+0.01)-erf_(a))/0.01 ;

    da = -res/dresda ;
    i++;
    a += da ;

  } while (fabs(res) > 0.0000001 && i < 20) ;
  return a*sqrt(2.0);
};


double NormalDistribution::PartialInnerVariance(double noSigmas) {
  if(noSigmas<=0) return 0;
  double extra = 2*noSigmas*exp(-noSigmas*noSigmas/2)/sqrt(2*M_PI) ;
  return 1 - extra/Erf(noSigmas) ;
};

double NormalDistribution::PartialOuterVariance(double noSigmas) {
  if(noSigmas<=0) return 1;
  double extra = 2*noSigmas*exp(-noSigmas*noSigmas/2)/sqrt(2*M_PI) ;
  return  1 + extra/(1-Erf(noSigmas)) ;
};

double NormalDistribution::PartialOuterWidth(double noSigmas) {
  return sqrt(PartialOuterVariance(noSigmas));
};

double NormalDistribution::PartialInnerWidth(double noSigmas) {
  return sqrt(PartialInnerVariance(noSigmas));
};


// int_0^a r^n exp(-r^2 /2) dr
double NormalDistribution::SpecialIntegral(double a, int n) {
  if(n==0) return erf_(a/sqrt(2.0))*sqrt(M_PI/2);
  if(n==1) return  1 - exp(-a*a/2) ;
  return - pow(a,n-1) * exp(-a*a/2) + (n-1) * SpecialIntegral(a,n-2);
};

// int_a^infty r^n exp(-r^2 /2) dr
double NormalDistribution::SpecialIntegralC(double a, int n) {
  if(n==0) return (1-erf_(a/sqrt(2.0)))*sqrt(M_PI/2);
  if(n==1) return  exp(-a*a/2) ;
  return pow(a,n-1) * exp(-a*a/2) + (n-1) * SpecialIntegralC(a,n-2);
};

double NormalDistribution::PartialInnerVariance(double noSigmas, int n)
{
  double a = noSigmas ;
  if(n==0) return 1 - 2*a*exp(-a*a/2)/(sqrt(2*M_PI)*erf_(a/sqrt(2.0)));
  return SpecialIntegral(a,n+2)/((n+1)*SpecialIntegral(a,n)) ;
}

double NormalDistribution::PartialOuterVariance(double noSigmas, int n)
{
  double a = noSigmas ;
  if(n==0) return 1 + 2*a*exp(-a*a/2)/(sqrt(2*M_PI)*(1-erf_(a/sqrt(2.0))));
  return SpecialIntegralC(a,n+2)/((n+1)*SpecialIntegralC(a,n)) ;
}

// double NormalDistribution::SpecialIntegral(double rmax, double rpow) {
//   float x = rmax*rmax/2 ;
//   float a = (rpow+1)/2 ;
//   return pow(2.0,(rpow-1)/2.) * gamdis_(x,a) * gammf_(a) ;
// };

// double NormalDistribution::SpecialIntegralC(double rmax, double rpow)
// {
//   float x = rmax*rmax/2 ;
//   float a = (rpow+1)/2 ;
//   return pow(2.0,(rpow-1)/2.) * (1-gamdis_(x,a)) * gammf_(a) ;
// };

// double NormalDistribution::PartialInnerVariance(double noSigmas, double n)
// {
//   double a = noSigmas ;
//   if(n==0) return 1 - 2*a*exp(-a*a/2)/(sqrt(2*M_PI)*erf_(a/sqrt(2.0)));
//   return SpecialIntegral(a,n+2)/((n+1)*SpecialIntegral(a,n)) ;
// }

// double NormalDistribution::PartialOuterVariance(double noSigmas, double n)
// {
//   double a = noSigmas ;
//   if(n==0) return 1 + 2*a*exp(-a*a/2)/(sqrt(2*M_PI)*(1-erf_(a/sqrt(2.0))));
//   return SpecialIntegralC(a,n+2)/((n+1)*SpecialIntegralC(a,n)) ;
// }

