#include <algorithm>
#include "wax/OTrEvent.hh"
#include "wax/OTrTrack.hh"
#include "wax/OTrTrackCandidate.hh"
#include "wax/OTrSuperCluster.hh"

#ifdef USE_ARTE
 #include <arte/arte.hh>
 #include <arte/artefunc.h>
#endif

#ifdef USE_ROOT
 #include <TH2.h>
 #include <TPad.h>
#endif

//__________________________________________________________________
// OTrEvent
//
// Container for:
// * SuperLayerEvents
// Functionality:
// * Reconstruction of tracks
//
// Author: Wouter Hulsbergen && Ivar Siccama
//__________________________________________________________________

#ifdef __OTR_CLASSDEF__
ClassImp(OTrEvent) ;
#endif //USE_ROOT

bool OTrEvent::fgAppendClustersToRSEG = false ;

//#define __OTR_RECOINFO__

OTrEvent::~OTrEvent() {
  // Destructor. 
  for( TrackContainer::iterator tr=fTrackCandidates.begin();
       tr != fTrackCandidates.end(); tr++) delete (*tr);
  fTrackCandidates.erase( fTrackCandidates.begin(), fTrackCandidates.end() );
} ;

OTrEvent::OTrEvent(int id) : fId(id) 
{
  // Constructor with event id.
} ;

OTrHit*
OTrEvent::Add( OTrHit* hit ) 
{
  // Add a hit to the event.
  return fSuperLayerEvents[hit->GetWire()->GetSuperLayerId()].Add(hit) ;
};

int
OTrEvent::SearchClusters() 
{
  // Search clusters in all superlayers.
  int n=0;
  for(SuperLayerEventContainer::iterator slEventIt = fSuperLayerEvents.begin();
      slEventIt != fSuperLayerEvents.end(); slEventIt++)
    //if((*slEventIt).first==2) // only reconstruct MC2 (need some speed :)
    n += (*slEventIt).second.SearchClusters() ;
  return n;
}
	
int
OTrEvent::GetNHits() const 
{
  // Return the number of hits in the event.
  int n=0;
  for(SuperLayerEventContainer::const_iterator slEventIt = fSuperLayerEvents.begin();
      slEventIt != fSuperLayerEvents.end(); slEventIt++)
    n += (*slEventIt).second.GetNHits() ;
  return n;
}

int
OTrEvent::GetNClusters() const 
{
  // Return the number of hits in the event.
  int n=0;
  for(SuperLayerEventContainer::const_iterator slEventIt = fSuperLayerEvents.begin();
      slEventIt != fSuperLayerEvents.end(); slEventIt++)
    n += (*slEventIt).second.GetClusters().size() ;
  return n;
}

void
OTrEvent::DumpHits(class ostream &os) const 
{
  // Call OTrHit::Dump for each hit.
  for(SuperLayerEventContainer::const_iterator slEventIt = fSuperLayerEvents.begin();
      slEventIt != fSuperLayerEvents.end(); slEventIt++)
    (*slEventIt).second.DumpHits(os) ;
};

void
OTrEvent::DumpClusters(class ostream &os) const 
{
  // Call OTrSuperCluster::Dump for each cluster.
  for(SuperLayerEventContainer::const_iterator slEventIt = fSuperLayerEvents.begin();
      slEventIt != fSuperLayerEvents.end(); slEventIt++)
    (*slEventIt).second.DumpClusters(os) ;
};

void
OTrEvent::Write(ostream& ofs) const
{
  // Write asci format to a file.
  static int eventId=0;
  ofs << eventId << " " << GetNHits() << " " ; 
  for(SuperLayerEventContainer::const_iterator 
	slEventIt = fSuperLayerEvents.begin();
      slEventIt != fSuperLayerEvents.end(); slEventIt++)
    for(OTrSuperLayerEvent::HitContainer::const_iterator
	  hitIt=(*slEventIt).second.GetHits().begin() ; 
	hitIt != (*slEventIt).second.GetHits().end() ; hitIt++) {
      OTrChannel* chanIt = (*hitIt)->GetChannel() ;
      int channelId = 128*chanIt->GetTdcId() + chanIt->GetChannelId() ;
      ofs <<  channelId << " " << (int)(*hitIt)->GetTdcCounts() << " " ;
    };
  cout << endl ;
  eventId++;
  return ;
};

void
OTrEvent::SortHits() 
{
  // Call OTrSuperLayerEvent::SortHits for each superlayer, which sort
  // hits in gede and wirenumber.
  for(SuperLayerEventContainer::iterator slEventIt = fSuperLayerEvents.begin();
      slEventIt != fSuperLayerEvents.end(); slEventIt++)
    (*slEventIt).second.SortHits();
};

#ifdef USE_ROOT
void
OTrEvent::PerformHistogramming() 
{
  // Call OTrSuperLayerEvent::PerformHistogramming for each superlayer.
  for(SuperLayerEventContainer::iterator slEventIt = fSuperLayerEvents.begin();
      slEventIt != fSuperLayerEvents.end(); slEventIt++)
    (*slEventIt).second.PerformHistogramming() ;
};
#endif

inline bool CompareTargetChi2( OTrTrackCandidate* lhs, 
			       OTrTrackCandidate* rhs) {
  return (lhs->GetClusters().size()> 
	  rhs->GetClusters().size()) ||
    ((lhs->GetClusters().size()==
      rhs->GetClusters().size()) && 
     lhs->GetTargetChi2() < rhs->GetTargetChi2()) ;
}

int
OTrEvent::SearchTrackCandidates() 
{
  // The track search algorithm defines tarcks in the following way:
  // 1) A track is made up of hit clusters 
  // 2) A track has no cluster or one cluster per superlayer 
  // 3) No track can be `sub-track' of another track
  // It combines clusters in different superlayers into tracks.
#ifdef USE_ARTE
  AppendClustersToRSEG() ;
#endif //USE_ARTE


#ifdef USE_ROOT
  static TH2F* rawTrackXYH2 = new TH2F("rawTrackXYH2","",100,-100,100,100,-100,100) ;
  static TH2F* tcTxVsTxH2 = 
    new TH2F("tcTxVsTxH2","",100,-0.26,0.06,100,-0.14,-0.04) ;
  //new TH2F("tcTxVsTxH2","",100,-0.3,0.06,100,-0.14,-0.0) ;
  static TH2F* tcTxVsTxCutH2 = 
    new TH2F("tcTxVsTxCutH2","",100,-0.26,0.06,100,-0.14,-0.04) ;
  //new TH2F("tcTxVsTxH2","",100,-0.3,0.06,100,-0.14,-0.0) ;
#endif
  if(fSuperLayerEvents.size()<2) { return 0 ; } ;

  // ******** Delete previous *********************
  for(TrackContainer::iterator trackIt=fTrackCandidates.begin() ;
      trackIt != fTrackCandidates.end() ; trackIt++)
    delete *trackIt ;
  fTrackCandidates.erase(fTrackCandidates.begin(),fTrackCandidates.end()) ;
  TrackContainer trackSeeds ;

  /** In the following the difference between a seed and a candidate
      is quite arbitrary: I just need two containers. **/
  //for(SuperLayerEventContainer::reverse_iterator sleIt=fSuperLayerEvents.rbegin() ;
  //    sleIt!=fSuperLayerEvents.rend(); sleIt++) {
  for(SuperLayerEventContainer::iterator sleIt=fSuperLayerEvents.begin() ;
      sleIt!=fSuperLayerEvents.end(); sleIt++) {
    vector<bool> clusterIsAdded((*sleIt).second.GetClusters().size(),false) ;
    /*** Loop over all seedtracks, combine with clusters and if valid we
         have new candidate ***/
    for(TrackContainer::iterator trackIt=fTrackCandidates.begin() ;
	trackIt!=fTrackCandidates.end() ; trackIt++) {
      OTrTrackCandidate* trackSeed = new OTrTrackCandidate(**trackIt) ;
      bool clusterFound=false ;
      int clId=0;
      for(OTrSuperLayerEvent::ClusterContainer::const_iterator clIt=
	    (*sleIt).second.GetClusters().begin() ;
	  clIt!=(*sleIt).second.GetClusters().end() ; clIt++, clId++) {
	if(trackSeed->AddIfValid(*clIt)) {
	  clusterIsAdded[clId] = true ;
	  clusterFound=true ;
	  trackSeeds.push_back(trackSeed) ;
	  trackSeed = new OTrTrackCandidate(**trackIt) ;
	} ;
#ifdef USE_ROOT
	tcTxVsTxH2->Fill(float((*clIt)->GetTx()), 
			 float(trackSeed->GetClusters().front()->GetTx())) ;
	if(fabs((*clIt)->GetTy()-trackSeed->GetClusters().front()->GetTy())<0.05)
	  tcTxVsTxCutH2->Fill(float((*clIt)->GetTx()), 
			      float(trackSeed->GetClusters().front()->GetTx())) ;
#endif //USE_ROOT
      }
      /** If no corresponding cluster is found, the original track is
          added to the list of seeds **/
      if(!clusterFound) trackSeeds.push_back(trackSeed) ;
      else delete trackSeed ;
    }
    /** If there is no track found with a cluster, the cluster itself
        is a new trackcandidate **/
    int clId=0;
    for(OTrSuperLayerEvent::ClusterContainer::const_iterator clIt=
	  (*sleIt).second.GetClusters().begin() ;
	clIt!=(*sleIt).second.GetClusters().end() ; clIt++, clId++)
      if(!clusterIsAdded[clId])
	trackSeeds.push_back(new OTrTrackCandidate(*clIt)) ;
    /** All seed track become candidates, and all previous candidates
        are deleted **/
    fTrackCandidates.swap(trackSeeds) ;
    for(TrackContainer::iterator trackIt=trackSeeds.begin() ;
	trackIt != trackSeeds.end() ; trackIt++)
      delete *trackIt ;
    trackSeeds.erase(trackSeeds.begin(),trackSeeds.end()) ;
  }
  /*** (Remove all track candidates which have less than 2 clusters
       assigned to them) ***/
  for(TrackContainer::iterator trackIt=fTrackCandidates.begin() ;
      trackIt!=fTrackCandidates.end() ; trackIt++)
    if((*trackIt)->GetClusters().size()<2) {
      delete *trackIt ;
      fTrackCandidates.erase(trackIt--) ;
    } 
#ifdef __OTR_RECOINFO__
  cout << "Found " << fTrackCandidates.size() << " track candidates " << endl ;
#endif

  /*** Fit all track candidates in global trackfit and remove if fit fails ***/
  for(TrackContainer::iterator trackIt=fTrackCandidates.begin() ;
      trackIt!=fTrackCandidates.end() ; trackIt++)
#ifdef USE_ARTE
    if(!((*trackIt)->RangerFit())) {
#else
    if(!((*trackIt)->Fit(true))) {
#endif
      delete *trackIt ;
      fTrackCandidates.erase(trackIt--) ;
    } ;
#ifdef __OTR_RECOINFO__
  cout << "Left " << fTrackCandidates.size() << " after track fit . " << endl ;
#endif

  /*** Order tracks in `target compatibility' ***/
  //fTrackCandidates.sort(CompareTargetChi2) ;
  sort(fTrackCandidates.begin(),fTrackCandidates.end(),CompareTargetChi2) ;
  
  /*** Clean up overlapping tracks ***/
  for(TrackContainer::iterator trackIt=fTrackCandidates.begin() ;
      trackIt!=fTrackCandidates.end() ; trackIt++) {
    TrackContainer::iterator trackIt2 = trackIt;
    trackIt2++;
    for( ; trackIt2!=fTrackCandidates.end() ; trackIt2++) {
      if((*trackIt2)->CleanUp(*(*trackIt)) && !((*trackIt2)->Valid())) {
	delete *trackIt2 ;
	fTrackCandidates.erase( trackIt2-- ) ;
      } 
    }
  }
#ifdef __OTR_RECOINFO__
  cout << "and " << fTrackCandidates.size() << " after clean up." << endl ;
#endif
  return fTrackCandidates.size() ;
}
  
int
OTrEvent::FitTracks()
{  
  int n=0;
  for(TrackContainer::iterator trackIt = fTrackCandidates.begin() ;
      trackIt != fTrackCandidates.end() ; trackIt++)
    //  (*trackIt).RangerFit() ;
    if((*trackIt)->Fit()) {
      n++;
      //(*trackIt)->RangerFitWouterTrack() ;
    };
  return n ;
};

#ifdef USE_ARTE
void
OTrEvent::AppendRSEG()  
{
  // Copy reconstructed tracks into the RSEG bank. If
  // fgAppendClustersToRSEG is set, not the tracks, but the clusters
  // are appended to RSEG.
  if(!fgAppendClustersToRSEG) {
    for(TrackContainer::iterator trackIt = fTrackCandidates.begin() ;
	trackIt != fTrackCandidates.end() ; trackIt++)
      (*trackIt)->AppendRSEG();
  } else {
    AppendClustersToRSEG() ;
  }
}

void
OTrEvent::AppendClustersToRSEG()
{
  for(SuperLayerEventContainer::iterator slEventIt = fSuperLayerEvents.begin();
      slEventIt != fSuperLayerEvents.end(); slEventIt++)
    (*slEventIt).second.AppendRSEG() ;
}

void 
OTrEvent::AppendHITB()  
{
  // The hits should be ordered before calling this method.
  for(SuperLayerEventContainer::iterator slEventIt = fSuperLayerEvents.begin();
      slEventIt != fSuperLayerEvents.end(); slEventIt++)
    (*slEventIt).second.AppendHITB() ;
}

#else //USE_ARTE

void
OTrEvent::AppendRSEG()  
{
  //Dummy routine for use outside ARTE.
}

void
OTrEvent::AppendHITB()  
{
  //Dummy routine for use outside ARTE.
}

void
OTrEvent::AppendClustersToRSEG()
{
  //Dummy routine for use outside ARTE.
}

#endif //USE_ARTE


#ifdef USE_ROOT

void
OTrEvent::DrawHits() {
  for(SuperLayerEventContainer::iterator slEventIt = fSuperLayerEvents.begin();
      slEventIt != fSuperLayerEvents.end(); slEventIt++)
    (*slEventIt).second.DrawHits() ;
};

void
OTrEvent::DrawClusters() {
  for(SuperLayerEventContainer::iterator slEventIt = fSuperLayerEvents.begin();
      slEventIt != fSuperLayerEvents.end(); slEventIt++)
    (*slEventIt).second.DrawClusters() ;
};

void
OTrEvent::DrawTrackCandidates( TPad* pad ) {
  int color = 2;
  for( TrackContainer::iterator tr = fTrackCandidates.begin();
       tr != fTrackCandidates.end(); tr++)
    (*tr)->Draw( pad, color ) ;
};

void
OTrEvent::Streamer(TBuffer &R__b)
{
  // Stream an object of class OTrEvent
  if (R__b.IsReading()) {
    int nsle;
    R__b >> fId >> nsle ;
    for(int i=0; i< nsle; i++) {
      int sleId ;
      R__b >> sleId ;
      fSuperLayerEvents[sleId].Streamer(R__b) ;
    };
  } else {
    R__b << fId << fSuperLayerEvents.size() ;
    for(SuperLayerEventContainer::iterator sleIt=fSuperLayerEvents.begin() ;
	sleIt!=fSuperLayerEvents.end() ; sleIt++) {
      R__b << (*sleIt).first ;
      (*sleIt).second.Streamer(R__b) ;
    } ;
  } ;
} ;
#endif
