#include <iomanip>
#include <CLHEP/Matrix/SymMatrix.h>
#include <CLHEP/Matrix/Vector.h>
#include "MyProfile.hh"

#include "OTrRTProjection.hh"
#include "OTrTrackDescription.hh"
#include "OTrTrack.hh"
#include "OTrParameterizedRTRelation.hh"
#include "NormalDistribution.hh"
#include "OTrSection.hh"
#include "OTrWire.hh"
#include "OTrReconstructedTrackInterface.hh"
#include "OTrNodeProjection.hh"
#include "OTrHit.hh"
#include "OTrChannelHistogrammer.hh"


#ifdef USE_ROOT
 #include <TProfile.h>
 #include <TH1.h>
 #include <TCanvas.h>
 #include <TGraphErrors.h>
#endif //USE_ROOT

#define MIN_RES_CUT2 (2000*2000) //ns^2
#define MAX_NO_SIGMAS 0.0

float OTrRTProjection::fgMaxNormalizedResidualForT = 1 ;
float OTrRTProjection::fgMaxNormalizedResidualForSigma2 = 4 ;
int OTrRTProjection::fgUpdateFrequency=10000 ; // residuals

#define TRACKFITT
#undef TRACKFITT

OTrRTProjection::OTrRTProjection(OTrParameterizedRTRelation* rtr, bool fitT, bool fitSigma) :
  fRTRelation(rtr), fNTParameters(0), fNSParameters(0), fCanvas(0), fColor(1), fNProjectedHits(0)
{ 
  if(fitT) fNTParameters = fRTRelation->GetNoTParameters() ;
  if(fitSigma) fNSParameters = fRTRelation->GetNoSigma2Parameters() ;

  fDChi2DP_T = new HepVector(GetNTParameters(),0) ;
  fDChi2DP_S = new HepVector(GetNSParameters(),0) ;
  fD2Chi2DP2_T = new HepSymMatrix(GetNTParameters(),0) ;
  fD2Chi2DP2_S = new HepSymMatrix(GetNSParameters(),0) ;
  fPrevDelta_T = new HepVector(GetNTParameters(),0) ;
  fPrevDelta_S = new HepVector(GetNSParameters(),0) ;
  const int nBins = 50;
  fResidualProfile = new MyProfile(nBins,fRTRelation->GetRMin(),fRTRelation->GetRMax()) ;
   
  // Monitoring
#ifdef USE_ROOT
  //ROOTSTUFF
  char* name ;
  if(fRTRelation->GetRMax()>0.4) name = "10mm" ; else name = "05mm" ; //ugly ... :-)
  float rmax = fRTRelation->GetRMax() ; float rmin=fRTRelation->GetRMin();

  // Book histograms
  TString resVsRname(name) ;
  resVsRname.Prepend("ResVsR") ;
  fTResidualVsR = new TProfile(resVsRname,resVsRname,50,rmin,rmax,-100,100) ;
  fTResidualVsR->SetMinimum(-10) ;
  fTResidualVsR->SetMaximum(10) ;

  TString norresName(name) ;
  norresName.Prepend("NormRes2VsR") ;
  fNormalizedResidual2VsR = new TProfile(norresName,norresName,nBins,rmin,rmax,0,50) ;
  fNormalizedResidual2VsR->SetMinimum(0) ;
  fNormalizedResidual2VsR->SetMaximum(2) ;

  TString dName(name) ;
  dName.Prepend("recD") ;
  fRecDistanceH1 = new TH1F(dName,dName,nBins,-rmax,rmax) ;

  TString ndName(name) ;
  ndName.Prepend("nonrecD") ;
  fNonRecDistanceH1 = new TH1F(ndName,ndName,nBins,-rmax,rmax) ;
 
  TString tName(name) ;
  tName.Prepend("recTime") ;
  fRecTimeH1 = new TH1F(tName,tName,64,0,256) ;

  TString ntName(name) ;
  ntName.Prepend("nonRecTime") ;
  fNonRecTimeH1 = new TH1F(ntName,ntName,64,0,256) ;

  TString rresName(name) ;
  rresName.Prepend("RResidual") ;
  fRResidualH1 = new TH1F(rresName,rresName,50,-0.3,0.3) ;

  TString rcorresName(name) ;
  rcorresName.Prepend("RCorrectedResidual") ;
  fRCorrectedResidualH1 = new TH1F(rcorresName,rcorresName,50,-0.3,0.3) ;

  TString tresName(name) ;
  tresName.Prepend("Tresidual") ;
  fTResidualH1 = new TH1F(tresName,tresName,50,-70,70) ;

  TString chi2probName(name) ;
  chi2probName.Prepend("chi2prob") ;
  fChi2ProbH1 = new TH1F(chi2probName,chi2probName,100,0,1) ;

  TString rescorName(name) ;
  rescorName.Prepend("ResidualCorrection") ;
  fResCorrectionH1 = new TH1F(rescorName,rescorName,100,0,20) ;

  TString tempName(name) ;
  tempName.Prepend("temp") ;
  fTempH1 = new TH1F(tempName,tempName,50,-0.3,0.3) ;

  float upperSigmaLimit = 0.10 ;
  #ifdef TRACKFITT
  upperSigmaLimit = 30 ;
  #endif //TRACKFITT
  // Create canvas and draw everything
  TString canvasName(name) ;
  canvasName.Prepend("OTrRTProjection ") ;
  fCanvas = new TCanvas(canvasName,canvasName,1) ;
  fCanvas->Show(); fCanvas->Clear();
  fCanvas->Divide(3,3);

  fCanvas->cd(1); gPad->DrawFrame(rmin,0,rmax,256) ; 
  fRTRelation->GetTGraph(50)->Draw("lp") ;
  
  fCanvas->cd(4); gPad->DrawFrame(rmin,0,rmax,upperSigmaLimit) ; 
  fRTRelation->GetSigmaGraph(50)->Draw("lp") ;
  
  fCanvas->cd(2); fTResidualVsR->Draw() ;
  fCanvas->cd(5); fNormalizedResidual2VsR->Draw() ;
  
  fCanvas->cd(3) ; fRecDistanceH1->Draw() ;
  fNonRecDistanceH1->Draw("same") ;
  fNonRecDistanceH1->SetLineColor(2) ;

  fCanvas->cd(6) ; fRecTimeH1->Draw() ;
  fNonRecTimeH1->Draw("same") ;
  fNonRecTimeH1->SetLineColor(2) ;

  fCanvas->cd(7) ; fRResidualH1->Draw() ;
  fCanvas->cd(7) ; fRCorrectedResidualH1->Draw("same") ;
  fCanvas->cd(8) ; fTResidualH1->Draw() ;
  fCanvas->cd(9) ; fChi2ProbH1->Draw() ;

  fCanvas->Update();
#endif //USE_ROOT
};


void
OTrRTProjection::Reset()
{
  fNProjectedHits =0 ;

  for(int row=0; row<(*fDChi2DP_T).num_row(); row++) {
    (*fDChi2DP_T)[row]=0;
    for(int col=0; col<=row; col++)
      (*fD2Chi2DP2_T)[row][col]=0;
  };

  for(int row=0; row<(*fD2Chi2DP2_S).num_row(); row++) {
    (*fDChi2DP_S)[row]=0;
    for(int col=0; col<=row; col++)
      (*fD2Chi2DP2_S)[row][col]=0;
  };

#ifdef USE_ROOT
  fResidualProfile->Reset() ;
  fTResidualVsR->Reset() ;
  fNormalizedResidual2VsR->Reset() ;
  fRecDistanceH1->Reset() ;
  fNonRecDistanceH1->Reset() ;
  fRecTimeH1->Reset() ;
  fNonRecTimeH1->Reset() ;
  fRResidualH1->Reset() ;
  fRCorrectedResidualH1->Reset() ;
  fTResidualH1->Reset() ;
  fChi2ProbH1->Reset() ;
  fResCorrectionH1->Reset() ;
#endif //USE_ROOT
};


void
OTrRTProjection::UpdateCanvas()
{
#ifdef USE_ROOT
  if(fCanvas) {
    for(int i=0; i<6; i++) {
      fCanvas->cd(i+1); gPad->Modified() ; 
    };
    fCanvas->Modified() ;
    fCanvas->Update();
  };
#endif
}


bool
OTrRTProjection::Add(const OTrReconstructedTrackInterface& track, const OTrHit* hit)
{
  // first calculate these static correction factors
  static double* fRecCutSigmaCorFacVec=0;
  static double* fNonrecCutSigmaCorFacVec=0;
  if(fRecCutSigmaCorFacVec==0) {
    // Actually, GetMinChi2Probability() should be static but you
    // cannot do that with a virtual function. Therefore I assume that
    // all tracks that come in are of the same type :
    const int maxNoDofs = 32;
    fRecCutSigmaCorFacVec = new double[32] ;
    fNonrecCutSigmaCorFacVec = new double[32] ;
    for(int m=1; m<maxNoDofs; m++) {
      double maxchi2 = 
	NormalDistribution::InvProb(1-track.GetMinChi2Probability(),m);
      fRecCutSigmaCorFacVec[m] = 
	1/sqrt(NormalDistribution::PartialInnerVariance(sqrt(maxchi2),m-1));
      fNonrecCutSigmaCorFacVec[m] = 
	1/sqrt(NormalDistribution::PartialOuterVariance(sqrt(maxchi2),m-1));
    };
  };
  
  // Here starts the routine
  if(!fRTRelation) fRTRelation = hit->GetChannel()->GetRTRelation() ; // extremely ugly, but it works ...
  OTrNodeProjection nodeProjection(hit) ;
  nodeProjection.Project(&track) ;
  if(nodeProjection.OnTrack()) {
    fNProjectedHits++ ;
    int nDofs= track.GetNDof() ;
    double R = fabs(nodeProjection.GetD()) ;
    double residualT = nodeProjection.GetResidualT() ;
    double residualR = nodeProjection.GetResidualR() ;
    ///double sigma2   = nodeProjection.GetSigmaT2() ;
    /// I want sigma at reconstructed distance instead
    double dTdR = fRTRelation->GetT(R).ddx ;
#ifdef TRACKFITT
    double sigmaT2 = fRTRelation->GetSigma2(R).GetValue() ;
    double sigmaR2 = sigmaT2 / sqr(dTdR) ;
#else
    double sigmaR2 = fRTRelation->GetSigma2(R).GetValue() ;
    double sigmaT2 = sigmaR2 * sqr(dTdR) ;
#endif //TRACKFITT
    bool hit_in_trackfit = track.FindHit(hit)!=0 ;
#ifdef USE_ROOT
    hit->GetChannel()->GetHistogrammer()->ProcessNodeProjection(&nodeProjection) ;
#endif
    double this_f2(1); //residual scaling factor
    double this_g(1); //factor for second derivative
    double df = track.GetCov().similarity(nodeProjection.GetNormdResdQ()) ;
    //HepDouble globaldf = 1 - (double)nDofs / (double)track.GetHits().size() ;
    //df=globaldf ;

    if(hit_in_trackfit) {
      this_g = 1 + df ;
      if(1-df>0) this_f2 = 1/(1-df) ;
      else       this_f2 = track.GetHits().size()/(double)nDofs ;
    } else {
      this_f2 = 1/(1+df) ;
      this_g  = 1 ;
    }
  
    HepDouble weight = 1/sigmaT2 ;//this_f2/sigma2 ; 
    HepDouble weightRes = 1-df ;

    //All reconstructed hits and all `nonreconstructed, but close'
    //hits are used for RTcalibration
    if(hit_in_trackfit || sqr(residualT/fgMaxNormalizedResidualForT) < sigmaT2 ) {
      const HepVector& dTdPar = fRTRelation->GetdTdPar(R) ;
      for(int k=0; k<GetNTParameters();k++)
	if(dTdPar[k]>0.) {
	  //Add contribution from dChi2dP_T
	  (*fDChi2DP_T)[k] += 2*weight*residualT*dTdPar[k] ; 
	  //Add contribution from partial d2Chi2dP2
	  for(int l=0; l<k+1;l++)
	    if(dTdPar[l]>0.)
	      (*fD2Chi2DP2_T)[k][l] += 2*weight*dTdPar[k]*dTdPar[l]*this_g ;
	}
    }
    
    // First correction: the res.scal.factor from `fitting'
    double residualScalFac = sqrt(this_f2) ;
    // Second correction: the chi^2 cut
    if(hit_in_trackfit) residualScalFac *= fRecCutSigmaCorFacVec[nDofs] ;
    else                residualScalFac *= fNonrecCutSigmaCorFacVec[nDofs+1] ;
#ifndef TRACKFITT
    // Third correction: lrsign close to wire
    //double sigmaR = sqrt(sigmaR2) ;
    //residualScalFac *= sigmaR/sqrt(PositiveNormalDistribution::Variance(R,sigmaR)) ;
#endif
    double correctedResidualT = residualScalFac*residualT ;
    double correctedResidualR = residualScalFac*residualR ;
    double normalizedResidual = correctedResidualT/sqrt(sigmaT2) ;
    //double normalizedResidual = correctedResidualR/sqrt(sigmaR2) ;
    double normalizedResidual2 = sqr(normalizedResidual) ;
    cout << "residualT: " << residualT << endl 
	 << "correctedResidualT: " << correctedResidualT << endl
	 << "normalizedResidualT: " << normalizedResidual << endl
	 << "sqrt(this_f2): " << sqrt(this_f2) << endl
	 << "sigmaT2: " << sigmaT2 << endl ;

    // For the sigma2 fit, hits with an unlikely high
    // residual-scaling-factor are a problem. Therefore we cut also
    // reconstructed hits.
    bool withinResidualCutForSigma2 = 
      normalizedResidual2 < sqr(fgMaxNormalizedResidualForSigma2 ) ; 
    if(withinResidualCutForSigma2) {
      //fResidualProfile->Fill(R,correctedResidualR,weightRes) ;
      fResidualProfile->Fill(R,correctedResidualR,weightRes) ;
      fNormalizedResidualProfile->Fill(R,normalizedResidual,weightRes) ;
    } 
    ////fResidualProfile->Fill(R,correctedResidualR) ;
    //fNormalizedResidualProfile->Fill(R,normalizedResidual) ;
#ifdef USE_ROOT
    // Fill histograms
    if(withinResidualCutForSigma2) {
      fRecDistanceH1->Fill((float)(nodeProjection.GetD())) ;
      fRecTimeH1->Fill((float)(hit->GetDriftTime())) ;
      fTResidualVsR->Fill(R,float(residualT)) ;
      fNormalizedResidual2VsR->Fill(R,float(normalizedResidual2)) ;
    } else {
      fNonRecDistanceH1->Fill((float)(nodeProjection.GetD())) ;
      fNonRecTimeH1->Fill((float)(hit->GetDriftTime())) ;
    }
    //if(hit->GetWire()->GetGede()->GetEStereoAngle()==OTrGede::kZeroDegrees)
    fRCorrectedResidualH1->Fill(correctedResidualR) ;     
    fRResidualH1->Fill(residualT/dTdR) ;
    fTResidualH1->Fill(residualT) ;
    
    fChi2ProbH1->Fill(track.GetChi2Prob()) ;
    fResCorrectionH1->Fill(this_f2) ;
#endif
    return true ;
  } //else cout << " not on track ! " << endl ; // OnTrack

  return false ;
};

bool
OTrRTProjection::FitSigma2()
{
  
  float thischi2=0; float thisweight =0; int N=0;
#ifdef USE_ROOT
  for(int i=1; i<=fNormalizedResidual2VsR->GetNbinsX(); i++) {
    double sig = fNormalizedResidual2VsR->GetBinError(i) ;
    if(sig>0) {
      double res = fNormalizedResidual2VsR->GetBinContent(i)-1 ;
      thischi2 += res*res/(sig*sig);
      thisweight += 1/(sig*sig);
      N++;
    };
  };
#endif
  cout << "Average res(chi2) for sigma2: " << sqrt(thischi2/thisweight) << "  "
       << thischi2/N << endl;
    
  bool convergedSigma2=true;
  if(GetNSParameters()>0) {//&& redo_tracksearch_flag) {
    int ifail;
    //  HepSymMatrix D2Chi2DP2Inv = D2Chi2DP2_S.inverse(ifail) ;
    //       if(ifail) cout << "Matrix inversion error for D2Chi2DP2_Sigma2!" << endl ;
    //       HepVector deltaParameters = - D2Chi2DP2Inv * DChi2DP_S ; 
    //       HepSymMatrix covParameters = 2 * D2Chi2DP2Inv ;
        
    fResidualProfile->Repair();
    //fResidualProfile->Print();
    for(int bin=0; bin<fResidualProfile->GetNoBins(); bin++) {
      HepDouble weight = 
	fResidualProfile->GetBinContent(bin).GetNoEntries() ;
      HepDouble measuredsigma2 = //cutVarCorFac*
	fResidualProfile->GetBinContent(bin).GetSecondMoment() ;
      //fResidualProfile->GetBinContent(bin).GetBareSecondMoment() ;
      HepDouble R = fResidualProfile->GetBinCenter(bin);
      HepDouble sigma2 = fRTRelation->GetSigma2(R).value ;
      if(weight>2) {
	const HepVector& dSigma2dPar = fRTRelation->GetdSigma2dPar(R) ;
	//OTrFunctionValue T = fRTRelation->GetT(R) ;
	HepDouble sigma2residual = sigma2-measuredsigma2 ;
	//HepDouble sigma2residual = 
	//  fTSigma2ResidualProfile->GetFirstMoment() ;
	weight /= sqr(2*measuredsigma2) ;
	int k;
	for(k=0; k<GetNSParameters();k++)
	  (*fDChi2DP_S)[k] += 2*weight*sigma2residual*dSigma2dPar[k] ;
	for(k=0; k<GetNSParameters();k++)
	  for(int l=0; l<k+1;l++)
	    (*fD2Chi2DP2_S)[k][l] += 2*weight*dSigma2dPar[k]*dSigma2dPar[l] ;
      } else cout << "Bin " << R 
		  << " has only weight " << weight << "!!!" << endl;
    };
    
    HepSymMatrix D2Chi2DP2Inv = fD2Chi2DP2_S->inverse(ifail) ;
    if(ifail) cout << "Matrix inversion error for D2Chi2DP2_Sigma2_2!"
		   << endl ;
    HepVector deltaParameters = - D2Chi2DP2Inv * (*fDChi2DP_S) ; 
    HepSymMatrix covParameters = 2 * D2Chi2DP2Inv ;
    
    HepVector oldParameters(fRTRelation->GetSigma2Parameters()) ;
    HepVector newParameters = oldParameters ;
    
    // Alternative -- `integration' method:
    HepVector sumBx(GetNSParameters(),0),sumBxSigma2(GetNSParameters(),0);
    
    for(int bin=0; bin<fResidualProfile->GetNoBins(); bin++) {
      float measuredsigma2 = 
	//fResidualProfile->GetBinContent(bin).GetSecondMoment() ;
	fResidualProfile->GetBinContent(bin).GetBareSecondMoment() ;
      float R = fResidualProfile->GetBinCenter(bin);
      HepVector basisFuncValues = fRTRelation->GetdSigma2dPar(R) ;
      sumBx += basisFuncValues ;
      sumBxSigma2 += measuredsigma2*basisFuncValues ;
    };
    HepVector newParametersB(GetNSParameters()) ;
    
    HepSymMatrix extraCov(GetNSParameters(),0) ;
    cout << "sigma2_parameters(old,new,delta,error)" << endl;
    for(int i=0; i<GetNSParameters(); i++)
      {
	HepDouble delta = deltaParameters[i] ;
	HepDouble error = sqrt(covParameters[i][i]) ;
	//covParameters[i][i] += delta*delta ;
	extraCov[i][i] = delta*delta ;
	//if(fabs(delta)>5*error) delta *=0.5; //!! to prevent oscillations
	delta*=0.5 ;
	if(delta*(*fPrevDelta_S)[i]<0) delta*=0.5 ;
	newParameters[i]  = oldParameters[i] + delta ;
	if(newParameters[i]<0) newParameters[i] -= delta/2 ;

	newParametersB[i] = sumBxSigma2[i]/sumBx[i] ; 

	convergedSigma2 = convergedSigma2 && fabs(delta) < error ;
	
	cout << setw(12) << setprecision(5) << oldParameters[i]
	     << setw(12) << setprecision(5) << newParameters[i]
	     << setw(12) << setprecision(5) << deltaParameters[i] 
	     << setw(12) << setprecision(5) << error
	     << setw(12) << setprecision(5) << sqrt((covParameters+extraCov)[i][i])
	     << setw(12) << setprecision(5) << newParametersB[i]
	     << endl ;
      }
    fRTRelation->UpdateSigma2Parameters(newParameters,
					covParameters+extraCov) ;
    (*fPrevDelta_S) = deltaParameters ;
  }
  return convergedSigma2 ;
}
  
bool OTrRTProjection::FitT()
{
  double thischi2=0, thisweight =0;
  int N=0;
  #ifdef USE_ROOT
  for(int i=1; i<=fTResidualVsR->GetNbinsX(); i++) {
    double sig = fTResidualVsR->GetBinError(i) ;
    if(sig>0) {
      double res = fTResidualVsR->GetBinContent(i) ;
      thischi2 += res*res/(sig*sig);
      thisweight += 1/(sig*sig);
      N++;
    };
  };
  #endif
  cout << "# projected hits : " << fNProjectedHits << endl ;
  cout << "Average res(chi2) for T: " << sqrt(thischi2/thisweight) << "  "
       << thischi2/N << endl;

  bool convergedT(true) ;

  if(GetNTParameters()>0) {
    int ifail ;
    HepSymMatrix D2Chi2DP2Inv = fD2Chi2DP2_T->inverse(ifail) ;
    if(ifail) {
      cout << "Matrix inversion error for D2Chi2DP2_T! " 
	   << *fD2Chi2DP2_T ;
    } else {
      HepVector deltaParameters = - D2Chi2DP2Inv * (*fDChi2DP_T) ; 
      HepSymMatrix covParameters = 2 * D2Chi2DP2Inv ;
    
      HepVector oldParameters(fRTRelation->GetTParameters()) ;
      HepVector newParameters = oldParameters ;
      
      HepSymMatrix extraCov(GetNTParameters(),0) ;
      cout << "rt_parameters(old,new,delta,error)" << endl;
      for(int i=0; i<GetNTParameters(); i++) {
	HepDouble delta = deltaParameters[i] ;
	HepDouble error = sqrt(covParameters[i][i]) ;
	extraCov[i][i] = delta*delta ;
	
	//if(fabs(delta)>5*error) delta *=0.5; //!! to prevent oscillations
	//delta *= 0.5 ;
	if((*fPrevDelta_T)[i]*delta<0) delta*=0.5 ;
	newParameters[i] = oldParameters[i] + delta ;
	
	//if(0.5*error > fabs(newParameters[i])) newParameters[i] = 0;
	//else 
	convergedT = convergedT && fabs(delta) < 0.5*error ;
	
	cout << setw(12) << setprecision(5) << oldParameters[i]
	     << setw(12) << setprecision(5) << newParameters[i]
	     << setw(12) << setprecision(5) << delta
	     << setw(12) << setprecision(5) << error
	     << endl ;
      }
      fRTRelation->UpdateTParameters(newParameters,
				     covParameters+extraCov);
      (*fPrevDelta_T) = deltaParameters ;
      //IF THINGS GO REALLY WRONG:
      //  if(!redo_tracksearch_flag && fRTRelation->GetT(0.).ddx<0) {
      // 	 fRTRelation->UpdateTParameters(oldParameters);
      // 	 //return;
      //        }
    }
  };
  return convergedT ;
};

void
OTrRTProjection::Draw()
{
#ifdef USE_ROOT
  if(fCanvas) { 
    TGraph* tgr;
    fColor++ ;
    fCanvas->cd(1) ;
    tgr = fRTRelation->GetTGraph(50) ;
    tgr->SetLineColor(fColor) ; tgr->Draw() ;
    fCanvas->cd(4);
    tgr = fRTRelation->GetSigmaGraph(50) ;
    tgr->SetLineColor(fColor) ; tgr->Draw() ;
  };
#endif
};
