#ifdef USE_ROOT
#include <math.h>
#include <TCanvas.h>
#include "wax/OTrResidualProfile.hh"

inline double sqr(double x) { return x*x ; }

OTrResidualProfile::OTrResidualProfile(TString& name, TString& title, 
				       int nxbins, float xmin, float xmax, 
				       int nybins, float ymin, float ymax) 
  : TH2F(name,title,nxbins,xmin,xmax,nybins,ymin,ymax)
{
  fWidth    = new double[nxbins] ;
  fWidthErr = new double[nxbins] ;
  fFitFlag  = new int[nxbins] ;
  fSumW     = new double[nxbins+2] ;
  fSumY     = new double[nxbins+2] ;
  fSumY2    = new double[nxbins+2] ;
  Reset() ;
}

Int_t OTrResidualProfile::Fill(Axis_t x, Axis_t y, Stat_t w)
{
  if(GetYaxis()->GetXmin()< y && y <GetYaxis()->GetXmax() ) {  
    int bin = GetXaxis()->FindBin(x) ;
    fSumW[bin] += w;
    fSumY[bin] += y;
    fSumY2[bin] += y*y;
  }
  return TH2F::Fill(x,y,w) ;
}

void OTrResidualProfile::Reset()
{
  for(int bin=0; bin<=GetNbinsX()+1; bin++) {
    fSumW[bin]=0; fSumY[bin]=0; fSumY2[bin]=0;
  }
  TH2F::Reset() ;
}

void
OTrResidualProfile::FitWidthsBare()
{
  int cut=0;
  for(int bin=1; bin<=GetNbinsX(); bin++) 
    if(fSumW[bin]>cut) {
      fWidth[bin-1]    = sqrt(fSumY2[bin]/fSumW[bin] - sqr(fSumY[bin]/fSumW[bin])) ;
      fWidthErr[bin-1] = fWidth[bin-1]/sqrt(fSumW[bin]) ;
      fFitFlag[bin-1] = 0;
    } else fFitFlag[bin-1] = 1;
}

Double_t DoubleGaus(double* x, double* par)
{
  const double mean = par[0] ;
  const double normA = par[1] ;
  const double widthA = par[2] ;
  const double normB = par[3] ;
  const double widthB = par[4] ;
  return normA*exp(-0.5*sqr((*x - mean)/widthA)) + normB*exp(-0.5*sqr((*x - mean)/widthB)) ;
}
Double_t AlternativeDoubleGaus(double* x, double* par)
{
  const double meanA = par[0] ;
  const double normA = par[1] ;
  const double varianceA = par[2] ;
  const double meanB = par[3] ;
  const double normB = par[4] ;
  const double varianceB = par[5] ;
  return normA*exp(-0.5*sqr(*x - meanA)/varianceA) + normB*exp(-0.5*sqr(*x - meanB)/varianceB) ;
}

Double_t GausOnBackGround(double* x, double* par)
{
  const double mean = par[0] ;
  const double norm = par[1] ;
  const double width = par[2] ;
  const double offset = par[3] ;
  return offset + norm*exp(-0.5*sqr((*x - mean)/width))  ;
}

Double_t AlternativeGausOnBackGround(double* x, double* par)
{
  const double mean     = par[0] ;
  const double norm     = par[1] ;
  const double variance = par[2] ;
  const double offset   = par[3] ;
  return offset + norm*exp(-0.5*sqr(*x - mean)/variance)  ;
}

Double_t AsymmetricGaus(double* x, double* par)
{
  const double mean     = par[0] ;
  const double norm     = par[1] ;
  const double varianceA = par[2] ;
  const double varianceB = par[3] ;
  double netx = *x-mean ;
  if(netx>0) return norm*exp(-0.5*sqr(netx)/varianceB)  ;
  return norm*exp(-0.5*sqr(netx)/varianceA)  ;
}

void
OTrResidualProfile::FitWidths()
{
  // Adapted from root ...

  Int_t nbins  = GetXaxis()->GetNbins();

  const int cut = 0 ;
  
  TF1* f1 = new TF1("f1",AlternativeDoubleGaus,-10,10,6) ;
  //TF1* f1 = new TF1("f1",AlternativeGausOnBackGround,-0.3,0.3,4) ;
  //TF1* f1 = new TF1("f1",AsymmetricGaus,-0.3,0.3,4) ;
  const char *fname = f1->GetName();
  Int_t npar = f1->GetNpar();
  
  static TCanvas* fitCanvas = 0;
  if(!fitCanvas) {
    fitCanvas = new TCanvas("fitCanvas") ;
    fitCanvas->Draw() ;
  }
  fitCanvas->cd() ;

  //Loop on all bins in X, generate a projection along Y
  Int_t nentries;
  for (int bin=1;bin<=nbins;bin++) {
    TH1D *hpy = ProjectionY("_temp",bin,bin,"e");
    if(hpy) {
      nentries = Int_t(hpy->GetEntries());
      if( nentries > cut ) {
	float width = hpy->GetRMS() ;
 	float mean  = hpy->GetMean() ;
	cout << mean << " " << width << " " << endl ; 
 	float sum(0) ;
	float sumY(0) ;
	float sumY2(0) ;
 	for(int biny=1; biny<=hpy->GetNbinsX(); biny++) {
	  sum+= hpy->GetBinContent(biny) ;
	  sumY+= hpy->GetBinContent(biny)*hpy->GetBinCenter(biny) ;
	  sumY2+= hpy->GetBinContent(biny)*sqr(hpy->GetBinCenter(biny)) ;
	}
	float norm = sum*hpy->GetBinWidth(1)/(sqrt(2*M_PI)*width) ;
	
	fWidth[bin-1] = width ;
	fWidthErr[bin-1] = width/sqrt(sum) ;
	fFitFlag[bin-1] = 0 ;
		  
// 	// f1->SetParameter(0,mean) ;
// 	// 	f1->SetParameter(1,norm) ;
// 	// 	//f1->SetParameter(2,0.8*width) ;
// 	// 	f1->SetParameter(2,sqr(0.8*width)) ;
// 	// 	f1->SetParLimits(2,0,5*sqr(width)) ;
// 	// 	f1->SetParameter(3,0) ;
// 	// 	f1->SetParLimits(3,0,norm) ;
// 	// 	f1->SetParameter(0,mean) ;
// 	// 	f1->SetParameter(1,norm) ;
// 	// 	f1->SetParameter(2,sqr(width)) ;
// 	// 	f1->SetParameter(3,sqr(width)) ;
// 	f1->SetParameter(0,mean) ;
// 	f1->SetParameter(1,norm) ;
// 	f1->SetParameter(2,sqr(0.5*width)) ;
// 	f1->SetParameter(3,mean) ;
// 	f1->SetParameter(4,0.1*norm) ;
// 	f1->SetParameter(5,sqr(3*width)) ;

// 	//hpy->Fit(fname,"QNRL");
// 	hpy->Fit(fname,"ILQ");
	
// 	Int_t npfits = f1->GetNumberFitPoints();
// 	//cout << "Fitting ...: " << npfits << " " <<  npar << " " << cut << endl ;
// 	if (npfits > npar && npfits >= cut) {
// 	  //fWidth[bin-1] = f1->GetParameter(2) ;
// 	  //fWidthErr[bin-1] = f1->GetParError(2) ;
// 	  //fWidth[bin-1] = sqrt(f1->GetParameter(2)) ;
// 	  //fWidthErr[bin-1] = f1->GetParError(2)/(2*fWidth[bin-1]) ;
// 	  // 	  fWidth[bin-1] = (sqrt(f1->GetParameter(2))+ sqrt(f1->GetParameter(3)))/2 ;
// 	  // 	  fWidthErr[bin-1] = sqrt(sqr(f1->GetParError(2))/f1->GetParameter(2) +
// 	  // 				  sqr(f1->GetParError(3))/f1->GetParameter(3))/4 ;
// 	  if(f1->GetParameter(4)<f1->GetParameter(1)) {
// 	    fWidth[bin-1] = sqrt(f1->GetParameter(2)) ;
// 	    fWidthErr[bin-1] = f1->GetParError(2)/(2*fWidth[bin-1]) ;
// 	  } else {
// 	    fWidth[bin-1] = sqrt(f1->GetParameter(5)) ;
// 	    fWidthErr[bin-1] = f1->GetParError(5)/(2*fWidth[bin-1]) ;
// 	  }
// 	  cout << bin << " " << hpy->GetEntries() << " " << mean << " " << width << " "
// 	       << fWidth[bin-1] << endl ;
// 	  fFitFlag[bin-1] = 0;
// 	} else fFitFlag[bin-1] = 4;
// 	hpy->Draw() ;
// 	fitCanvas->Update() ;
// 	getchar() ;
	
      } else fFitFlag[bin-1] = 3;
      delete hpy;
    } else fFitFlag[bin-1] = 2;
  }
  delete f1 ;
}

#endif //USE_ROOT
