#include <iomanip.h>
#include <algorithm>
#include "wax/OTrSector.hh"
#include "wax/OTrLayer.hh"
#include "wax/OTrEvent.hh"
#include "wax/OTrChannelHistogrammer.hh"
#include "wax/OTrTdcHistogram.hh"

#ifdef USE_ROOT
 #include <TH1.h>
 #include <TPad.h>
 #include <TText.h>
 #include <TAxis.h>
#endif

//__________________________________________________________________
// OTrSector
//
// Container for:
// * OTrGedes
// * OTrWires
// Functionality:
//
// Author: Wouter Hulsbergen && Ivar Siccama
//__________________________________________________________________

#ifdef __OTR_CLASSDEF__
ClassImp(OTrSector) ;
//void OTrSector::Streamer(TBuffer& b) {} ;
//void OTrSector::ShowMembers(TMemberInspector &, char *) {} ;
#endif

float OTrSector::fgMaximumOccupancy = 0.10 ;

inline bool CompareSectionXPosition(const OTrSection* lhs, const OTrSection* rhs)
{
  return lhs->GetRawX(0) < rhs->GetRawX(0) ;
}

OTrSector::OTrSector() : fLayer(0)
#ifdef USE_ROOT
  , fPlot(0) 
#endif
{
  cout << "** OTrSector::OTrSector(): do not call !!" << endl ;
}


OTrSector::OTrSector( OTrLayer* lay, ComponentId_t id ) 
  : OTrDetectorComponent(id), fLayer(lay)
#ifdef USE_ROOT
  , fPlot(0)
#endif
{
  //cout << "OTrSector::OTrSector(...) : " << GetSectorId() << endl ;
  for(OTrLayer::SectionContainer::const_iterator sectionIt=fLayer->GetSections().begin() ;
      sectionIt!=fLayer->GetSections().end() ; sectionIt++) {
    //cout << (*sectionIt).GetSectionId() << " " << (*sectionIt).GetSectorId() << endl ;
    if((*sectionIt).GetSectorId()==GetSectorId()) 
      fSections.push_back(&(*sectionIt)) ;
  }
  sort(fSections.begin(), fSections.end(), CompareSectionXPosition) ;
  //cout << "# sections: " << fSections.size() << endl ;
} 

OTrSector::~OTrSector() {
#ifdef USE_ROOT
  if(fPlot) delete fPlot ;
#endif
}

int OTrSector::GetNWires() const
{
  int n=0;
  for(SectionContainer::const_iterator sectionIt=GetSections().begin() ;
      sectionIt!=GetSections().end() ; sectionIt++) n+= (*sectionIt)->GetWires().size() ;
  return n;
}

// void 
// OTrSector::Translate(const ThreeVector& vec) 
// {
//   for(SectionContainer::iterator secIt=GetSections().begin();
//       secIt!=GetSections().end() ; secIt++) (*secIt)->Translate(vec) ;
// }

HepPoint3D
OTrSector::GetPMin() const
{
  Point3D thepmin(0,0,0) ;
  SectionContainer::const_iterator secIt=GetSections().begin() ;
  if(secIt!=GetSections().end() ) {
    thepmin = (*secIt)->GetPMin() ;
    secIt++ ;
    for(;secIt!=GetSections().end(); secIt++) {
      Point3D pmin((*secIt)->GetPMin()) ;
      if(thepmin.x() > pmin.x()) thepmin.setX(pmin.x()) ;
      if(thepmin.y() > pmin.y()) thepmin.setY(pmin.y()) ;
      if(thepmin.z() > pmin.z()) thepmin.setZ(pmin.z()) ;
    }
  }
  return thepmin ;
}

HepPoint3D
OTrSector::GetPMax() const
{
  Point3D thepmax(0,0,0) ;
  SectionContainer::const_iterator secIt=GetSections().begin() ;
  if(secIt!=GetSections().end() ) {
    thepmax = (*secIt)->GetPMax() ;
    secIt++ ;
    for(;secIt!=GetSections().end(); secIt++) {
      Point3D pmax((*secIt)->GetPMax()) ;
      if(thepmax.x()<pmax.x()) thepmax.setX(pmax.x()) ;
      if(thepmax.y()<pmax.y()) thepmax.setY(pmax.y()) ;
      if(thepmax.z()<pmax.z()) thepmax.setZ(pmax.z()) ;
    }
  }
  return thepmax ;
}


#ifdef USE_ROOT

float
OTrSector::GetOccupancy( bool useMask ) const
{
  float occSum = 0;
  int   nSum  = 0;
  for(SectionContainer::const_iterator sectionIt = GetSections().begin() ;
      sectionIt != GetSections().end() ; sectionIt++)
    for( vector<OTrWire>::const_iterator wireIt = (*sectionIt)->GetWires().begin() ;
	 wireIt != (*sectionIt)->GetWires().end() ; wireIt++) {
      const OTrChannel* chan = wireIt->GetChannel() ;
      if (chan && (!useMask || chan->GoodMask()) && chan->GetOccupancy()>0 ) {
	occSum += chan->GetOccupancy();
	nSum++; 
      }
    } 
  float occup = 0;
  if ( nSum > 0 ) {
    occup = occSum/float(nSum);
  }
  return occup ;
}

float
OTrSector::OccupancySummary( ostream& ofs, bool useMask )
{
  const int nWires = GetNWires() ;
  float occup = 0;
  int   ntot  = 0;
  ofs << "###############" << GetSectorId() 
      << "###############" << endl;
  for(SectionContainer::const_iterator sectionIt = GetSections().begin() ;
      sectionIt != GetSections().end() ; sectionIt++)
    for( vector<OTrWire>::const_iterator wireIt = (*sectionIt)->GetWires().begin() ;
	 wireIt != (*sectionIt)->GetWires().end() ; wireIt++) {
      const OTrChannel* chan = wireIt->GetChannel() ;
      float occ;
      if (!useMask || chan->GoodMask()) {
	occ = chan->GetOccupancy();
	ntot++;
	occup += occ ;
      } else {
	occ = -1 ;
      }
      //ofs << setw(3) << wireIt->GetWireNumber() << "  " 
      //	  << setw(6) << setprecision(4) << occ; 
      //nprinted++;
      //if ( nprinted%8 == 0 ) { ofs << endl; nprinted = 0; }
      //if ( occ < 0.001 || occ > 0.30 ) continue;
      //cout << wireIt->GetWireNumber() << ": " << chan->occupancy() << endl;
    } 
  
  if ( ntot > 0 ) {
    occup = occup/float(ntot);
  } else {
    occup = 0;
  }  
  if ( GetLayer()->GetEStereoAngle() == 0 ) ofs << "-5 degrees: ";
  if ( GetLayer()->GetEStereoAngle() == 1 ) ofs << " 0 degrees: ";
  if ( GetLayer()->GetEStereoAngle() == 2 ) ofs << "+5 degrees: "; 
  ofs << GetSectorId() << " ntot, occup = " << ntot << " " 
      << setprecision(2) << occup << endl;
  const OTrWire* firstwire = &(GetSections().front()->GetWires().front()) ;
  const OTrWire* lastwire  = &(GetSections().back()->GetWires().back()) ;

  if ( firstwire == 0 || lastwire == 0 ) {
    ofs << "*** ERROR wire not found" << endl;
  } else {
    ofs << "position of 1st wire: " << setprecision(4) 
	<< firstwire->GetCentralX() << " " 
	<< firstwire->GetCentralY() << " " 
	<< firstwire->GetRawZ() << endl 
	<< "position of " << nWires << "th wire: "
	<< lastwire->GetCentralX() << " " 
	<< lastwire->GetCentralY() << " " 
	<< lastwire->GetRawZ() << endl ;
  }
  return occup;
}

TString OTrSector::GetName() const
{
  char name[16];
  sprintf(name,"Sector_%.3d%.3d%.3d",GetSuperLayerId(),GetLayerId(),GetSectorId()) ;
  return TString(name) ;
}

TH1*
OTrSector::GetChannelH1(bool useMask, float (OTrChannelHistogrammer::* channelFunc)() const ) const
{
  const int nWires = GetNWires() ;
  TString h1Name(GetName());
  TH1* histo = new TH1F(h1Name, "", nWires, 0, nWires );
  int bin=1;
  for( SectionContainer::const_iterator sectionIt = fSections.begin() ;
       sectionIt!=fSections.end() ; sectionIt++) {
    //(*sectionIt)->DumpBadChannels(baddies);
    for( OTrSection::WireContainer::const_iterator wireIt=(*sectionIt)->GetWires().begin() ;
	 wireIt!=(*sectionIt)->GetWires().end() ; wireIt++, bin++) {
      const OTrChannel* chan = wireIt->GetChannel() ;
      float x =0 ;
      if(chan && (!useMask || chan->GoodMask())) {
	const OTrChannelHistogrammer* chanhis = chan->GetHistogrammer() ;
	x = ((*chanhis).*channelFunc)() ;
      }
      histo->SetBinContent(bin,x) ;
    }
  }
  return histo ;
}

TH1*
OTrSector::GetOccupancyVariationH1(bool useMask) const
{
  const int nWires = GetNWires() ;
  TString h1Name(GetName());
  TH1* histo = new TH1F(h1Name, "", nWires, 0, nWires );
  int bin=1;
  for( SectionContainer::const_iterator sectionIt = fSections.begin() ;
       sectionIt!=fSections.end() ; sectionIt++) {
    for( OTrSection::WireContainer::const_iterator wireIt = 
	   (*sectionIt)->GetWires().begin() ;
	 wireIt!=(*sectionIt)->GetWires().end() ; wireIt++, bin++) {
      const OTrChannel* chan = wireIt->GetChannel() ;
      float diff = 0 ;
      if(chan && (!useMask || chan->GoodMask())) {
	OTrSection::WireContainer::const_iterator nextWireIt = wireIt + 1;
        if ( nextWireIt != (*sectionIt)->GetWires().end() ) {
	  const OTrChannel* nextChan = nextWireIt->GetChannel();
	  if(nextChan && (!useMask || nextChan->GoodMask())) {  
	    diff = (*nextChan).GetOccupancy() - (*chan).GetOccupancy() ;
	  }
	}
      }
      histo->SetBinContent(bin,diff) ;
    }
  }
  return histo ;
}

TH1*
OTrSector::GetTZeroH1(bool useMask) const
{
  const int nWires = GetNWires() ;
  TString h1Name(GetName());
  h1Name.Append("_TZero") ;
  TH1* histo = new TH1F(h1Name, "", nWires, 0, nWires );
  int bin=1;
  for( SectionContainer::const_iterator sectionIt = fSections.begin() ;
       sectionIt!=fSections.end() ; sectionIt++)
    for( OTrSection::WireContainer::const_iterator wireIt=(*sectionIt)->GetWires().begin() ;
	 wireIt!=(*sectionIt)->GetWires().end() ; wireIt++, bin++) {
    const OTrChannel* chan = wireIt->GetChannel() ;
    if(chan && (!useMask || chan->GoodMask())) {
      histo->SetBinContent(bin,chan->GetTZero().GetValue()) ;
      histo->SetBinError(bin,chan->GetTZero().GetError()) ;
    } 
  }
  return histo ;
}
  
TH1*
OTrSector::GetTimeH1() const 
{
  TString timeH1Name(GetName());
  timeH1Name.Append("_TimeH1") ;
  TH1* timeH1 = new TH1D(timeH1Name,timeH1Name,OTrChannel::GetNTimeBins(),
			 OTrChannel::GetMinimumTimeAxis(),
			 OTrChannel::GetMaximumTimeAxis()) ;
  for( SectionContainer::const_iterator sectionIt = fSections.begin() ;
       sectionIt!=fSections.end() ; sectionIt++)
    for( OTrSection::WireContainer::const_iterator wireIt=(*sectionIt)->GetWires().begin() ;
	 wireIt!=(*sectionIt)->GetWires().end() ; wireIt++) {
      const OTrChannel* chan = wireIt->GetChannel() ;
      if(chan) {
	TH1D* histo = chan->GetTimeH1() ;
	timeH1->Add(histo,1) ;
	delete histo ;
      }
    }
  return timeH1;
} 

TH1*
OTrSector::GetAvTimeH1( bool useHistogram, bool useMask )
{
  const int nWires = GetNWires() ;
  TString h1Name(GetName());
  h1Name.Append("_AvTime") ;
  TH1* histo = new TH1F( h1Name, "", nWires, 0, nWires );

  int bin=1;
  for( SectionContainer::const_iterator sectionIt = fSections.begin() ;
       sectionIt!=fSections.end() ; sectionIt++)
    for( OTrSection::WireContainer::const_iterator wireIt=(*sectionIt)->GetWires().begin() ;
	 wireIt!=(*sectionIt)->GetWires().end() ; wireIt++, bin++) {
      const OTrChannel* chan = wireIt->GetChannel() ;
    float average = 0;
    if(chan && (!useMask || chan->GoodMask())) {
      if(useHistogram) average = chan->GetAverageTime() ;
      else average = chan->GetAverageTime() ;
    }
    histo->SetBinContent(bin, average );
  } 
  return histo ;
}

TH1*
OTrSector::GetSigmaTimeH1( bool useHistogram, bool useMask )
{
  const int nWires = GetNWires() ;
  TString h1Name(GetName());
  h1Name.Append("_SigmaTime") ;
  TH1* histo = new TH1F( h1Name, "", nWires, 0, nWires );

  int bin=1;
  for( SectionContainer::const_iterator sectionIt = fSections.begin() ;
       sectionIt!=fSections.end() ; sectionIt++)
    for( OTrSection::WireContainer::const_iterator wireIt=(*sectionIt)->GetWires().begin() ;
	 wireIt!=(*sectionIt)->GetWires().end() ; wireIt++, bin++) {
      const OTrChannel* chan = wireIt->GetChannel() ;
      float sigma = 0;
      if(chan && (!useMask || chan->GoodMask())) {
	if(useHistogram) {
	  TH1D* tdch1 = chan->GetTimeH1() ;
	  sigma = tdch1->GetRMS();
	  delete tdch1 ;
	} else sigma = chan->GetSigma() ;
      }
      histo->SetBinContent( bin, sigma ) ; 
    } 
  return histo ;
}

// TH1*
// OTrSector::GetOccupancyH1( bool useMask ) const
// {
//   TH1* histo = GetChannelH1(useMask, &(OTrChannel::GetOccupancy) ) ;
//   TString name(histo->GetName()) ;
//   name.Append("_Occupancy") ;
//   histo->SetName(name) ;
//   return histo;
// }

int OTrSector::Plot(TPad* p, const int choice, char* option, const int cl,
	  	     const OTrSuperLayerEvent* slEvent, const bool useMask )
{

  p->cd();
  // Delete existing plot
  if ( fPlot != 0 ) {
    delete fPlot;
    fPlot = 0;
  } 
  const int nWires = GetNWires() ;
  TText* ttext = 0;
  switch (choice)
    {
    case 1: // occupancy for each channel 
      fPlot = GetChannelH1(useMask, &(OTrChannelHistogrammer::GetOccupancy) ) ;
      fPlot->SetName((TString(fPlot->GetName())).Append("_Occupancy")) ;
      //fPlot = GetOccupancyH1( useMask ) ;
      //fPlot->SetMaximum(1.2);
      fPlot->SetMaximum(fgMaximumOccupancy);
      //fPlot->SetMaximum(0.15);
      fPlot->SetFillColor(10);
      char text[256]; sprintf(text,"Av: %8.2e ",GetOccupancy()) ;
      float xpos,ypos ;
      xpos = fPlot->GetXaxis()->GetXmin()
	+0.1*(fPlot->GetXaxis()->GetXmax()-fPlot->GetXaxis()->GetXmin()) ;
      ypos = (0.55+0.15*int(GetSections().front()->GetEStereoAngle()))*fgMaximumOccupancy ;
      //cout << xpos << " " << ypos << " " << fPlot->GetYaxis()->GetXmax() << " " << endl ;
      ttext = new TText(xpos,ypos,text) ;
      //ttext->SetTextSize(0.05*(fPlot->GetXaxis()->GetXmax()-fPlot->GetXaxis()->GetXmin())/32) ;
      ttext->SetTextSize(0.12) ;
      break ;
      
    case 2: // tdc count for each hit channel
      char name[30];
      sprintf( name, "%s%d", "hit_channels_", GetId() );
      fPlot = new TH1D( name, "", nWires, 0, nWires );
      if ( slEvent ) {
	const OTrSuperLayerEvent::HitContainer& hits = slEvent->GetHits();
	//cout << "OTrSector::plot(), hits.size() = " << hits.size() << endl;
	for( OTrSuperLayerEvent::HitContainer::const_iterator hitp = hits.begin();
	     hitp != hits.end(); hitp++) {
	  const OTrWire* wire = (*hitp)->GetWire();
	  if(wire->GetSectorId() == GetSectorId() ) {
	    OTrChannel* chan = wire->GetChannel() ;
	    if ( useMask && !chan->GoodMask()  ) continue; 
	    fPlot->Fill( wire->GetWireNumber()+1, (*hitp)->GetTdcCounts() );
	  }
	}
      }
      fPlot->SetMaximum(260);
      break ;
  
    case 3: // average tdc value for each channel
      fPlot = GetAvTimeH1(false, useMask) ;
      //fPlot->GetYaxis()->SetLimits(120,200);
      fPlot->SetMinimum(120);
      fPlot->SetMaximum(210);
      break ;
  
    case 4: // tdc spectrum for this sector
      TH1D* tdchisto ; tdchisto = (TH1D*)GetTimeH1() ; 
      fPlot = new OTrTdcHistogram( *tdchisto );
      delete tdchisto ;
      break ;
      
    case 5: // RMS tdc count for each channel
      fPlot = GetSigmaTimeH1(false, useMask ) ;
      //fPlot->SetMinimum(100);
      //fPlot->SetMaximum(200);
      break ;
      
    case 6: 
      fPlot = GetTZeroH1( useMask ) ;
      fPlot->SetMinimum(-30);
      fPlot->SetMaximum(30);
      break ;

    case 7:
      fPlot = GetOccupancyVariationH1( useMask ) ;
      fPlot->SetMinimum(-0.01);
      fPlot->SetMaximum(0.01);
      break ;

    default:
      fPlot = new TH1F("EMPTY","NO SUCH CHOICE!",1,0,1) ;
      break ;
    }
  
  if ( fPlot ) {
    fPlot->SetLabelSize(0.05);
    fPlot->SetFillColor(0);
    fPlot->SetMarkerStyle(21);
    fPlot->SetMarkerSize(0.4);
    fPlot->SetMarkerColor(cl); 
    fPlot->SetLineColor(cl);
  }

  // fit an asymmtric Gaussian to the TDC spectrum
  if ( choice==4 ) {
    OTrTdcHistogram* tdcHisto = (OTrTdcHistogram*)fPlot ;
    //OTrTdcHistogram SectorTdcHisto( *fPlot );
    OTrFObservable width;
    tdcHisto->FitAsymmetricGaussian( width );

    if ( width.GetValue() > 0 ) {
      char text[256]; 
      int val = int(width.GetValue());
      int eval = int(width.GetError());
      sprintf(text,"width: %d %s %d ", val," +/- ",eval);
      float xpos,ypos ;
      float xrange = tdcHisto->GetXaxis()->GetXmax() -
	tdcHisto->GetXaxis()->GetXmin() ;
      xpos  = tdcHisto->GetXaxis()->GetXmin() + 0.1*xrange; 
      ypos = 0.55+0.15*int(GetSections().front()->GetEStereoAngle());
      ypos = ypos*tdcHisto->GetMaximum();
      //fgMaximumOccupancy ;
      cout << text << " at " 
	   << xpos << " " << ypos << " " << tdcHisto->GetYaxis()->GetXmax()
	   << " " << endl ;
      ttext = new TText(xpos,ypos,text) ;
      //ttext->SetTextSize(0.05*(tdcHisto->GetXaxis()->GetXmax()-tdcHisto-
      //    >GetXaxis()->GetXmin())/32) ;
      ttext->SetTextSize(0.08) ;
    }
    cout << "** maximum = " << tdcHisto->GetYaxis()->GetXmax() << endl;
  }

  // histogram drawing option
  char opt[20];
  sprintf(opt,"%s%s",option,"");
  fPlot->Draw(opt) ;
  if(ttext) {
    ttext->SetTextColor(cl) ;
    ttext->Draw() ;
  }
  return 0;
}
#endif // USE_ROOT

ostream & operator << (ostream & os, const OTrSector& s )
{
  return os << "sector: " << s.GetId() << " " << s.GetSectorId();
}
