#include <algorithm>
#include "wax/OTrSuperCluster.hh"
#include "wax/OTrTrackDescription.hh"
#include "wax/OTrReconstructedHit.hh"
#include "wax/OTrParameterizedRTRelation.hh"

#ifdef USE_ROOT
 #include <TPolyMarker3D.h>
 #include <TProfile.h>
#endif

//__________________________________________________________________
// OTrSuperCluster
//
// Container for:
// * 3 OTrSectionClusters (but not owner)
// Functionality:
// * Reconstruction of one spacepoint
//
// Author: Wouter Hulsbergen 
//__________________________________________________________________

//#ifdef __OTR_CLASSDEF__
// ClassImp(OTrSuperCluster) ;
//#endif

float OTrSuperCluster::fgMinChi2Probability = 0.001 ;

OTrSuperCluster::OTrSuperCluster(OTrSectionCluster* plusIt, OTrSectionCluster* minusIt) 
  : OTrTrackDescription2D(), fXDif(99), fClosestPointMarker(0)
{
  float sindif 
    //sin(plusIt->GetStereoAngle()-minusIt->GetStereoAngle()) ;
    = plusIt->GetSinStereoAngle()*minusIt->GetCosStereoAngle() -
    plusIt->GetCosStereoAngle()*minusIt->GetSinStereoAngle() ;
  float tx = (plusIt->GetSinStereoAngle()*minusIt->GetRawTx() - 
	      minusIt->GetSinStereoAngle()*plusIt->GetRawTx())/sindif ;
  float ty = (plusIt->GetCosStereoAngle()*minusIt->GetRawTx() - 
	      minusIt->GetCosStereoAngle()*plusIt->GetRawTx())/sindif ;
  SetTx(tx) ; SetTy(ty) ;
  fPlusClusters.push_back(plusIt) ;
  fMinusClusters.push_back(minusIt) ;
};

OTrSuperCluster::OTrSuperCluster(const OTrSuperCluster& rhs) 
  : OTrReconstructedTrackInterface(rhs),
    OTrTrackDescription2D(rhs)
{
  fXDif = rhs.fXDif ;
  fPlusClusters = rhs.fPlusClusters;
  fMinusClusters = rhs.fMinusClusters;
  fZeroClusters = rhs.fZeroClusters;
  fSectionClusters = rhs.fSectionClusters;
  fClosestPointMarker=0 ;
#ifdef USE_ROOT
  if(rhs.fClosestPointMarker) 
    fClosestPointMarker = new TPolyMarker3D(*rhs.fClosestPointMarker) ;
#endif
} ;

OTrSuperCluster::~OTrSuperCluster()
{
#ifdef USE_ROOT
  if(fClosestPointMarker) {
    TCollection::StartGarbageCollection();
    TCollection::GarbageCollect(fClosestPointMarker) ;
    delete fClosestPointMarker ;
  };
#endif
};

bool
OTrSuperCluster::AddZeroIfValid(OTrSectionCluster* zeroClus) {
  bool added(false) ;
  float xdif(0) ;
  if(zeroClus->WithinYAcceptance(GetTx(),GetTy()) &&
     fabs(xdif = zeroClus->GetXDif(GetTx())) < 
     (fPlusClusters.front()->GetGede()->GetWirePitch()/
      fPlusClusters.front()->GetCosStereoAngle() +
      zeroClus->GetGede()->GetWirePitch())/2) {
    //      (fPlusClusters.front()->GetGede()->GetWirePitch()/
    //       cos(fPlusClusters.front()->GetStereoAngle())+
    //       zeroClus->GetGede()->GetWirePitch())) {
    //zeroClus->GetGede()->GetWirePitch()) {
    bool foundGede(false) ;
    for(ClusterContainer::iterator zeroIt=fZeroClusters.begin() ;
	zeroIt!=fZeroClusters.end(); zeroIt++)
      if((*zeroIt)->GetGede()==zeroClus->GetGede()) {
	foundGede=true ;
	if(fabs(xdif)< fabs((*zeroIt)->GetXDif(GetTx()))) {
	  *zeroIt = zeroClus ;
	  if(fabs(xdif)<fabs(fXDif)) fXDif = xdif ;
	  added = true ;
	};
	break ;
      };
    if(!foundGede) {
      fZeroClusters.push_back(zeroClus) ;
      if(fabs(xdif)<fabs(fXDif)) fXDif = xdif ;
      added=true ;
    }; 
  };
#ifdef USE_ROOT
  static TH1F* xdifH1 = new TH1F("xdifB","",500,-3,3) ;
  if(zeroClus->WithinYAcceptance(GetTx(),GetTy()))
     xdifH1->Fill(float(zeroClus->GetXDif(GetTx())/
			zeroClus->GetGede()->GetWirePitch())) ;
#endif //USE_ROOT
  return added ;
};

inline bool
OTrSuperCluster::AddZero(OTrSectionCluster* zeroClus) {
  bool foundGede(false) ;
  for(ClusterContainer::iterator zeroIt=fZeroClusters.begin() ;
      zeroIt!=fZeroClusters.end(); zeroIt++)
    if((*zeroIt)->GetGede()==zeroClus->GetGede()) {
      foundGede=true ;
      break ;
    };
  if(!foundGede) fZeroClusters.push_back(zeroClus) ;
  return !foundGede ;
};

inline bool
OTrSuperCluster::AddMinus(OTrSectionCluster* minusClus) {
  bool foundGede(false) ;
  for(ClusterContainer::iterator minusIt=fMinusClusters.begin() ;
      minusIt!=fMinusClusters.end(); minusIt++)
    if((*minusIt)->GetGede()==minusClus->GetGede()) {
      foundGede=true ;
      break ;
    };
  if(!foundGede) fMinusClusters.push_back(minusClus) ; 
  return !foundGede ;
};

inline bool
OTrSuperCluster::AddPlus(OTrSectionCluster* plusClus) {
  bool foundGede(false) ;
  for(ClusterContainer::iterator plusIt=fPlusClusters.begin() ;
      plusIt!=fPlusClusters.end(); plusIt++)
    if((*plusIt)->GetGede()==plusClus->GetGede()) {
      foundGede=true ;
      break ;
    };
  if(!foundGede) fPlusClusters.push_back(plusClus) ; 
  return !foundGede ;
};

inline float
OTrSuperCluster::CalcXDif() {
  if(fZeroClusters.size()) {
    ClusterContainer::iterator zeroIt=fZeroClusters.begin() ;
    float newxdif = GetXDif(*zeroIt) ; zeroIt++;
    for( ; zeroIt!=fZeroClusters.end(); zeroIt++) {
      float xdif = GetXDif(*zeroIt) ;
      if(fabs(xdif)<fabs(newxdif)) newxdif = xdif ;
    };
    //   if(fabs(newxdif) < fabs(fXDif))
    //       cout << "%^(%(^%^%^*%*%^%&*%^&*" << newxdif << " " << fXDif << endl ;
    fXDif = newxdif ;
  };
  return fXDif ;
} ;

bool
OTrSuperCluster::Merge(OTrSuperCluster& rhs)
{

  // The idea is this: If clusters overlap in more than one
  // stereolayer, only one of them can be real or they form one
  // cluster. This is of course not completely true, but in any case
  // almost completely.

  // Merge clusters that overlap in zero && plus or in zero && minus
  
  // first remove on rhs all clusters which also appear on lhs
  int nZero(0), nPlus(0), nMinus(0) ;
  for(ClusterContainer::const_iterator lhsPlusIt=fPlusClusters.begin() ;
      lhsPlusIt!=fPlusClusters.end(); lhsPlusIt++)
    for(ClusterContainer::iterator rhsPlusIt=rhs.fPlusClusters.begin() ;
	rhsPlusIt!=rhs.fPlusClusters.end(); rhsPlusIt++)
      if(*lhsPlusIt==*rhsPlusIt) {
	nPlus++;
	//rhs.fPlusClusters.erase(rhsPlusIt--) ;
      };
  for(ClusterContainer::const_iterator lhsMinusIt=fMinusClusters.begin() ;
      lhsMinusIt!=fMinusClusters.end(); lhsMinusIt++)
    for(ClusterContainer::iterator rhsMinusIt=rhs.fMinusClusters.begin() ;
	rhsMinusIt!=rhs.fMinusClusters.end(); rhsMinusIt++)
      if(*lhsMinusIt==*rhsMinusIt) {
	nMinus++;
	//rhs.fMinusClusters.erase(rhsMinusIt--) ;
      };
  for(ClusterContainer::iterator lhsZeroIt=fZeroClusters.begin() ;
      lhsZeroIt!=fZeroClusters.end(); lhsZeroIt++)
    for(ClusterContainer::iterator rhsZeroIt=rhs.fZeroClusters.begin() ;
	rhsZeroIt!=rhs.fZeroClusters.end(); rhsZeroIt++)
      if(*lhsZeroIt==*rhsZeroIt) {
	nZero++;
	//rhs.fZeroClusters.erase(rhsZeroIt--) ;
      };

// 	if(nPlus||nMinus||fabs(GetXDif(*lhsZeroIt))<fabs(rhs.GetXDif(*lhsZeroIt))) {
// 	  nZero++;
// 	  rhs.fZeroClusters.erase(rhsZeroIt--) ;
// 	} else {
// // 	  cout << "Should not happen too often: " << *rhsZeroIt << " "
// // 		<< fZeroClusters.size() << " "
// // 		<< GetXDif(*rhsZeroIt) << " " << rhs.GetXDif(*rhsZeroIt) << " "
// // 		<< CalcXDif() << " " << rhs.CalcXDif() << endl 
// // 		<< GetTx() << " " << GetTy() << " "
// // 		<< rhs.GetTx() << " " << rhs.GetTy() << endl ;
// 	   fZeroClusters.erase(lhsZeroIt--) ; 
// 	  break ;
// 	};
// // 	nZero++;
// // 	rhs.fZeroClusters.erase(rhsZeroIt--) ;
//       }

  // If there is an overlap in more than one stereolayer, the two
  // clusters are merged.
  bool merged = (nZero&&(nMinus||nPlus))||(nPlus && nMinus) ;
  if(merged) {
    //if(nPlus && nMinus) cout << " Overlap in minus and plus " << endl ;
    //if(nZero&&nPlus)  cout << " Overlap in zero and plus " << endl ;
    //if(nZero&&nMinus) cout << " Overlap in zero and minus " << endl ;
    for(ClusterContainer::iterator rhsMinusIt=rhs.fMinusClusters.begin() ;
  	rhsMinusIt!=rhs.fMinusClusters.end(); rhsMinusIt++) {
      //if(AddMinus(*rhsMinusIt)) rhs.fMinusClusters.erase(rhsMinusIt--) ;
      AddMinus(*rhsMinusIt) ;
      //rhs.fMinusClusters.erase(rhsMinusIt--) ;
    };
    for(ClusterContainer::iterator rhsZeroIt=rhs.fZeroClusters.begin() ;
  	rhsZeroIt!=rhs.fZeroClusters.end(); rhsZeroIt++) {
      //if(AddZero(*rhsZeroIt)) rhs.fZeroClusters.erase(rhsZeroIt--) ;
      AddZero(*rhsZeroIt) ;
      //rhs.fZeroClusters.erase(rhsZeroIt--) ;
    };
    for(ClusterContainer::iterator rhsPlusIt=rhs.fPlusClusters.begin() ;
	rhsPlusIt!=rhs.fPlusClusters.end(); rhsPlusIt++) {
      //if(AddPlus(*rhsPlusIt)) rhs.fPlusClusters.erase(rhsPlusIt--) ;
      AddPlus(*rhsPlusIt) ;
      //rhs.fPlusClusters.erase(rhsPlusIt--) ;
    };      
  };
  return merged ;
};

 //   if(nZero) rhs.CalcXDif() ;
//   if(!Valid()) {
//     cout << "?????????????????" 
// 	 << GetXDif() << " " << rhs.GetXDif() << endl ;
//     cout << nZero << " " <<  nPlus << " " <<  nMinus << endl ;
//     CalcXDif(); rhs.CalcXDif() ;
//     cout << GetXDif() << " " << rhs.GetXDif() << endl ;
//     cout << "lhs" << endl ;
//     for(ClusterContainer::iterator lhsZeroIt=fZeroClusters.begin() ;
//       lhsZeroIt!=fZeroClusters.end(); lhsZeroIt++)
//       cout << *lhsZeroIt<< " "
// 	   << GetXDif(*lhsZeroIt) << " " << rhs.GetXDif(*lhsZeroIt) << endl ;
//     cout << "rhs" << endl ;
//     for(ClusterContainer::iterator rhsZeroIt=rhs.fZeroClusters.begin() ;
// 	rhsZeroIt!=rhs.fZeroClusters.end(); rhsZeroIt++)
//       cout << *rhsZeroIt<< " "
// 	   << GetXDif(*rhsZeroIt) << " " << rhs.GetXDif(*rhsZeroIt) << endl ;	
//   };

void
OTrSuperCluster::Connect() 
{
  fSectionClusters.erase(fSectionClusters.begin(),fSectionClusters.end()) ;
  for(ClusterContainer::const_iterator zeroIt=fZeroClusters.begin() ;
      zeroIt!=fZeroClusters.end(); zeroIt++) {
    (*zeroIt)->SetSuperCluster(this) ;
    fSectionClusters.push_back(*zeroIt) ;
  };
  for(ClusterContainer::const_iterator plusIt=fPlusClusters.begin() ;
      plusIt!=fPlusClusters.end(); plusIt++) {
    (*plusIt)->SetSuperCluster(this) ;
    fSectionClusters.push_back(*plusIt) ;
  };
  for(ClusterContainer::const_iterator minusIt=fMinusClusters.begin() ;
      minusIt!=fMinusClusters.end(); minusIt++) {
    (*minusIt)->SetSuperCluster(this) ;
    fSectionClusters.push_back(*minusIt) ;
  };
};

#ifdef USE_ROOT
void
OTrSuperCluster::PerformHistogramming() const
{
  for(ClusterContainer::const_iterator clusIt=fSectionClusters.begin() ;
      clusIt!=fSectionClusters.end() ; clusIt++)
    (*clusIt)->PerformHistogramming() ;
} ;
#endif   

void 
OTrSuperCluster::RawWireFit()
{
  HepSymMatrix d2Chi2dQ2(2,0) ;
  HepVector dChi2dQ(2,0) ;
  HepVector dResdQ(2,0) ;
  float res ;
  float weight ;
  for(ClusterContainer::const_iterator clusIt=fSectionClusters.begin() ;
      clusIt!=fSectionClusters.end() ; clusIt++) {
    weight = 12/sqr((*clusIt)->GetGede()->GetWirePitch()) ;
    res = - (*clusIt)->GetRawX() ;
    dResdQ[0] = (*clusIt)->GetRawZ()* (*clusIt)->GetCosStereoAngle() ;
    dResdQ[1] = (*clusIt)->GetRawZ()*-(*clusIt)->GetSinStereoAngle() ;
    dChi2dQ += 2*res*dResdQ ;
    d2Chi2dQ2[0][0] += 2*dResdQ[0]*dResdQ[0] ;
    d2Chi2dQ2[0][1] += 2*dResdQ[0]*dResdQ[1] ;
    d2Chi2dQ2[1][1] += 2*dResdQ[1]*dResdQ[1] ;
  }
  int ierr;
  HepSymMatrix covOver2 = d2Chi2dQ2.inverse(ierr) ;
  if(ierr) {
    cout << "Matrix inversion error in OTrSuperCluster::WireFit()" << endl ;
    return ;
  };
  SetQ(-covOver2*dChi2dQ) ;
  SetCov(2*covOver2) ;
};

enum FitMode {
  kNormal=0,
  kCompareChi2=1,
  kCheckChi2=3
};

extern "C" float prob_(const float&, const int&) ;

bool
OTrSuperCluster::OneFitIteration(int fitmode)
{
  HepDouble chi2 = 0;
  static HepVector dChi2dQ(GetDim(),0) ;
  static HepSymMatrix d2Chi2dQ2(GetDim(),0) ;
  for(int row=0; row<GetDim(); row++) {
    dChi2dQ[row]=0;
    for(int col=0; col<=row; col++)
      d2Chi2dQ2[row][col]=0;
  }
  OTrNodeProjection* nodeProjection;
  for(HitContainer::iterator hitIt=GetHits().begin() ; 
      hitIt!=GetHits().end(); hitIt++) {
    nodeProjection = (*hitIt)->GetNodeProjection() ;
    nodeProjection->Project(this) ;
    chi2      += nodeProjection->GetChi2() ;
    dChi2dQ   += nodeProjection->GetdChi2dQ() ;
    d2Chi2dQ2 += nodeProjection->Getd2Chi2dQ2() ;
  }
  int ierr;
  HepSymMatrix d2Chi2dQ2Inv = d2Chi2dQ2.inverse(ierr) ;
  if(ierr) {
    printf("Matrix inversion error in OTrSuperCluster::OneFitIteration()") ;
    return false ;
  }
  chi2 += -d2Chi2dQ2Inv.similarity(dChi2dQ)/2 ;// This is the expected change in chi2
  switch (fitmode)
    {
    case kCompareChi2: 
      if(chi2>GetChi2()) return false ;
      break;
    case kCheckChi2:
      if(!ValidChi2()) return false ;
      break;
    }
  SetQ(GetQ() - d2Chi2dQ2Inv*dChi2dQ ) ;
  SetCov(2*d2Chi2dQ2Inv) ;
  SetChi2(chi2) ;
  return ValidChi2() ;
}

extern "C" float prob_(const float&, const int&) ;

inline bool 
compareHitDistance(OTrReconstructedHit* lhs, OTrReconstructedHit* rhs) {
  float lhsr = lhs->GetNodeProjection()->GetRm()/lhs->GetMeasuredHit()->
    GetChannel()->GetRTRelation()->GetRMax() ;
  float rhsr = rhs->GetNodeProjection()->GetRm()/rhs->GetMeasuredHit()->
    GetChannel()->GetRTRelation()->GetRMax() ;
  return lhsr > rhsr ;
}

inline bool 
compareHitChi2(OTrReconstructedHit* lhs, OTrReconstructedHit* rhs) {
  float lhsChi2 = lhs->GetNodeProjection()->GetChi2() ;
  float rhsChi2 = rhs->GetNodeProjection()->GetChi2() ;
  return lhsChi2 > rhsChi2 ;
}

bool
OTrSuperCluster::CreateReconstructedHits()
{
  bool success(false) ;
  EraseHits();
  for(ClusterContainer::const_iterator clusIt=fSectionClusters.begin() ;
      clusIt!=fSectionClusters.end() ; clusIt++)
    for(OTrSectionCluster::HitIterator hitIt=(*clusIt)->begin(); 
	hitIt!=(*clusIt)->end() ; hitIt++) {
      OTrReconstructedHit::LRSign_t lrsign = (*clusIt)->GetLRSign(*hitIt) ;
      GetHits().push_back(new OTrReconstructedHit(*hitIt,lrsign)) ;
      //if(lrsign==0) nonFixedHits.push_back(&(fReconstructedHits.back())) ;
    }

  // Sort hits in measured distance and determine lrsign one-by-one.
  sort(GetHits().begin(),GetHits().end(),compareHitDistance) ;
  for(HitContainer::iterator hitIt=GetHits().begin() ;
      hitIt!=GetHits().end() ; hitIt++) 
    if(!(*hitIt)->GetLRSignIsFixed()) {
      (*hitIt)->SetLRSign(-1.0) ;
      OneFitIteration(kNormal) ; 
      (*hitIt)->FlipLRSign() ;
      if(!OneFitIteration(kCompareChi2)) (*hitIt)->FlipLRSign() ; // Fit worse, flip back
    }
  OneFitIteration(kNormal) ; 

  // This does not really make it better ...
  for(HitContainer::iterator hitIt=GetHits().begin() ;
      hitIt!=GetHits().end(); hitIt++) (*hitIt)->UpdateInternal() ;
  OneFitIteration(kNormal) ;
  
  // Flip all hits once to see if we can improve and get rid of really bad hits
  success = OneFitIteration(kNormal) ; // We need this second to set the residuals right !!!
  sort(GetHits().begin(), GetHits().end(), compareHitChi2) ;
  HitContainer badHits ;
  for(HitContainer::iterator hitIt=GetHits().begin() ; 
      hitIt!=GetHits().end(); hitIt++) {
    //cout << (*hitIt)->GetChi2() << " " ;
    (*hitIt)->FlipLRSign() ;
    success = OneFitIteration(kCompareChi2) ;
    bool changed=success;
    if(!ValidChi2() && ValidNHits(*hitIt)) {
      //cout << " removed. " ;
      badHits.push_back(*hitIt) ;
      GetHits().erase(hitIt--) ;
      OneFitIteration(kNormal) ;
      changed=true ;
    } else {
      if(!success) (*hitIt)->FlipLRSign() ;
      //else cout << " lrsign changed " ;
    }
    //cout << "|||" ;
    // If you need the residuals longer, you have to take care they are up-to-date:
    if(changed) {
      OneFitIteration(kNormal) ;
      sort(hitIt+1,GetHits().end(),compareHitChi2) ;
    }
  }
  success = ValidChi2() ;
  for(HitContainer::iterator hitIt=badHits.begin() ;
      hitIt!=badHits.end(); hitIt++) delete *hitIt ;

#ifdef USE_ROOT
  static TH1F* chi2H1 = new TH1F("scchi2H1","",1000,0,100) ;
  chi2H1->Fill(GetChi2OverNDof()) ;
  static TH1F* chi2probH1 = new TH1F("scchi2probH1","",1000,0,1) ;
  //if(fChi2 < 100*chisin_(1-fgMinChi2Prob,GetNDof()))
  chi2probH1->Fill(GetChi2Prob()) ;
  static TH1F* clusterSigmaXH1 = new TH1F("clusterSigmaXH1","",1000,0,0.1) ;
  clusterSigmaXH1->Fill(sqrt(GetCov()[0][0])*GetCentralZ()) ;
  static TH1F* clusterSigmaYH1 = new TH1F("clusterSigmaYH1","",1000,0,10.0) ;
  clusterSigmaYH1->Fill(sqrt(GetCov()[1][1])*GetCentralZ()) ;  
#ifdef USE_ARTE
  if(GetChi2Prob() > 0.001) {
    static TH1F* lrsignNH1 = new TH1F("lrsignNH1","",3,-1.5,1.5) ;
    static TProfile* lrsignNPR = new TProfile("lrsignNPR","",20,-1,1) ;
    for(HitContainer::iterator hitIt=GetHits().begin() ;
	hitIt!=GetHits().end(); hitIt++) {
      // 	static TProfile* lrSignPR = new TProfile("lrSignPR","",20,-1,1) ;
      ArtePointer<HITB> hitb = (*hitIt)->GetMeasuredHit()->GetHITB() ;
      int realSign = int(hitb->q / fabs(hitb->q)) ;
      int recSign = (*hitIt)->GetLRSign() ;
      lrsignNH1->Fill(recSign*realSign) ;
      lrsignNPR->Fill(hitb->q/((*hitIt)->GetMeasuredHit()->
			       GetChannel()->GetRTRelation()->GetRMax()),
		      float(recSign*realSign)) ;
    }
  }
#endif // USE_ARTE
#endif // USE_ROOT
//   HitContainer::iterator hitIt= GetHits().begin() ;
//   ArtePointer<MTRA> mtra((ArtePointer<MIMP>(ArteRelationIterator<MIMP>((*hitIt)->GetMeasuredHit()->GetHITB()->mimp)))->mtra)   ;
//   for(; hitIt!=GetHits().end() ; hitIt++) 
//     if(mtra!=
//        ArtePointer<MTRA>((ArtePointer<MIMP>(ArteRelationIterator<MIMP>((*hitIt)->GetMeasuredHit()->GetHITB()->mimp)))->mtra)) break ;
//   if(hitIt!=GetHits().end()) return false ;

  return success ;
} 

bool
OTrSuperCluster::ValidNHits(const OTrReconstructedHit* doNotTakeThisHit) const
{
  //cout << "OTrTrack::ValidNHits()" << endl ;
  float slId[] = {0,0,0} ;
  for(HitContainer::const_iterator hitIt=GetHits().begin() ;
      hitIt!=GetHits().end(); hitIt++) 
    if(*hitIt!=doNotTakeThisHit)
      slId[(*hitIt)->GetMeasuredHit()->GetWire()->GetGede()->GetEStereoAngle()]++;
  
  return (slId[0]>0) && (slId[1]>0) && (slId[2]>0) ; 
};


void
OTrSuperCluster::Dump() const
{
  cout << "OTrSuperCluster::Dump()" << endl 
       << "Zero: " ;
  for(ClusterContainer::const_iterator zeroIt=fZeroClusters.begin() ;
       zeroIt!=fZeroClusters.end(); zeroIt++) (*zeroIt)->Dump() ;
  cout << "Plus: " ;
  for(ClusterContainer::const_iterator plusIt=fPlusClusters.begin() ;
       plusIt!=fPlusClusters.end(); plusIt++) (*plusIt)->Dump() ;
  cout << "Minus: " ;
  for(ClusterContainer::const_iterator minusIt=fMinusClusters.begin() ;
       minusIt!=fMinusClusters.end(); minusIt++) (*minusIt)->Dump() ;
};

#ifdef USE_ROOT
void
OTrSuperCluster::Draw(int color) 
{
  for(ClusterContainer::const_iterator clusIt=fSectionClusters.begin() ;
      clusIt!=fSectionClusters.end() ; clusIt++) (*clusIt)->Draw(color) ;
  
  //#ifdef USE_ROOT
  //cout << "OTrHitCluster::Draw(int color)" << endl ;
  //   if(!fClosestPointMarker)
  //     fClosestPointMarker = new TPolyMarker3D(1,20) ;
  //   fClosestPointMarker->SetPoint(0,GetTx()*GetZ(),GetTy()*GetZ(),GetZ()) ;
  //   fClosestPointMarker->SetMarkerColor(2) ;
  //   fClosestPointMarker->SetMarkerSize(10) ;
  //   fClosestPointMarker->Draw() ;
  //#endif
};
#endif
