#include <CLHEP/Matrix/DiagMatrix.h>
#include <CLHEP/Matrix/Matrix.h>
#include <CLHEP/Matrix/SymMatrix.h>
#include <CLHEP/Matrix/Vector.h>
#include <iostream.h>
#include <iomanip.h>
#include <functional>
#include <algorithm>

#ifdef USE_ROOT
 #include <TProfile.h>
#endif

#include "wax/OTrTrack.hh"
#include "wax/OTrNodeProjection.hh"
#include "wax/OTrHit.hh"
#include "wax/OTrTrackCandidate.hh"
#include "wax/OTrSuperCluster.hh"
#include "wax/OTrParameterizedRTRelation.hh"
#include "wax/OTrReconstructedHit.hh"

//__________________________________________________________________
// OTrTrack
//
// Container for:
// * reconstructed hits
// Functionality:
// * track fit
// * removement of bad hits
//
// Author: Wouter Hulsbergen 
//__________________________________________________________________

//#ifdef __OTR_CLASSDEF__
//ClassImp(OTrTrack) ;
//#endif

//
// compare call
//
template<class T,class Result>
class comparisonCall:
  public binary_function<T*,T*,int >
{
public:
  typedef T* argument_type;
  typedef int result_type;
  comparisonCall( Result (T::* m)() const ): member( m ) {}
  int operator()( const T* el1, const T* el2 ) const {
    return ((*el1).*member)() < ((*el2).*member)();
  }
private:
  Result (T::* member)() const;
};

template<class T,class Result>
comparisonCall<T, Result> compareCall( Result (T::* m)() const )
{ return comparisonCall<T,Result>( m ); }


int OTrTrack::fgMaxNSteps = 5 ;
float OTrTrack::fgMinChi2Probability = 0.00001 ;
int OTrTrack::fgMaxNStepsForEtaConvergence = 1 ;
//Obsolete
int OTrTrack::fgMaxNStepsForChi2Convergence = 3 ;
unsigned int OTrTrack::fgMinNHits = 5 ;

bool
OTrTrack::Valid() {
  return (GetError()==0) && (GetHits().size()>=fgMinNHits) ;
} ;

bool
OTrTrack::ValidNHits(const OTrReconstructedHit* doNotTakeThisHit) const
{
  //cout << "OTrTrack::ValidNHits()" << endl ;
  float slId[] = {-1,-1,-1} ;
  for(HitContainer::const_iterator hitIt=GetHits().begin() ;
      hitIt!=GetHits().end(); hitIt++) 
    if(*hitIt!=doNotTakeThisHit) {
      int stereoAngle = 
	(*hitIt)->GetMeasuredHit()->GetWire()->GetGede()->GetEStereoAngle() ;
      int thisSlId = (*hitIt)->GetMeasuredHit()->GetWire()->GetSuperLayerId() ;
      if(slId[stereoAngle]==-1) slId[stereoAngle] = thisSlId ;
      else if(slId[stereoAngle]!=-2 && slId[stereoAngle]!=thisSlId) 
	slId[stereoAngle]=-2 ;
    };
  return (slId[0]==-2) && (slId[1]==-2) && (slId[2]==-2) ; 
};

/*
bool
OTrTrack::CloseEnough(const ReconstructedHit* hit)
{
  float d = 0 ;
  hit->SetLRSign(d) ;
  return hit->CalcChi2(this) < fgMaxNSigmas ;
};

*/

OTrTrack::OTrTrack(const OTrTrackCandidate* trc) : OTrTrackDescription4D()
{
//   for(OTrTrackCandidate::ClusterContainer::const_iterator clIt=trc->GetClusters().begin() ;
//       clIt!=trc->GetClusters().end(); clIt++)
//     for(OTrSuperCluster::ClusterContainer::const_iterator secIt=(*clIt)->GetSectionClusters().begin() ;
// 	secIt!=(*clIt)->GetSectionClusters().end() ; secIt++) 
//       for(OTrSectionCluster::HitIterator hitIt=(*secIt)->begin(); hitIt!=(*secIt)->end() ; hitIt++) {
// 	signed char lrsign = (*secIt)->GetLRSign(*hitIt) ;
// 	//fHits.push_back(OTrReconstructedHit(*hitIt,0)) ;
// 	static TH1F* lrsignAH1 = new TH1F("lrsignAH1","",3,-1.5,1.5) ;
// 	static TProfile* lrSignPR = new TProfile("lrSignPR","",20,-1,1) ;
// 	ArtePointer<HITB> hitb = (*hitIt)->GetHITB() ;
// 	int realSign = hitb->q / fabs(hitb->q) ;
// 	int recSign = (*secIt)->GetLRSign(*hitIt)/2 ;
// 	lrsignAH1->Fill(recSign*realSign) ;
// 	float t = (*hitIt)->GetDriftTime() ;
// 	float ri = (*hitIt)->GetChannel()->GetRTRelation()->GetR(t) ;
// 	//cout << hitb->q << " " << ri <<  " " << recSign*realSign << endl ;
// 	float r = hitb->q/(*hitIt)->GetChannel()->GetRTRelation()->GetRMax() ;
// 	lrSignPR->Fill(r,(float)recSign*realSign) ;
// 	//cout << hitb.index() << " " << hitb->q << " " << t << " " << ri << endl ;
// 	cout << hitb << " " << (int)lrsign ; 
// 	fHits.push_back(OTrReconstructedHit(*hitIt,lrsign)) ;
// 	cout << (int)fHits.back().GetLRSign() << endl ;
//       };
  
  for(OTrTrackCandidate::ClusterContainer::const_iterator 
	clIt=trc->GetClusters().begin() ;
      clIt!=trc->GetClusters().end(); clIt++)
    for(OTrSuperCluster::HitContainer::const_iterator 
	  hitIt=(*clIt)->GetHits().begin() ;
	hitIt!=(*clIt)->GetHits().end(); hitIt++) {
      GetHits().push_back(new OTrReconstructedHit(**hitIt)) ;
    };
  
  OTrTrackCandidate::ClusterContainer::const_iterator 
    clIt=trc->GetClusters().begin() ;
  if(clIt!=trc->GetClusters().end()) {
    SetTx((*clIt)->GetTx()) ;
    SetTy((*clIt)->GetTy()) ;
  }
} 

enum FitMode {
  kAmbiguity,
  kCheckChi2,
  kCompareChi2,
  kNormal,
  kNormalLRSignFixed
};

bool
OTrTrack::Fit(int fitMode)
{
  //cout << "OTrTrack::FitNice(...)" << endl ;
  
  if(GetNDof()<1) return false ;

  OTrTrackDescription* newTrack = new OTrTrackDescription4D(*this) ;
  int dim = GetDim() ;
  int nSteps(0), ierr;

  HepVector deltaQ(dim), dChi2dQ(dim) ;
  HepSymMatrix d2Chi2dQ2(dim), d2Chi2dQ2Inv(dim) ;

  HepDouble chi2(0), dchi2, prev_chi2;
   
  bool etaError(false) ;
  bool converged(false) ;
  bool updateInternalFlag ;
  bool measurementZero = (fitMode == kAmbiguity) ;
  bool updateInternalAllowed = (measurementZero || fitMode == kNormal) ;
  OTrNodeProjection* nodeProjection ;

  do{
    nSteps++ ;

    // Reset some values
    prev_chi2  = chi2 ;
    chi2 = 0 ;
    etaError=false ;
    for(int row=0; row<dim; row++) {
      dChi2dQ[row]=0;
      for(int col=0; col<=row ; col++) {
	d2Chi2dQ2[row][col]=0;
      };
    };

    updateInternalFlag =  (fitMode == kAmbiguity) || 
      (fitMode == kNormal && nSteps > fgMaxNStepsForEtaConvergence) ;

    // Loop over hits
    for(HitContainer::iterator hitIt=GetHits().begin() ; 
	hitIt!=GetHits().end(); hitIt++) {
      nodeProjection = (*hitIt)->GetNodeProjection() ;
      nodeProjection->Project(newTrack, updateInternalFlag) ;
      chi2      += nodeProjection->GetChi2() ;
      dChi2dQ   += nodeProjection->GetdChi2dQ() ;
      d2Chi2dQ2 += nodeProjection->Getd2Chi2dQ2() ;
      etaError = etaError || nodeProjection->GetEtaError() ;
      //cout << nodeProjection->GetChi2() << " " << chi2 << endl ;
    };

    // using qrsolve will be little bit faster, but that's for the future
    d2Chi2dQ2Inv = d2Chi2dQ2.inverse(ierr) ;
    if(ierr) { 
      cout << "shit: ierr" << d2Chi2dQ2 << endl ;
      delete newTrack ;
      return false;
    };
    deltaQ = - d2Chi2dQ2Inv * dChi2dQ ;
    
    // Check for oscillations
    if(nSteps>1 && chi2 > prev_chi2)  deltaQ *= 0.5 ;

    newTrack->SetQ(newTrack->GetQ() + deltaQ ) ;
    dchi2 = -d2Chi2dQ2Inv.similarity(dChi2dQ)/2 ; // That is the expected change in chi2
    chi2 += dchi2 ;
    converged =  -dchi2 < 0.1 && !(etaError && updateInternalAllowed)  ;
    //cout << "chi2: " << dchi2 << " " << chi2 << endl ; 
  } while (!converged && nSteps<fgMaxNSteps) ;
  
  //cout << "noSteps" << nSteps << endl ;
  
  switch(fitMode)
    {
    case kCheckChi2:
      if(!ValidChi2(chi2)) {
	//cout << "chi2 too high, " << chi2  << endl;
	delete newTrack ;
	return false ;
      }
      break ;
    case kCompareChi2:
      if(chi2 > 0.9*GetChi2()) { // jesus, that much handwaving ?!
	//cout << "worse than current " << chi2 << endl ; 
	delete newTrack ;
	return false;
      };
      break ;
    default:
      break ;
    };
  
  //newTrack.SetCov(2*d2Chi2dQ2Inv) ;
  SetQ(newTrack->GetQ()) ;
  SetCov(2*d2Chi2dQ2Inv) ;
  SetChi2(chi2) ;
  delete newTrack ;
  SetError(0) ;
  return true ;
} ;

inline bool 
compareHitDistance(OTrReconstructedHit& lhs, OTrReconstructedHit& rhs) {
  //return fabs(lhs.GetNodeProjection()->GetD()) > fabs(lhs.GetNodeProjection()->GetD()) ;
  return lhs.GetNodeProjection()->GetRm() > lhs.GetNodeProjection()->GetRm() ;
};

bool
OTrTrack::SearchHitsAndFit()
{
  bool success = false ;
  if(!ValidNHits()) return false ;
#ifdef USE_ARTE
#ifdef USE_ROOT
  static TH1F* lrsignBH1 = new TH1F("lrsignBH1","",3,-1.5,1.5) ;
  for(HitContainer::const_iterator hitIt=GetHits().begin() ;
      hitIt!=GetHits().end(); hitIt++) {
    // 	static TProfile* lrSignPR = new TProfile("lrSignPR","",20,-1,1) ;
    ArtePointer<HITB> hitb = (*hitIt)->GetMeasuredHit()->GetHITB() ;
    int realSign = int(hitb->q / fabs(hitb->q)) ;
    int recSign = (*hitIt)->GetLRSign() ;
    lrsignBH1->Fill(recSign*realSign) ;
  };    
#endif
#endif
  /// Removes all hits which are suspected of not belonging to track.
  HitContainer badHits ;
  Fit(kNormalLRSignFixed) ;
  for(HitContainer::iterator hitIt=GetHits().begin() ; hitIt!=GetHits().end(); hitIt++)
    (*hitIt)->UpdateInternal() ;
  Fit(kNormalLRSignFixed) ;
  success = ValidChi2() ;
  sort(GetHits().begin(), GetHits().end(), 
       not2( compareCall( &OTrReconstructedHit::GetChi2 ) )) ;

  for(HitContainer::iterator hitIt=GetHits().begin() ; hitIt!=GetHits().end(); hitIt++) {
    (*hitIt)->FlipLRSign() ;
    success = Fit(kCompareChi2) ;
    if(!ValidChi2() && ValidNHits(*hitIt)) {
      //fHits.erase(hitIt--) ;
      HitContainer::iterator prevIt = hitIt ; prevIt--;
      badHits.push_back(*hitIt) ;
      GetHits().erase(hitIt) ;
      hitIt = prevIt ;
      //cout << "Removed: " 
      //	   << badHits.front().GetMeasuredHit()->GetHITB() << endl ;
      Fit(kNormalLRSignFixed) ;
    } else if(!success) {
      (*hitIt)->FlipLRSign() ;
      //Fit(kNormalLRSignFixed) ;
    };
  };
  success = ValidChi2() ;
#ifdef USE_ARTE
#ifdef USE_ROOT
  static TH1F* lrsignCH1 = new TH1F("lrsignCH1","",3,-1.5,1.5) ;
  static TProfile* lrsignMPR = new TProfile("lrsignCPR","",20,-1,1) ;
  for(HitContainer::const_iterator hitIt=GetHits().begin() ; 
      hitIt!=GetHits().end(); hitIt++) {
    // 	static TProfile* lrSignPR = new TProfile("lrSignPR","",20,-1,1) ;
    ArtePointer<HITB> hitb = (*hitIt)->GetMeasuredHit()->GetHITB() ;
    int realSign = int(hitb->q / fabs(hitb->q)) ;
    int recSign = (*hitIt)->GetLRSign() ;
    lrsignCH1->Fill(recSign*realSign) ;
    lrsignMPR->Fill(hitb->q/((*hitIt)->GetMeasuredHit()->
			     GetChannel()->GetRTRelation()->GetRMax()),
		    float(recSign*realSign)) ;
  };
#endif
#endif
  return success ;
};

// bool
// OTrTrack::SearchHitsAndFit()
// {
//   bool success = false ;
//   if(!ValidNHits()) return false ;

//   vector<OTrReconstructedHit*> nonFixedHits;
//   for(HitContainer::iterator hitIt=fHits.begin() ;
//       hitIt!=fHits.end() ; hitIt++)
//     if(!(hitIt->GetLRSignIsFixed())) nonFixedHits.push_back(&(*hitIt)) ;
  
//   cout << "NoHits(fixed)" << fHits.size() << " " << nonFixedHits.size() << endl ;
//   int maxLRSign = pow(2,nonFixedHits.size()) ;
//   cout << "maxLRSign: " << maxLRSign << endl ;
//   int besti;
//   for(int i=0; i<maxLRSign; i++) {
//     int j=i;
//     for(vector<OTrReconstructedHit*>::iterator hitIt=nonFixedHits.begin();
// 	hitIt!=nonFixedHits.end() ; hitIt++) {
//       if(j%2) (*hitIt)->SetLRSign(1.0) ;
//       else (*hitIt)->SetLRSign(-1.0) ;
//       j /=2;
//     };
//     if(!success) success=Fit(kNormalLRSignFixed) ;
//     else if(success=Fit(kCompareChi2)) besti=i;
//   } ;
//   int j=besti;
//   for(HitContainer::iterator hitIt=fHits.begin() ;
//       hitIt!=fHits.end() ; hitIt++) {
//     if(j%2) hitIt->SetLRSign(1.0) ;
//     else hitIt->SetLRSign(-1.0) ;
//   };
//   success=Fit(kNormalLRSignFixed) ;
//   HitContainer badHits ;
//   while(!(success=ValidChi2())) {
//     fHits.sort( not2( compareCall( &OTrReconstructedHit::GetChi2 ) ) ) ;
//     badHits.splice(badHits.begin(),fHits,fHits.begin()) ;
//     if(!ValidNHits()) {
//       fHits.splice(fHits.begin(),badHits,badHits.begin()) ;
//       break ;
//     };
//     success = Fit(kNormalLRSignFixed) ;
//   };
//   return success;
// };


// bool
// OTrTrack::SearchHitsAndFit()
// {
//   bool success = false ;
//   if(!ValidNHits()) return false ;
//   if(Fit(kNormalLRSignFixed)) {
//     fHits.sort(compareHitDistance) ;
//     cout << "Chi2: " << GetChi2() << " " << GetChi2Prob() << endl ;
//     for(HitContainer::iterator hitIt=fHits.begin() ;
// 	hitIt!=fHits.end() ; hitIt++) 
//       if(!hitIt->GetLRSignIsFixed()) {
// 	cout << hitIt->GetMeasuredHit()->GetHITB() << " " 
// 	     << (int)hitIt->GetLRSign() << " " ;
// 	hitIt->SetLRSign(1.0) ;
// 	bool thissuccess = Fit(kNormalLRSignFixed) ;
// 	hitIt->FlipLRSign() ;
// 	if(!Fit(kCompareChi2)) hitIt->FlipLRSign() ; // Fit wrong, flip back
// 	cout << (int)hitIt->GetLRSign() << " "
// 	     << hitIt->GetNodeProjection()->GetD() << " "<< thissuccess << " "
// 	     << " " << GetChi2() << " " << GetChi2Prob() << endl ;
//       };
//     static TH1F* lrsignBH1 = new TH1F("lrsignBH1","",3,-1.5,1.5) ;
//     for(HitIterator hitIt=fHits.begin() ; hitIt!=fHits.end(); hitIt++) {
//       // 	static TProfile* lrSignPR = new TProfile("lrSignPR","",20,-1,1) ;
//       ArtePointer<HITB> hitb = (*hitIt).GetMeasuredHit()->GetHITB() ;
//       int realSign = hitb->q / fabs(hitb->q) ;
//       int recSign = hitIt->GetLRSign() ;
//       lrsignBH1->Fill(recSign*realSign) ;
//     };    
//     /// Removes all hits which are suspected of not belonging to track.
//     HitContainer badHits ;
//     success = ValidChi2() ;
//     fHits.sort( not2( compareCall( &OTrReconstructedHit::GetChi2 ) ) ) ;
//     for(HitContainer::iterator hitIt=fHits.begin() ; hitIt!=fHits.end(); hitIt++) {
//       hitIt->FlipLRSign() ;
//       success = Fit(kCompareChi2) ;
//       if(!ValidChi2() && ValidNHits(&(*hitIt))) {
// 	HitContainer::iterator prevIt = hitIt ; prevIt--;
// 	badHits.splice(badHits.begin(),fHits,hitIt) ;
// 	hitIt = prevIt ;
// 	cout << "Removed: " 
// 	     << badHits.front().GetMeasuredHit()->GetHITB() << endl ;
//       } else if(!success) hitIt->FlipLRSign() ;
//     };
//     // Deze verslechtert het alleen maar.
//     success = Fit(kNormalLRSignFixed) ;

//     static TH1F* lrsignCH1 = new TH1F("lrsignCH1","",3,-1.5,1.5) ;
//     for(HitIterator hitIt=fHits.begin() ; hitIt!=fHits.end(); hitIt++) {
//       // 	static TProfile* lrSignPR = new TProfile("lrSignPR","",20,-1,1) ;
//       ArtePointer<HITB> hitb = (*hitIt).GetMeasuredHit()->GetHITB() ;
//       int realSign = hitb->q / fabs(hitb->q) ;
//       int recSign = hitIt->GetLRSign() ;
//       lrsignCH1->Fill(recSign*realSign) ;
//     };
//   };
//   return success ;
// };

// bool
// OTrTrack::SearchHitsAndFit()
// {
//   if(fHits.size()<fgMinNHits) return false ;
//   HitContainer badHits ;
//   bool success = false ;
//   /// First perform fit with all times zero and set ambiguity to sign of distance
//   //cout << "nHits before fit: " << fHits.size()<< endl  ;
//   if(true /*Fit(kAmbiguity)*/) {
//     static TH1F* lrsignAH1 = new TH1F("lrsignBH1","",3,-1.5,1.5) ;
//     for(HitIterator hitIt=fHits.begin() ; hitIt!=fHits.end(); hitIt++) {
//       // 	static TProfile* lrSignPR = new TProfile("lrSignPR","",20,-1,1) ;
//       ArtePointer<HITB> hitb = (*hitIt).GetMeasuredHit()->GetHITB() ;
//       int realSign = hitb->q / fabs(hitb->q) ;
//       int recSign = hitIt->GetLRSign() ;
//       lrsignAH1->Fill(recSign*realSign) ;
//     };

//     /// Now perform normal fit 
//     if(Fit(kNormalLRSignFixed)) { 
//       cout << "Chi2 now: " << GetChi2() << " " << GetChi2Prob() << endl ;
//       //  for(HitContainer::iterator hitIt=fHits.begin() ;
//       // 	  hitIt!=fHits.end() ; hitIt++)
//       // 	hitIt->SetLRSign(hitIt->GetNodeProjection()->GetD()) ;
//       //       Fit(kNormal) ;
//       /// Removes all hits which are suspected of not belonging to track.
//       while(!(success=ValidChi2())) {
// 	fHits.sort( not2( compareCall( &OTrReconstructedHit::GetChi2 ) ) ) ;
// 	badHits.splice(badHits.begin(),fHits,fHits.begin()) ;
// 	if(!ValidNHits()) {
// 	  fHits.splice(fHits.begin(),badHits,badHits.begin()) ;
// 	  break ;
// 	};
// 	//fHits.erase(fHits.begin()) ; 
// 	Fit(kNormalLRSignFixed) ;
//       };
//       static TH1F* lrsignBBH1 = new TH1F("lrsignBBH1","",3,-1.5,1.5) ;
//       for(HitIterator hitIt=fHits.begin() ; hitIt!=fHits.end(); hitIt++) {
// 	// 	static TProfile* lrSignPR = new TProfile("lrSignPR","",20,-1,1) ;
// 	ArtePointer<HITB> hitb = (*hitIt).GetMeasuredHit()->GetHITB() ;
// 	int realSign = hitb->q / fabs(hitb->q) ;
// 	int recSign = hitIt->GetLRSign() ;
// 	lrsignBBH1->Fill(recSign*realSign) ;
//       };
      
//       /// Change all signs once to see whether we can make better estimate of lrsign
//       fHits.sort(compareHitDistance) ;
//       for(HitContainer::iterator hitIt=fHits.begin() ;
// 	  hitIt!=fHits.end() ; hitIt++) {
// 	hitIt->FlipLRSign() ;   // Flip
// 	if(!Fit(kCompareChi2)) hitIt->FlipLRSign() ; // Fit wrong, flip back
// 	else cout << "***** FIT CHANGED *****" << GetChi2() << endl ;
//       };
 
//       //cout << "nHits after erasure bad hits: " << fHits.size()<< endl  ;
//       //cout << "Chi2: " << GetChi2() << endl ;
//       /// Change all signs once to see whether we can make better estimate of lrsign
//       // fHits.sort(compareHitDistance) ;
//       //       for(HitContainer::iterator hitIt=fHits.begin() ;
//       // 	  hitIt!=fHits.end() ; hitIt++) {
//       // 	hitIt->FlipLRSign() ;   // Flip
//       // 	if(!Fit(kCompareChi2)) hitIt->FlipLRSign() ; // Fit wrong, flip back
//       // 	//else cout << "***** FIT CHANGED *****" << GetChi2() << endl ;
//       //       };
//       static TH1F* lrsignCH1 = new TH1F("lrsignCH1","",3,-1.5,1.5) ;
//       for(HitIterator hitIt=fHits.begin() ; hitIt!=fHits.end(); hitIt++) {
// 	// 	static TProfile* lrSignPR = new TProfile("lrSignPR","",20,-1,1) ;
// 	ArtePointer<HITB> hitb = (*hitIt).GetMeasuredHit()->GetHITB() ;
// 	int realSign = hitb->q / fabs(hitb->q) ;
// 	int recSign = hitIt->GetLRSign() ;
// 	lrsignCH1->Fill(recSign*realSign) ;
//       };
//       if(success=ValidChi2()) {
// 	/// Now add dustbin hits to see whether there is a chance they fit afterall.
// 	while(badHits.begin()!=badHits.end()) {
// 	  fHits.splice(fHits.begin(),badHits,badHits.begin()) ;
// 	  // update lrsign
// 	  fHits.begin()->GetNodeProjection()->Project(*this, true, false) ;
// 	  if(!Fit(kCheckChi2)) fHits.pop_front() ;
// 	};
//       };
//       static TH1F* lrsignDH1 = new TH1F("lrsignDH1","",3,-1.5,1.5) ;
//       for(HitIterator hitIt=fHits.begin() ; hitIt!=fHits.end(); hitIt++) {
// 	// 	static TProfile* lrSignPR = new TProfile("lrSignPR","",20,-1,1) ;
// 	ArtePointer<HITB> hitb = (*hitIt).GetMeasuredHit()->GetHITB() ;
// 	int realSign = hitb->q / fabs(hitb->q) ;
// 	int recSign = hitIt->GetLRSign() ;
// 	lrsignDH1->Fill(recSign*realSign) ;
//       };
//     };
//   };
//   cout << "Chi2 after: " << GetChi2() << " " << GetChi2Prob() << " " << success << endl ;
//   if (!success) cout << "Track not valid" << endl ;
//   return success ;
// } ;




/*********** JUNK **************
Uit ranger :
rseg->cvf[ 0] = cov()(1,1);
  rseg->cvf[ 1] = cov()(1,2);
  rseg->cvf[ 2] = cov()(2,2);
  rseg->cvf[ 3] = cov()(3,1);
  rseg->cvf[ 4] = cov()(3,2);
  rseg->cvf[ 5] = cov()(3,3);
  rseg->cvf[ 6] = cov()(4,1);
  rseg->cvf[ 7] = cov()(4,2);
  rseg->cvf[ 8] = cov()(4,3);
  rseg->cvf[ 9] = cov()(4,4);
  rseg->cvf[10] = cov()(5,1);
  rseg->cvf[11] = cov()(5,2);
  rseg->cvf[12] = cov()(5,3);
  rseg->cvf[13] = cov()(5,4);
  rseg->cvf[14] = cov()(5,5);

   //Check extremum
    //normalizedDeterminant = 1 - d2Chi2dQ2[1][0]*d2Chi2dQ2[1][0]/
    // (d2Chi2dQ2[0][0]*d2Chi2dQ2[1][1]) ;
    //if((d2Chi2dQ2[0][0]<=0)||(d2Chi2dQ2[1][1]<=0)||(normalizedDeterminant<=0))
    // {
    //   fError = 1;
    //   //cout << "shit: no minimum" << endl ;
    //   return false ;
    // };
// bool
// OTrTrack::Fit(int fitMode)
// {
//   // As angles are always small, the fit is almost linear. We
//   // therefore neglect the second derivatives. 
//   if(GetNDof()<1) return false ;

//   cout << "OTrTrack::OTrTrackFit(...)" << endl ;
//   // First fill some arrays for speed
//   int nHits = GetHits().size() ;
//   //cout << nHits << endl ;
//   //HepDouble t0 = fgRTRelation->GetT0();
//   vector<HepDouble> costhetaw(nHits),sinthetaw(nHits),
//     zwcosthetaw(nHits),zwsinthetaw(nHits),xw(nHits), tm(nHits), t0(nHits);
//   vector<int> eps(nHits);
//   vector<OTrRTRelation*> rtrelation(nHits) ;
//   HepDiagMatrix W(nHits) ;

//   int i;
//   list<OTrReconstructedHit>::iterator hit=fHits.begin() ;
//   for( i=0; i<nHits; i++, hit++)
//     {
//       const OTrWire* wire = hit->GetMeasuredHit()->GetWire() ;
//       // The next minus sign makes this routine agree with the rest of
//       // wax. However, you will find the same tracks with a plus sign,
//       // except for a minus in y and ty.

//       double theta = -1*wire->GetSubSection()->GetSection()->GetGede()->GetStereoAngle() ;
//       double zw = wire->GetBeginPoint().z() ;
//       costhetaw[i] = cos(theta);
//       sinthetaw[i] = sin(theta);
//       zwcosthetaw[i] = zw*costhetaw[i] ;
//       zwsinthetaw[i] = zw*sinthetaw[i] ;
//       xw[i] =  wire->GetRawX() ;
//       eps[i] = hit->GetLRSign() ;
//       rtrelation[i] = hit->GetMeasuredHit()->GetChannel()->GetRTRelation() ;
//       t0[i] = rtrelation[i]->GetT(0).GetValue() ;
//       switch(fitMode)
// 	{
// 	case kAmbiguity:
// 	  tm[i] = t0[i] ;
// 	  break;
// 	default:
// 	  tm[i] = hit->GetMeasuredHit()->GetDriftTime() ;
// 	  break ;
// 	};
//       double R = rtrelation[i]->GetR(tm[i]) ;
//       W[i][i] = 1./fabs(rtrelation[i]->GetSigma2(R).GetValue()) ;
// //       cout << tm[i] << " " << R << " " << W[i][i] << " " 
// // 	   << wire->GetId() << " " << theta << endl ;
//     };

//   HepVector Q = GetQ() ;
//   HepDouble& x = Q[0]; HepDouble& y = Q[1]; HepDouble& tx = Q[2]; HepDouble& ty = Q[3]; 
//   int dim = Q.num_row() ;
//   int nSteps(0), ierr;
//   HepVector deltaQ(dim), dChi2dQ(dim) ;
//   HepSymMatrix d2Chi2dQ2(dim), d2Chi2dQ2Inv(dim) ;
//   HepVector Res(nHits) ;
//   HepVector Chi2(nHits) ;
//   HepMatrix dResdQ(dim,nHits) ;

//   HepDouble chi2(0), prev_chi2;
//   HepDouble dchi2,prev_dchi2 ;
  
//   //FunctionValue T;
//   HepDouble d,res ;
//   vector<int> signD(nHits) ;
//   int eta ;
//   bool etaError(false), converged(false);
//   bool suspectOfOscillations = false;
//   bool seriousSuspectOfOscillations = false ;
//   HepDouble noemer;

//   do{
//     nSteps++ ;
//     prev_dchi2 = dchi2 ;
//     prev_chi2  = chi2 ;
        
//     chi2 = 0 ;
//     etaError = false ;

//     //cout << "Step: " << noSteps 
//     //	 << " phi: " << phi 
//     //	 << " d0:  " << d0 << endl ;   
    
//     for(i=0; i<nHits; i++)
//       {
// 	noemer = sqrt( 1 + sqr(tx*costhetaw[i]+ty*sinthetaw[i])) ;
// 	d = ( x*costhetaw[i] + y*sinthetaw[i] +
// 	     tx*zwcosthetaw[i] + ty*zwsinthetaw[i] - xw[i] ) / noemer ;
// 	signD[i] = 1; if(d<0.) signD[i]=-1;
// 	eta = eps[i]*signD[i] ;
// 	if(nSteps > fgMaxNStepsForEtaConvergence && eta <0) { // DO I WANT THIS ?!!
// 	  eta=1; eps[i]*=-1;
// 	};
	  
// 	double R = d*signD[i] ;
// 	const OTrFunctionValue& T = rtrelation[i]->GetT(R) ;
// 	res    = (T.value-t0[i]) - eta * (tm[i]-t0[i]) ;
//   	cout << i << " " << d << " " << eps[i] 
//   	     << " res: " << res
//   	     << " T.value" << T.value
//   	     << " tm: " << tm[i]
//  	     << " chi2: " << res*res*W[i][i] << endl ;
	
// 	//dChi2dD[i] = 2*W[i][i]*res*T.ddx*signD ;
// 	etaError = etaError || ((eta<0) ) ;//&& (fabs(d*dTdR)>2*tt.error)) ;
// 	Chi2[i] = res*res*W[i][i] ;
// 	Res[i] = res ;
// 	//Here we could use some ... 
// 	dResdQ[0][i] = T.ddx*signD[i] * costhetaw[i]/noemer ;
// 	dResdQ[1][i] = T.ddx*signD[i] * sinthetaw[i]/noemer ;
// 	dResdQ[2][i] = T.ddx*signD[i] * zwcosthetaw[i]/noemer ;
// 	dResdQ[3][i] = T.ddx*signD[i] * zwsinthetaw[i]/noemer ;
//       }
    
//     dChi2dQ   = 2* dResdQ * W * Res ;
//     d2Chi2dQ2 = 2* W.similarity(dResdQ) ;
//     chi2 = W.similarity(Res) ;
//     cout << "Chi2: " << chi2 <<endl ;

// //     if(nSteps > fgMaxNStepsForEtaConvergence && etaError) {
// //       cout << "shit: no eta convergence" << endl; 
// //       fError = 2; return false ; };

//     // Check for oscillations
//     if(nSteps>1) {
// 	dchi2 = chi2 - prev_chi2 ;
// 	seriousSuspectOfOscillations = suspectOfOscillations || seriousSuspectOfOscillations ;
// 	suspectOfOscillations = dchi2 > 0 ;
// 	//cout << seriousSuspectOfOscillations << " " << suspectOfOscillations << endl ;
// 	if((nSteps > fgMaxNStepsForChi2Convergence) && suspectOfOscillations && 
// 	   seriousSuspectOfOscillations) 
// 	  break ;
//     };

//    d2Chi2dQ2Inv = d2Chi2dQ2.inverse(ierr) ;
//    if(ierr) cout << "shit: ierr" << endl ;
//    deltaQ = - d2Chi2dQ2Inv * dChi2dQ ;
//    if(suspectOfOscillations) deltaQ *= 0.5 ;

//    cout << "Q:" << Q[0] << " " << Q[1] << " " << Q[2] << " "<< Q[3] << endl ;
//    Q += deltaQ;
   
//    // cout << dChi2dQ << endl
//    // 	<< d2Chi2dQ2Inv << endl ;
//    double convfactor = d2Chi2dQ2.similarity(deltaQ) ;
//    cout << convfactor << endl ;
//    //converged = (fabs(deltaQ[0])< 0.5*d2Chi2dQ2Inv[0][0]) 
//    //  && (fabs(deltaQ[1])< 0.5*d2Chi2dQ2Inv[1][1]) ; // && !etaError ;
//    converged = convfactor < 0.1 && !etaError;
//   } while (!converged && nSteps<fgMaxNSteps) ;
  
//   cout << "noSteps" << nSteps << endl ;
  
//   switch(fitMode)
//     {
//       //case kAmbiguity:
//       //hit=fHits.begin() ;
//       //for( i=0; i < nHits; i++, hit++) hit->SetLRSign(signD[i]) ;
//       //break ;
//     case kCheckChi2:
//       if(!ValidChi2(chi2)) {
// 	cout << "chi2 too high, " << chi2  << endl;
// 	return false ;
//       }
//       break ;
//     case kCompareChi2:
//       if(chi2 > GetChi2()) {
// 	cout << "worse than current" << endl ; 
// 	return false;
//       };
//       break ;
//     default:
//       break ;
//     };
  
//   hit=fHits.begin() ;
//   for( i=0; i < nHits; i++, hit++) hit->SetLRSign(signD[i]) ;
//   SetQ(Q) ;
//   SetCov(2*d2Chi2dQ2Inv) ;
//   SetChi2(chi2) ;

//   if(etaError) { 
//     cout << "still eta error!!!!!!!!!!!!!!!!!!!!!!!" << endl;
//     fError=2; //return false ;
//   };
  
//   fError = 0;
//   hit=fHits.begin() ;
//   for( i=0; i<nHits; i++, hit++) {
//     hit->SetResidual(Res[i]) ;
//     hit->SetChi2(Chi2[i]) ;
//   };
//   return true ;
// } ;

    */
