#include <math.h>
#include <CLHEP/Utilities/fortran.h>
#include <CLHEP/Matrix/Matrix.h>
#include <CLHEP/Matrix/DiagMatrix.h>
#include "wax/OTrTrackCandidate.hh"
#include "wax/OTrHit.hh"
#include "wax/OTrReconstructedHit.hh"
#include "wax/OTrTrack.hh"
#include "wax/OTrWire.hh"
#include "wax/OTrGede.hh"
#include "wax/OTrEvent.hh"
#include "wax/OTrSuperCluster.hh"
#include "wax/OTrSection.hh"

#ifdef USE_ROOT
 #include <TPolyLine3D.h>
 #include <TPad.h>
 #include <TH2.h>
#endif

#ifdef USE_ARTE
#include <iomanip>
#include <list>
#include <arte/arte.hh>
#include <arte/artefunc.h>
#include <rang/Track.hh>
#include <rang/Node.hh>
#include <rang/Hit.hh>
#include <rang/HitWire.hh>
#include <rang/Geom.hh>
#endif //USE_ARTE

//#define MC_DEBUG

OTrTrackCandidate::OTrTrackCandidate() : 
  fRangerTrack(0), fTrack(0), fPolyLine3D(0)
{
  cout << "creating a default OTrTrackCandidate" << endl;
}

OTrTrackCandidate::OTrTrackCandidate(const OTrSuperCluster* cl) :
  fRangerTrack(0), fTrack(0), fPolyLine3D(0), fQ(4,0), fCov(4,0), fChi2(0)
{
  fClusters.push_back(cl) ;
  fQ.sub(3,cl->GetQ()) ;
  fCov[0][0] = sqr(1.0) ;//cm, almost arbitrary value
  fCov[1][1] = fCov[0][0] ;//cm
  fCov.sub(3,cl->GetCov()) ;
}

OTrTrackCandidate::OTrTrackCandidate( const OTrTrackCandidate& rhs )
{
  fClusters = rhs.fClusters;
#ifdef USE_ROOT
  if(rhs.fPolyLine3D)
    fPolyLine3D = new TPolyLine3D( *(rhs.fPolyLine3D) );
  else
    fPolyLine3D = 0 ;
#else
  fPolyLine3D = rhs.fPolyLine3D;
#endif
  fTargetChi2 = rhs.fTargetChi2 ;
  fRSEG = rhs.fRSEG;
  if(rhs.fTrack) fTrack = new OTrTrack(*(rhs.fTrack)) ;
  else fTrack = 0;
  fQ = rhs.fQ ;
  fCov = rhs.fCov ;
  fChi2 = rhs.fChi2 ;
#ifdef  USE_ARTE
  if(rhs.fRangerTrack) fRangerTrack = new rangerTrack(*(rhs.fRangerTrack)) ;
  else fRangerTrack = 0;
#endif
}


OTrTrackCandidate::~OTrTrackCandidate()
{
  //  cout << "in ~OTrTrackCandidate()" << endl;
#ifdef USE_ROOT
  if ( fPolyLine3D ) {
    //cout << " destructing OTrTrackCandidate " << endl;
    TCollection::StartGarbageCollection();
    TCollection::GarbageCollect(fPolyLine3D);
    delete fPolyLine3D;
    fPolyLine3D = 0;
    //cout << " done " << endl;
  }
#endif //USE_ROOT
#ifdef  USE_ARTE
  if(fRangerTrack) delete fRangerTrack ;
#endif //USE_ARTE
  if(fTrack) delete fTrack ;
}

bool
OTrTrackCandidate::Valid() const
{
  return fClusters.size()>1  ;
// && fabs(GetX())< 10 //5 /*cm*/ 
//     && fabs(GetY())< 100 /*cm*/ ;
};

bool
OTrTrackCandidate::CleanUp(const OTrTrackCandidate& rhs)
{
  bool changed=false ;
  for(ClusterContainer::const_iterator rhsIt=rhs.fClusters.begin() ;
      rhsIt!=rhs.fClusters.end() ; rhsIt++) {
    for(ClusterContainer::iterator lhsIt = fClusters.begin() ;
	lhsIt!=fClusters.end(); lhsIt++) {
      if((*rhsIt)==(*lhsIt)) {
	fClusters.erase(lhsIt) ;
	lhsIt--;
	changed=true;
      };
    };
  };
  return changed ;
} 

void
OTrTrackCandidate::Add(const OTrSuperCluster* cluster)
{
  fClusters.push_back(cluster) ;
};

#define WITH_KALMAN
#ifdef WITH_KALMAN
bool
OTrTrackCandidate::AddIfValid(const OTrSuperCluster* cluster)
{
  static const float fgMaxChi2 = 5 ;
  static HepSymMatrix* Qk = 0;
  static HepMatrix* Hk;
  static HepVector* mk;
  static const HepSymMatrix* Vk;
  static HepSymMatrix* Wk;
  static HepMatrix* Kk;
  static HepSymMatrix* Ck;
  static HepSymMatrix* Cj;
  static HepDiagMatrix* I4 ;
  static HepVector* xk;
  static HepVector* dxk;
  static HepVector* rk;
  if(Qk==0) {
    xk = new HepVector(4,0) ;
    dxk = new HepVector(4,0) ;
    Ck = new HepSymMatrix(4) ;
    Cj = new HepSymMatrix(4) ;
    Qk = new HepSymMatrix(4,0) ;
    (*Qk)[2][2] = sqr(0.0010); // quite a reasonable guess (1mm/m)
    (*Qk)[3][3] = (*Qk)[2][2];
    Hk = new HepMatrix(2,4,0) ;
    (*Hk)[0][0]=1; (*Hk)[1][1]=1;
    Kk = new HepMatrix(4,2,0);
    Wk = new HepSymMatrix(2,0);
    mk = new HepVector(2,0) ;
    rk = new HepVector(2,0) ;
    I4 = new HepDiagMatrix(4,1) ;
  }
  bool added(false) ;
  float z  = cluster->GetCentralZ() ;
  (*mk)[0] = cluster->GetCentralX() ;
  (*mk)[1] = cluster->GetCentralY() ;
  (*Hk)[0][2] = z; (*Hk)[1][3] = z;
  (*rk) = (*Hk)*fQ - (*mk) ;
  (*Cj)    = fCov + (*Qk) ;
  Vk = &(cluster->GetCov()) ;
  int ierr;
  (*Wk) = ((*Vk)+(*Cj).similarity(*Hk)).inverse(ierr) ;
  float chi2 = 2*(*Wk).similarity(*rk) ;
  //float chi2 = (*Cj).inverse(ierr).similarity(*dxk) + (*Wk).similarity(*rk) ; // this makes it slow !!

#ifdef MC_DEBUG
  ArtePointer<MTRA> firstmtra((ArtePointer<MIMP>(ArteRelationIterator<MIMP>(fClusters.front()->GetHits().front()->GetMeasuredHit()->GetHITB()->mimp)))->mtra)   ;
  ArtePointer<MTRA> newmtra((ArtePointer<MIMP>(ArteRelationIterator<MIMP>(cluster->GetHits().front()->GetMeasuredHit()->GetHITB()->mimp)))->mtra)   ;

  static TH1F* sameTrackChi2H1 = new TH1F("sameTrackChi2H1","sameTrackChi2H1",100,0,20) ;
  static TH1F* wrongTrackChi2H1 = new TH1F("wrongTrackChi2H1","wrongTrackChi2H1",100,0,20) ;
  static TH1F* sameTrackChi23clH1 = new TH1F("sameTrackChi23clH1","sameTrackChi23clH1",100,0,20) ;
  static TH1F* wrongTrackChi23clH1 = new TH1F("wrongTrackChi23clH1","wrongTrackChi23clH1",100,0,20) ;
  if(firstmtra.index()==newmtra.index()) {
    if(fClusters.size()>1) sameTrackChi23clH1->Fill(chi2);
    else sameTrackChi2H1->Fill(chi2) ;
#else 
  if(chi2<fgMaxChi2) {
#endif
    (*Kk) = (*Cj)*(*Hk).T()*(*Wk) ;
    //(*Ck) = (*Cj) - (*Kk)*(*Hk)*(*Cj) ; // so we don't have to define identity matrix :)
    (*Ck) = (*Cj) - (*Wk).similarityT((*Hk)*(*Cj)) ;
    (*Ck) = (*Cj).similarity((*I4)-(*Kk)*(*Hk)) + (*Vk).similarity(*Kk) ;
    (*dxk) = - (*Kk)*(*rk) ;
    (*xk) = fQ + (*dxk) ;
    fCov = (*Ck) ;
    fQ   = (*xk) ;
    fChi2 = chi2 ;
    fClusters.push_back(cluster) ;
    added = true ;
  } 
#ifdef MC_DEBUG
  else {
    if(fClusters.size()>1) wrongTrackChi23clH1->Fill(chi2) ;
    else wrongTrackChi2H1->Fill(chi2) ;
  }
#endif
#ifdef USE_ROOT
  static TH1F* tcchi2H1 = new TH1F("tcchi2H1","",500,0,100) ;
  tcchi2H1->Fill(chi2) ;
#endif
  return added ;
}
  
#else //WITH_KALMAN

bool
OTrTrackCandidate::AddIfValid(const OTrSuperCluster* cluster)
{
  static const float fgMaxChi2 = 5 ;
  static HepSymMatrix* multipleScatCov = 0;
  if(multipleScatCov==0) {
    multipleScatCov = new HepSymMatrix(2,1) ;
    *multipleScatCov *= sqr(0.005); //just a guess !
  }
  bool added(false) ;
  int ierr;
  //const OTrSuperCluster* referenceCluster = fClusters.front() ;
  const OTrSuperCluster* referenceCluster = fClusters.back() ;
  float chi2 = ((*multipleScatCov + referenceCluster->GetCov()+cluster->GetCov()).inverse(ierr)).
    similarity(referenceCluster->GetQ()-cluster->GetQ()) ;
  //int slId1 = cluster->GetSectionClusters().back()->GetSection()->GetSuperLayerId() ;
  //int slId2 = referenceCluster->GetSectionClusters().back()->GetSection()->GetSuperLayerId() ;
  //cout << slId1 << " " << slId2 << " " << referenceCluster->GetTx() << " " 
  //     << cluster->GetTx() << " " << chi2 << endl ;
  cout << referenceCluster->GetRSEG().index() << " " 
       << cluster->GetRSEG().index() << " " << chi2 << endl ;
  if(ierr) { cout << "OTrTrackCandidate::AddIfValid(...) matrix inversion error" << endl ;
  } else {
    if(chi2<fgMaxChi2) {
      fClusters.push_back(cluster) ;
      added = true ;
    };
#ifdef USE_ROOT
    static TH1F* tcchi2H1 = new TH1F("tcchi2H1","",1000,0,20) ;
    tcchi2H1->Fill(chi2) ;
#endif
  };
  return added ;
};

#endif // WITH_KALMAN

bool
OTrTrackCandidate::Fit(bool redoTrackSearch)
{ 
  bool success(false) ;
  if(fClusters.size()<2) {
    success = false ;
  } else if(redoTrackSearch || !fTrack) {
    if(fTrack) delete fTrack ;
    fTrack = new OTrTrack(this) ;
    success = fTrack->SearchHitsAndFit() ;
    //if(!success) {
    //  delete fTrack ;
    //  fTrack = 0; 
    //}
  } else {
    success = fTrack->Fit(5) ;
  } 
  if(fTrack) {
    int ierr ;
    HepSymMatrix weightMatrixXY = 
      fTrack->GetCov().sub(1,2).inverse(ierr) ;
    fTargetChi2 =
      weightMatrixXY.similarity(fTrack->GetQ().sub(1,2)) ;
  }
  return success ;
}

#ifdef USE_ROOT
int
OTrTrackCandidate::Draw( TPad* pad, const int color )
{
  if ( fPolyLine3D == 0 ) {
    fPolyLine3D = new TPolyLine3D( fClusters.size() );
    int npoint = 0;
    for(ClusterContainer::iterator clus = fClusters.begin() ;
	clus!=fClusters.end(); clus++) {
      fPolyLine3D->SetPoint(npoint,(*clus)->GetX(),(*clus)->GetY(),(*clus)->GetZ());
      npoint++;
      fPolyLine3D->SetLineColor( color );
    }
  }
  pad->cd();
  //pad->Clear();
  fPolyLine3D->AppendPad();
  //  fPolyLine3D->Draw();
  return fClusters.size();
}
#endif

#ifdef USE_ARTE
ArtePointerRSEG
OTrTrackCandidate::AppendRSEG() {
  ArtePointer<RSEG> rseg(0) ;
  //cout << "OTrTrackCandidate::AppendRSEG()" << endl ;
  //   RSEG newrseg ;
  //   rseg = ArteTable<RSEG>::push_back(newrseg) ;
  //   for(ClusterContainer::const_iterator clusIt = GetClusters().begin() ;
  //       clusIt!= GetClusters().end() ; clusIt++) (*clusIt)->AppendRSEG(rseg) ;
  //   //if(fTrack) rseg = fTrack->AppendRSEG() ;
  //   rseg->fit = 1;
  if(fRangerTrack) rseg = fRangerTrack->storeRseg() ;
  return rseg ;
}

bool CompareMeasurement(const ArtePointer<HITB>& lhs, const ArtePointer<HITB>& rhs) {
  return lhs->q > rhs->q ;
};

bool
OTrTrackCandidate::RangerFit()
{
  static float fgMinChi2Prob = 0.01 ;
  static bool rangerInited = false ;
  if( !rangerInited ) {
    rangerInited = true;
    // for( ArteTable<GEDE>::iterator gede = ArteTable<GEDE>::begin();
    // 	 gede != ArteTable<GEDE>::end(); ++gede ) {
    //       (*gede)->ms = 3.6e-7;
    //     }
    //rangerGeom::reset();
    //rangerGeom::build();
    //rangerGeom::sort();
    rangerGeom::update();
  }
  bool success = false ;
  vector<ArtePointer<RSEG> > rsegVec;
  for(ClusterContainer::const_iterator clusIt = GetClusters().begin() ;
      clusIt!= GetClusters().end() ; clusIt++) rsegVec.push_back((*clusIt)->GetRSEG()) ; 
  if(fRangerTrack) delete fRangerTrack ;
  fRangerTrack = new rangerTrack( rsegVec, ModeL );
  int fitstat = fRangerTrack->fit(1,0,0,0) ;
  float chi2 = fRangerTrack->totalChiSq() ;
  const rangerState& state = (*(fRangerTrack->last()))->getState() ;
  int ndof   =  fRangerTrack->getNodes().size()-state.vec().num_row() ;
  float chi2prob = prob_(chi2,ndof) ;
  int ierr;
  HepSymMatrix weightMatrixXY = state.cov().sub(1,2).inverse(ierr) ;
  fTargetChi2 = weightMatrixXY.similarity(state.vec().sub(1,2)) ;
  success = !fitstat && chi2prob > fgMinChi2Prob ;

#ifdef USE_ROOT
  cout << "Number of clusters: " << GetClusters().size() 
       << " chi2: " << chi2 << " " << ndof << " " << chi2prob << endl ;
  cout << "hitb(mtra): " ;
  for(rangerTrack::rangerNodeIterator nodeIt=fRangerTrack->begin() ;
      nodeIt != fRangerTrack->end() ; nodeIt++)
    if((*nodeIt)->hasHitb()) {
      ArtePointer<HITB> hitb = (*nodeIt)->getHitb() ;
      ArtePointer<MTRA> mtra = 
	ArtePointer<MTRA>((ArtePointer<MIMP>(ArteRelationIterator<MIMP>(hitb->mimp)))->mtra);
      cout << hitb << "(" << mtra << ") " ;
    }      
  cout << endl ;
#endif
  return success ;
}

#endif //USE_ARTE



// void
// OTrTrackCandidate::RangerFit()
// {
//   static bool rangerInited = false;
//   if( !rangerInited ) {
//     rangerInited = true;
//     for( ArteTable<GEDE>::iterator gede = ArteTable<GEDE>::begin();
// 	 gede != ArteTable<GEDE>::end(); ++gede ) {
//       (*gede)->ms = 3.6e-7;
//     }
//     rangerGeom::reset();
//     rangerGeom::build();
//     rangerGeom::sort();
//   }
  
//   ArtePointer<RSEG> rseg = fRSEG ;

//   cout << " track: " << setw(4) << rseg << endl
//        << " x_i : " << setw(8) << setprecision(3) << rseg->xf - rseg->zf * rseg->txf
//        << " y_i : " << setw(8) << setprecision(3) << rseg->yf - rseg->zf * rseg->tyf
//        << " tx_i : " << setw(8) << setprecision(3) << rseg->txf
//        << " ty_i : " << setw(8) << setprecision(3) << rseg->tyf
//        << endl ;
  
//   // Make rangertrack
//    // for(rangerTrack::rangerNodeIterator nodeIt=t->begin() ; nodeIt!=t->end();
// //       nodeIt++) 
// //     if((*nodeIt)->hasHitb())
// //       cout << (*nodeIt)->zNode() << " "
// // 	   << (*((rangerHit*)(*nodeIt))->getMeasurement())(1) << " " 
// // 	   << (*nodeIt)->getHitb()->q << endl ;
// //     else cout << "dead node" << endl ;
  
//   //This is the trick for solving the ambiguities.

//   /*
//   //A. Set all distances to zero
//   HepVector* vec ;
//   for(rangerTrack::rangerNodeIterator nodeIt=t->begin() ; nodeIt!=t->end();
//       nodeIt++) 
//     if((*nodeIt)->hasHitb() && (*nodeIt)->getHitb()->gede->cmp==2 &&
//        (vec=((rangerHit*)(*nodeIt))->getMeasurement()))
//       (*vec)[0] = 0 ;
//       */
//   //B. Fit


//   rangerTrack* t = new rangerTrack( rseg, ModeL );
//   int fitstat=0;
//   fitstat = t->fit(1,0,0,0) ;
//   t->dump() ;
//   t->storeRtra() ;
//   fRangerTrack = t ;

//     if(fitstat<0) return ;
//   do {      
//     prevChi2 = t->totalChiSq() ;
//     fitstat = t->fit() ;
//   }  while ( prevChi2 - t->totalChiSq() > 0.01*prevChi2 ) ;
//   cout << "After first fit: " << t->totalChiSq() << endl ;
  
//   //C. Set hitb sign to current sign of residual
  
//   for(rangerTrack::rangerNodeIterator nodeIt=t->begin() ; nodeIt!=t->end();
//       nodeIt++)
//     if((*nodeIt)->hasHitb() && (*nodeIt)->getHitb()->gede->cmp==2) {
//       rangerHitWire* hitWire = (rangerHitWire*)(*nodeIt) ;
//       if((*(hitWire->getResidual()))[0]>0)
// 	(*nodeIt)->getHitb()->q *= -1 ;
//     };
  
//   //D. Delete track, remake with current hitb values and fit;
//   delete t;
//   t = new rangerTrack( rseg, ModeL );
//   cout << "Before first refit: " << endl ;
//   fitstat = t->fit() ; if(fitstat<0) return ;
//   do {      
//     prevChi2 = t->totalChiSq() ;
//     t->fit() ;
//   }  while ( prevChi2 - t->totalChiSq() > 0.01*prevChi2 ) ;
//   double bestChi2 =  t->totalChiSq() ;
//   cout << "After first refit: " << bestChi2 << endl ;
  
//   // E. sort, such that largest measurements (residuals ?!) come first
//   list<ArtePointer<HITB> > hitWires ; 
//   for(rangerTrack::rangerNodeIterator nodeIt=t->begin() ; nodeIt!=t->end();
//       nodeIt++)
//     if((*nodeIt)->hasHitb() && (*nodeIt)->getHitb()->gede->cmp==2) {
//       hitWires.push_back((*nodeIt)->getHitb()) ;
//     };
//   hitWires.sort(CompareMeasurement) ;
  
//   //F. Flip all signs until right combi found
  
//   for(list<ArtePointer<HITB> >::iterator wireIt=hitWires.begin() ;
//       wireIt != hitWires.end() ; wireIt++) {
//     (*wireIt)->q *= -1 ;
//     rangerTrack* tnew = new rangerTrack( rseg, ModeL );
//     fitstat = tnew->fit() ;
//     if(fitstat>=0) {
//       do {      
// 	prevChi2 = tnew->totalChiSq() ;
// 	tnew->fit() ;
//       }  while ( prevChi2 - tnew->totalChiSq() > 0.01*prevChi2 ) ;
//       prevChi2 = tnew->totalChiSq() ;
//     } else prevChi2=99 ;

//     if(fitstat>=0 && prevChi2<bestChi2) {
//       delete t ;
//       t = tnew ;
//       bestChi2 = prevChi2 ;
// 	cout << "Flipped sign ! Chi2 now " << bestChi2 << endl ;
//     } else {
//       cout << "Sign not flipped ! Chi2 would have been " << prevChi2 << endl ;
// 	delete tnew ;
// 	(*wireIt)->q *= -1 ;
//     };
    
//   };
  
  //ArtePointer<RTRA> rp = t->storeRtra();
//};

// #ifdef USE_ARTE

// ArtePointerRSEG
// OTrTrackCandidate::AppendRSEG() {
//   RSEG r ;
//   const OTrHitCluster* clf = GetClusters()[0] ;
//   const OTrHitCluster* cle = GetClusters()[1] ;

//   r.zf = clf->z() ;
//   r.xf = clf->x() ;
//   r.yf = clf->y() ;
//   r.txf = GetTx() ;
//   r.tyf = GetTy() ;
//   r.pf = 1E-6 ;    // inverse momentum (currently unknown)
//   r.ze = cle->z() ;
//   r.xe = cle->x() ;
//   r.ye = cle->y() ;
//   r.txe = GetTx() ;
//   r.tye = GetTy() ;
//   r.pe = 1E-6 ;    // inverse momentum (currently unknown)
  
//   //cout << fCov << endl;
//   //cout << fCov(1,1) << " " << fCov(1,2) << endl;

//   /*
//     r.zf = 0 ;
//     r.xf = fX ;  // first element
//     r.yf = fY ;
//     r.txf = fTx ;
//     r.tyf = fTy ;
//     r.pf = 1E-6 ;    // inverse momentum (currently unknown)
    
//     r.ze = r.zf ;
//     r.xe = r.xf ;
//     r.ye = r.yf ;
//     r.txe = r.txf ;
//     r.tye = r.tyf ;
//     r.pe = r.pf ;
//   */

//   // covariance matrix, do I fill it correctly (Ivar)
//   for ( int i = 0; i < 15; i++ )  r.cvf[i] = 0 ;
//   for ( int i = 0; i < 15; i++ ) r.cve[i] = r.cvf[i] ;

//   /*
//     // x row 
//     r.cvf[0]=0; r.cvf[1]=0; r.cvf[3]=0; 
//     r.cvf[6]=0; r.cvf[10]=0;  
//     // y row
//     r.cvf[2]=0; r.cvf[4]=0; 
//     r.cvf[7]=0; r.cvf[11]=0; 
//     // tx row
//     r.cvf[5]=0; 
//     r.cvf[8]=0; r.cvf[12]=0;
//     // ty row
//     r.cvf[9]=0; r.cvf[13]=0; 
//     // p row 
//     r.cvf[14]=0;
//   */

//   r.time = 0 ;
//   r.dtim = 0 ;

//   r.chi = 0 ;
//   r.fit = 0 ;
//   r.cmp = 2 ;   // that's us

//   /*
//     r.phitb = 0 ;
//     r.prtra = 0 ;
//     r.prpnt = 0 ;
//     r.pmtra = 0 ;
//     r.prhit = 0 ;
//     r.prccl = 0 ;
//     r.prbeg = 0 ;
//     r.prend = 0 ;
//   */

// <<<<<<< OTrTrackCandidate.C
//   fRSEG = ArteTable<RSEG>::push_back( r ) ;
  
//   /*
//     Put hits in vector, sort them, ens

//     fHits.sort(not2( compareCall( &OTrHit::GetZ ) ) ) ;
//   for( list<OTrHit*>::iterator hitIt=fHits.begin() ; 
//        hitIt!=fHits.end() ; hitIt++ ) {
//     ArtePointerHITB hitbp = (*hitIt)->GetHITB();
//     if ( hitbp == 0 ) {
//       cout << "WARNING in OTrTrackCandidate::AppendRSEG(), "
// 	   << "pointer to HITB bank is zero" << endl;
//     } else {
      
//       // Establish relation between HITB and RSEG
//       createRelation(fRSEG,hitbp,&RSEG::hitb) ;

//       // Construct RHIT
//       RHIT rhit;
//       // I have to fill something here, but what exactly ???
//       rhit.x = hitbp->x;
//       rhit.y = (*hitIt)->GetWire()->GetMiddlePoint().y() ;//(*clIt)->y();
//       rhit.tx = GetTx();
//       rhit.ty = GetTy();
//       rhit.p = 1E-6;
//       rhit.q = 0.123;
      
//       ArteTable<RHIT>::push_back( rhit ) ;
//       ArtePointer<RHIT> rhitp( ArteTable<RHIT>::nrow() ) ;
      
//       createRelation(fRSEG,rhitp,&RSEG::rhit) ;
//       createRelation(hitbp,rhitp,&HITB::rhit) ;
// =======
//   ArtePointer<RSEG> fRSEG = ArtePointer<RSEG>::append( r );
   
//   for ( vector<const OTrHitCluster*>::const_iterator clIt=GetClusters().begin() ; 
// 	clIt != GetClusters().end() ; clIt++) {
//     //cout << (*clIt)->GetHits().size() << endl; 
//     for (vector<OTrHit*>::const_iterator hitIt= (*clIt)->GetHits().begin() ;
// 	 hitIt != (*clIt)->GetHits().end() ; hitIt++) {
//       //      (*hitIt)->DumpHITB();
//       ArtePointerHITB hitbp = (*hitIt)->GetHITB();
//       if ( hitbp == 0 ) {
// 	cout << "WARNING in OTrTrackCandidate::AppendRSEG(), "
// 	     << "pointer to HITB bank is zero" << endl;
//       } else {
// 	// Establish relation between HITB and RSEG
// 	unsigned int nhit = hitbp.index();
// 	unsigned int nseg = fRSEG.index();
// 	mmbrel_(ArtePointer<RSEG_HITB>::getLink().getRef(),nseg,nhit);
// 	// Construct RHIT
// 	RHIT rhit;
// 	// I have to fill something here, but what exactly ???
//         rhit.x = hitbp->x;
// 	rhit.y = (*clIt)->y();
// 	rhit.tx = fTx;
// 	rhit.ty = fTy;
// 	rhit.p = 1E-6;
// 	rhit.q = 0.123;
// 	rhit.rtra = 0;
// 	ArtePointer<RHIT> rhitp = ArtePointer<RHIT>::append(rhit) ;  
// 	unsigned int nrhit = rhitp.index() ;
// 	mmbrel_(ArtePointer<HITB_RHIT>::getLink().getRef(),nhit,nrhit) ;
// 	mmbrel_(ArtePointer<RSEG_RHIT>::getLink().getRef(),nseg,nrhit) ;
//       }
// >>>>>>> 1.10
//     }
//   }
//   */
//   return fRSEG ;
// }
// OTrTrackCandidate::OTrTrackCandidate(const OTrSuperCluster* cl1, const OTrSuperCluster* cl2) :
//   OTrTrackDescription4D(), fRangerTrack(0), fTrack(0), fPolyLine3D(0)
// {
//   fClusters.push_back(cl1) ;
//   fClusters.push_back(cl2) ;
//   float z1=cl1->GetZ(); float x1=z1*cl1->GetTx(); float y1=z1*cl1->GetTy();
//   float z2=cl2->GetZ(); float x2=z2*cl2->GetTx(); float y2=z2*cl2->GetTy(); 
  
//   const double zco = 0;
  
//   HepVector Q(4) ;
//   Q[2] = (x2-x1)/(z2-z1) ; //tx
//   Q[3] = (y2-y1)/(z2-z1) ; //ty
//   Q[0] = x1 + (zco-z1)*Q[2] ; //x
//   Q[1] = y1 + (zco-z1)*Q[3] ; //y
//   SetQ(Q) ;
//   //cout << Q << cl1->GetTx() << " " << cl2->GetTx() << endl ;
//   HepMatrix dQdp1(4,2,0) ;
//   dQdp1[0][0] = z2/(z2-z1) ; dQdp1[0][1] = 0         ; 
//   dQdp1[1][0] = 0          ; dQdp1[1][1] = z2/(z2-z1); 
//   dQdp1[2][0] = -1/(z2-z1) ; dQdp1[2][1] = 0         ; 
//   dQdp1[3][0] = 0          ; dQdp1[3][1] = -1/(z2-z1); 
						      
//   HepMatrix dQdp2(4,2,0) ;			      
//   dQdp2[0][0] = z1/(z1-z2) ; dQdp2[0][1] = 0         ; 
//   dQdp2[1][0] = 0          ; dQdp2[1][1] = z1/(z1-z2); 
//   dQdp2[2][0] = 1/(z1-z2)  ; dQdp2[2][1] = 0         ;
//   dQdp2[3][0] = 0          ; dQdp2[3][1] = -1/(z1-z2);
  
//   SetCov(z1*z1*cl1->GetCov().similarity(dQdp1) +
// 	 z2*z2*cl2->GetCov().similarity(dQdp2)) ;
//   int err;
  
//   HepSymMatrix weightMatrixXY = GetCov().sub(1,2).inverse(err) ;
//   fTargetChi2 = weightMatrixXY.similarity(GetQ().sub(1,2))/2 ;
// #ifdef USE_ROOT
//   static TH1F* minusLogTargetCompH1 = new TH1F("minusLogTargetCompH1","",100,0,100) ;
//   minusLogTargetCompH1->Fill(fTargetChi2) ;
// #endif
// } ;
