#ifndef __OTR_PARAMETERIZEDRTRELATION_HH__
#define __OTR_PARAMETERIZEDRTRELATION_HH__

#ifndef __SGI_STL_VECTOR
#include <vector>
#endif
#ifndef __OTR_FUNCTIONVALUE_HH__
#include "OTrFunctionValue.hh"
#endif
#ifndef __OTR_RTRELATION_HH__
#include "OTrRTRelation.hh"
#endif
#ifndef _Vector_H_
#include "CLHEP/Matrix/Vector.h"
#endif
#ifndef _SYMMatrix_H_
#include "CLHEP/Matrix/SymMatrix.h"
#endif

class TGraphErrors ;

class OTrParameterizedRTRelation : public OTrRTRelation
{
public:
  // enumerates, typedefs
  enum ParameterizedRTRelationType { kSpline, kChebychev } ;

  // constructors, destructor, assignment
  OTrParameterizedRTRelation() ;
  OTrParameterizedRTRelation(int noTpars, int noSigma2pars, double rmin, double rmax) ;
  virtual ~OTrParameterizedRTRelation() ;
  
  // setters and getters
  const HepVector& GetTParameters() const { return fTParameters ; } 
  const HepVector& GetSigma2Parameters() const { return fSigma2Parameters ; } 
  const HepSymMatrix& GetTCov() const { return fTCov ; } 
  const HepSymMatrix& GetSigma2Cov() const { return fSigma2Cov ; } 
  //static ParameterizedRTRelationType GetType() { return fgType; }
  const vector<HepVector>* GetdTdParTable() const { return &tdTdParTable ; } ;
  static void SetCorrectionLevel(int level=1) { fgCorrectionLevel=level; }
  static int GetCorrectionLevel() { return fgCorrectionLevel ; } 

  // other inline methods
  int GetNoTParameters() const { return GetTParameters().num_row() ; } ;
  int GetNoSigma2Parameters() const { return GetSigma2Parameters().num_row() ; } ;
  inline const HepVector& GetdTdPar(HepDouble) const ;
  inline const HepVector& GetdSigma2dPar(HepDouble) const ;

  // other methods
  void UpdateTParameters(const HepVector& = 0, const HepSymMatrix& = 0) ;
  void UpdateSigma2Parameters(const HepVector& = 0, const HepSymMatrix& = 0) ;
  void SetTParameters(const HepVector&, const HepSymMatrix& = 0) ;
  void SetSigma2Parameters(const HepVector&, const HepSymMatrix& = 0) ;
  void FillBasisFuncTables() ;
  HepVector CalcdTdPar(HepDouble) const ;
  HepVector CalcdSigma2dPar(HepDouble) const ;
  
#ifdef __OTR_CLASSDEF__
  ClassDef(OTrParameterizedRTRelation,0)
#endif
#ifdef USE_ROOT
  TGraphErrors *GetTGraph(int nPoints) ;
  TGraphErrors *GetSigmaGraph(int nPoints) ;
#endif

  // declared by base class
  virtual void Read(istream&) ; // Read from inputstream
  virtual void Write(ostream&) const ; // Writes the parameters to an outputstream
  virtual void Write(const char* name) const = 0 ;
  virtual OTrFunctionValue CalcFunctionT(const HepDouble R) const ;
  virtual OTrFunctionValue CalcFunctionSigma2(const HepDouble R) const ;
  virtual OTrRTResult CalcRTResult(const HepDouble R) const ;

  // purely virtual
  virtual vector<OTrFunctionValue>  CalcBasisFuncVector(HepDouble, int) const = 0;
  virtual void SetUniformSigma(HepDouble) = 0;
  virtual void CorrectT0() = 0;
  virtual void RepairTParameters() = 0;
  virtual void RepairSigma2Parameters() = 0;

private:
  // private methos
  OTrFunctionValue CalcFunctionValue(HepDouble,HepDouble,HepDouble,
				  const HepVector&) const ;
  HepVector CalcParDeriv(HepDouble,HepDouble,HepDouble,int) const ;

  // data members
  HepVector fTParameters ;      // t(r) parameters 
  HepVector fSigma2Parameters ; // V(r) parameters
  HepSymMatrix fTCov ;          // cov matrix for t(r) parameters 
  HepSymMatrix fSigma2Cov ;     // cov matrix for V(r) parameters 
  static int fgCorrectionLevel;
  //static const ParameterizedRTRelationType fgType ;

  // temporary data members
  vector<HepVector> tdTdParTable;      // table with t(r) basis functions
  vector<HepVector> tdSigma2dParTable; // table with V(r) basis functions
} ;

/*********** inlines *********************/

inline OTrFunctionValue
OTrParameterizedRTRelation::CalcFunctionT(const HepDouble R) const {
  return CalcFunctionValue(R,GetRMin(),GetRMax(),fTParameters) ; } 

inline OTrFunctionValue
OTrParameterizedRTRelation::CalcFunctionSigma2(const HepDouble R) const {
  return CalcFunctionValue(R,GetRMin(),GetRMax(),fSigma2Parameters) ; }

inline const HepVector&
OTrParameterizedRTRelation::GetdSigma2dPar(const HepDouble R) const {
  const int noBins = tdSigma2dParTable.size();
  int rbin = int(noBins * (R-GetRMin())/(GetRMax()-GetRMin())) ;
  if(rbin<0) rbin=0;
  if(rbin>noBins-1) rbin=noBins-1 ;
  return tdSigma2dParTable[rbin] ;
}

inline const HepVector&
OTrParameterizedRTRelation::GetdTdPar(const HepDouble R) const {
  const int noBins = tdTdParTable.size();
  int rbin = int(noBins * (R-GetRMin())/(GetRMax()-GetRMin())) ;
  if(rbin<0) rbin=0;
  if(rbin>noBins-1) rbin=noBins-1 ;
  return tdTdParTable[rbin] ;
}

#endif











