#ifndef __OTR_RTRELATION_HH__
#define __OTR_RTRELATION_HH__

#ifndef __OTR_FUNCTIONVALUE_HH__
#include "OTrFunctionValue.hh"
#endif
#ifndef __OTR_OBSERVABLE_HH__
#include "OTrObservable.hh"
#endif
#ifndef __SGI_STL_VECTOR
#include <vector>
#endif
#ifdef __OTR_CLASSDEF__
 #ifndef ROOT_Rtypes
 #include <Rtypes.h>
 #endif
#endif

class OTrRTResult ;

/** Interface base class for the rtrelation. Retrieval via tables for fast access. **/

class OTrRTRelation
{
public:
  // constructors, destructor, assignment
  OTrRTRelation() ;
  OTrRTRelation(double rmin, double rmax) ;
  virtual ~OTrRTRelation() ;

  // setters and getters
  HepDouble GetT0() const { return fT0; } 
  HepDouble GetRMin() const { return fRMin ; } 
  HepDouble GetRMax() const { return fRMax ; } 
  static int GetNTableBins() { return fgNTableBins ; }
  void SetRMax(const HepDouble rmax) { fRMax = rmax ; }  
  void SetT0(const HepDouble T0) { fT0 = T0 ; }  
  void SetRMin(const HepDouble rmin) { fRMin = rmin ; } 

  // other inline methods
  inline HepDouble GetR(const HepDouble time) const ;
  inline const OTrFunctionValue& GetT(const HepDouble) ;
  inline const OTrFunctionValue& GetSigma2(const HepDouble) const ;

  // other methods
  OTrDObservable GetObservedT(const HepDouble) ;
  HepDouble GetR(const int rbin) const { return GetRMin() + tDeltaR*rbin ; } ;
  inline int GetRBin(const HepDouble) const ;
  void FillTTable() ;
  void FillSigma2Table() ;
  void DumpTable() ;
  virtual void Read(istream&) ;
  virtual void Write(ostream&) const ;
  virtual void Write(const char* filename) const = 0 ;
  
  // purely virtual methods
  virtual OTrFunctionValue CalcFunctionT(const HepDouble R) const = 0 ;
  virtual OTrFunctionValue CalcFunctionSigma2(const HepDouble R) const = 0 ;
  virtual OTrRTResult CalcRTResult(const HepDouble R) const = 0 ;

#ifdef __OTR_CLASSDEF__
  ClassDef(OTrRTRelation,0)
#endif

private:
  HepDouble fRMin ; // the minimum distance to wire 
  HepDouble fRMax ; // the maximum distance to wire 
  HepDouble fT0;    // global shift in tzero (obsolete)
  static int fgNTableBins ; // number of bins in tables
  // temporary data :
  HepDouble tTRMax ;  // time belonging to the minimum distance to wire
  HepDouble tTRMin ;  // time belonging to the maximum distance to wire
  HepDouble tDeltaT ; // binwidth for R(t) table
  HepDouble tDeltaR ; // binwidth for T(r) and Sigma2(r) table
  vector<OTrFunctionValue> tTTable ;       // T(r) table
  vector<OTrFunctionValue> tSigma2Table ;  // Sigma2(t) table
  vector<float> tRTable ;                  // R(t) table
} ;

/*************** inlines ********************/


inline int OTrRTRelation::GetRBin(const HepDouble R) const {
  int rbin = int((R-GetRMin())/tDeltaR) ;
  if(rbin <=0) return 0;
  if(rbin >=GetNTableBins()-1) return GetNTableBins()-1 ;
  return rbin ;
}

inline HepDouble
OTrRTRelation::GetR(const HepDouble tm) const {
  int tbin = int((tm-tTRMin)/tDeltaT) ;
  if(tbin<=0) return GetRMin() ;
  else if(tbin>=(int)tRTable.size()) return GetRMax() ;
  return tRTable[tbin] ;
};

inline const OTrFunctionValue&
OTrRTRelation::GetT(const HepDouble R) {
  int rbin = int((R-GetRMin())/tDeltaR) ;
  if(rbin <=0) return tTTable[0] ;
  if(rbin >=GetNTableBins()) {
      const OTrFunctionValue& validT = tTTable[GetNTableBins()-1] ; 
      OTrFunctionValue& T = tTTable[GetNTableBins()] ; //overflowbin	
      T.value = validT.value + (R-fRMax)*validT.ddx ;
      //T.ddx = validT.ddx ;
      //T.d2dx2 = 0;
      return T ;
    }
  return tTTable[rbin] ;
};

inline const OTrFunctionValue&
OTrRTRelation::GetSigma2(const HepDouble R) const {
  int rbin = int((R-GetRMin())/tDeltaR) ;
  if(rbin <=0) return tSigma2Table[0] ;
  if(rbin >=(int)tSigma2Table.size()) return tSigma2Table.back() ;
  return tSigma2Table[rbin] ;
};

#endif







