#ifndef TRACK_DESCRIPTION
#define TRACK_DESCRIPTION

#ifndef _Vector_H_
#include "CLHEP/Matrix/Vector.h"
#endif
#ifndef _SYMMatrix_H_
#include "CLHEP/Matrix/SymMatrix.h"
#endif
#ifdef USE_ROOT
 #ifndef ROOT_Rtypes
 #include <Rtypes.h>
 #endif
#endif

/** OTrTrackDescription is a base class for describing tracks. It constains a vector of parameters and a corresponding covariance matrix. **/

class OTrTrackDescription 
{
public:
  virtual const HepVector& GetQ() const = 0 ;
  virtual const HepSymMatrix& GetCov() const = 0 ;
  virtual void SetQ(const HepVector& q ) = 0 ;
  virtual void SetCov(const HepSymMatrix& cov) = 0 ;
  virtual const HepDouble GetTx() const = 0 ;
  virtual const HepDouble GetTy() const = 0 ;
  virtual const HepDouble GetX() const = 0 ;
  virtual const HepDouble GetY() const = 0 ;
  virtual const HepDouble GetZ() const = 0 ;
  virtual const int GetDim() const = 0 ;
  virtual ~OTrTrackDescription() {} ;
#ifdef USE_ROOT
  //ClassDef(OTrTrackDescription,0) 
#endif
} ;

class OTrTrackDescriptionBase : public virtual OTrTrackDescription, public HepVector
{
private:
  ///
  HepSymMatrix fCov;
public:
  OTrTrackDescriptionBase(int dim=4) ;
  OTrTrackDescriptionBase(const HepVector&, const HepSymMatrix& = 0) ;
  virtual const HepVector& GetQ() const {return *this ; } ;
  virtual const HepSymMatrix& GetCov() const { return fCov; } ;
  virtual void SetCov(const HepSymMatrix& cov) { fCov=cov ; } ;
  virtual void SetQ(const HepVector& q ) { HepVector::operator=(q) ; } ; 
  virtual const int GetDim() const { return num_row() ; } ;
  virtual ~OTrTrackDescriptionBase() ;
#ifdef USE_ROOT
  //ClassDef(OTrTrackDescriptionBase,0) 
#endif
} ;

class OTrTrackDescription2D : public virtual OTrTrackDescription, 
			      private OTrTrackDescriptionBase
{
public:
  OTrTrackDescription2D() : OTrTrackDescriptionBase(2) {} ;
  OTrTrackDescription2D(const HepVector& q, const HepSymMatrix& cov= 0) : 
    OTrTrackDescriptionBase(q,cov) {} ;
  virtual ~OTrTrackDescription2D() {} ;
  
  // Declared by interface class
  using OTrTrackDescriptionBase::GetQ   ;
  using OTrTrackDescriptionBase::GetCov ;
  using OTrTrackDescriptionBase::SetQ   ;
  using OTrTrackDescriptionBase::SetCov ;
  using OTrTrackDescriptionBase::GetDim ;
  virtual const HepDouble GetX() const { return 0 ; } ;
  virtual const HepDouble GetY() const { return 0 ; } ;
  virtual const HepDouble GetZ() const { return 0 ; } ;
  virtual const HepDouble GetTx() const { return HepVector::operator[](0) ; } ;
  virtual const HepDouble GetTy() const { return HepVector::operator[](1) ; } ; 

  // Extra methods
  void SetTx(const HepDouble tx) { HepVector::operator[](0) = tx ; } ;
  void SetTy(const HepDouble ty) { HepVector::operator[](1) = ty ; } ;
#ifdef USE_ROOT
  //ClassDef(OTrTrackDescription2D,0) 
#endif
};

class OTrTrackDescription4D : private OTrTrackDescriptionBase,
			      public virtual OTrTrackDescription 
{
public:
  OTrTrackDescription4D() : OTrTrackDescriptionBase(4) {} ;  
  OTrTrackDescription4D(const HepVector& q, const HepSymMatrix& cov= 0) : 
    OTrTrackDescriptionBase(q,cov) {} ;
  virtual ~OTrTrackDescription4D() {} ;

  // Declared by interface class
  using OTrTrackDescriptionBase::GetQ   ;
  using OTrTrackDescriptionBase::GetCov ;
  using OTrTrackDescriptionBase::SetQ   ;
  using OTrTrackDescriptionBase::SetCov ;
  using OTrTrackDescriptionBase::GetDim ;
  virtual const HepDouble GetX() const  { return HepVector::operator[](0) ; } ;
  virtual const HepDouble GetY() const  { return HepVector::operator[](1) ; } ;
  virtual const HepDouble GetZ() const  { return 0 ; } ; 
  virtual const HepDouble GetTx() const { return HepVector::operator[](2) ; } ;
  virtual const HepDouble GetTy() const { return HepVector::operator[](3) ; } ; 

  // Extra methods
  void SetX(const HepDouble x)   { HepVector::operator[](0) = x  ; } ;
  void SetY(const HepDouble y)   { HepVector::operator[](1) = y  ; } ;
  void SetTx(const HepDouble tx) { HepVector::operator[](2) = tx ; } ;
  void SetTy(const HepDouble ty) { HepVector::operator[](3) = ty ; } ;
#ifdef USE_ROOT
  //ClassDef(OTrTrackDescription4D,0) 
#endif
};

#endif




